package org.processmining.xlu.experiment.discovery;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.plugins.etm.parameters.ETMParam;
import org.processmining.plugins.etm.parameters.ETMParamFactory;
import org.processmining.plugins.etm.ui.plugins.ETMwithoutGUI;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.conversion.ProcessTree2Petrinet;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.InvalidProcessTreeException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.NotYetImplementedException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.PetrinetWithMarkings;

public class ExpDiscoveryETM implements ExpDiscoveryAlgorithm {

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log) {
		XEventClassifier classifier = new XEventNameClassifier();
		return minePetrinet(context, log, classifier);

	}

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log,
			XEventClassifier classifier) {
//		ETMParam etmParam = ETMParamFactory.buildStandardParam(log, context);
		ETMParam etmParam = ETMParamFactory.buildParam(
				log,
				context,
				ETMParamFactory.STD_POPSIZE,
				ETMParamFactory.STD_ELITE_COUNT,
				ETMParamFactory.STD_RANDOMTREES_NR,
				ETMParamFactory.STD_CROSSOVER_CHANCE,
				ETMParamFactory.STD_MUTATION_RANDOM_CHANCE,
				ETMParamFactory.STD_PREVENT_DUPLICATES, // TODO: ask...
				ETMParamFactory.STD_MAXIMUM_NR_GENERATIONS,
				ETMParamFactory.STD_TARGET_FITNESS,
//				ETMParamFactory.STD_REPLAYFITNESS_WEIGHT,
				100, ETMParamFactory.STD_REPLAYFITNESS_MAXF,
				ETMParamFactory.STD_REPLAYFITNESS_MAXTIME,
				ETMParamFactory.STD_PRECISION_WEIGHT,
				ETMParamFactory.STD_GENERALIZATION_WEIGHT,
				ETMParamFactory.STD_SIMPLICITY_WEIGHT);
//		etmParam.addTerminationCondition(new GenerationCount(100));
		etmParam.addTerminationConditionMaxGen(50);
//		etmParam.getFitnessEvaluator().getFitness(candidate, population)

		etmParam.setLogModulo(100);

		ProcessTree tree = ETMwithoutGUI.minePTWithParameters(context, log, classifier,
				etmParam);

		MarkedPetrinet net = convertProcessTreeToPetrinet(context, tree);

		return net;
	}

	public static MarkedPetrinet convertProcessTreeToPetrinet(PluginContext context,
			ProcessTree tree) {
		PetrinetWithMarkings pn = null;
		try {
			pn = ProcessTree2Petrinet.convert(tree);
		} catch (NotYetImplementedException e) {
			e.printStackTrace();
		} catch (InvalidProcessTreeException e) {
			e.printStackTrace();
		}

		if (pn.initialMarking == null) {
			System.err.println("initial marking is null");
		}
		if (pn.finalMarking == null) {
			System.err.println("final marking is null");
		}

		MarkedPetrinet net = new MarkedPetrinet(pn.petrinet, pn.initialMarking,
				pn.finalMarking);

		//create Petri net connections
		context.addConnection(new InitialMarkingConnection(net.getPetrinet(), net
				.getInitialMarking()));
		context.addConnection(new FinalMarkingConnection(net.getPetrinet(), net
				.getFinalMarking()));

		return net;
	}

	public String toString() {
		return "ETM";
	}
}
