package org.processmining.xlu.experiment.discovery;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.causalactivitygraph.models.CausalActivityGraph;
import org.processmining.causalactivitygraphcreator.parameters.ConvertCausalActivityMatrixToCausalActivityGraphParameters;
import org.processmining.causalactivitygraphcreator.plugins.ConvertCausalActivityMatrixToCausalActivityGraphPlugin;
import org.processmining.causalactivitymatrix.models.CausalActivityMatrix;
import org.processmining.causalactivitymatrixminer.miners.MatrixMiner;
import org.processmining.causalactivitymatrixminer.miners.MatrixMinerParameters;
import org.processmining.causalactivitymatrixminer.miners.impl.HAFMiniMatrixMiner;
import org.processmining.framework.connections.Connection;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.hybridilpminer.models.lp.configuration.factories.LPMinerConfigurationFactory;
import org.processmining.hybridilpminer.models.lp.configuration.interfaces.LPMinerConfiguration;
import org.processmining.hybridilpminer.models.lp.configuration.parameters.DiscoveryStrategy;
import org.processmining.hybridilpminer.models.lp.configuration.parameters.DiscoveryStrategyType;
import org.processmining.hybridilpminer.models.lp.configuration.parameters.LPConstraintType;
import org.processmining.hybridilpminer.models.lp.miner.factories.LPMinerFactory;
import org.processmining.hybridilpminer.models.lp.miner.interfaces.LPMiner;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;

public class ExpDiscoveryHyberILP implements ExpDiscoveryAlgorithm {

	private boolean isDummyEndAdded = true;

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log) {
		return minePetrinet(context, log, new XEventNameClassifier());
	}

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log,
			XEventClassifier classifier) {

		if (isDummyEndAdded) {
			log = ExpDiscoveryDummyEndUtil.addDummyEndEvents(log, classifier);
		}

		// Create configuration
		LPMinerConfiguration configuration = LPMinerConfigurationFactory
				.createDefaultConfiguration();
		configuration.setEvenClassifier(classifier);
		// This is necessary to guaranty replay fitness 1 when having a single final marking.
		configuration.getLPConstraintTypes().add(LPConstraintType.EMPTY_AFTER_COMPLETION);
		
		DiscoveryStrategy discoveryStrategy = new DiscoveryStrategy(
				DiscoveryStrategyType.CAUSAL);
		CausalActivityGraph graph = createCausalActivityGraph(context, log, classifier);
		discoveryStrategy.setCausalActivityGraph(graph);
		configuration.setDiscoveryStrategy(discoveryStrategy);

		// Run miner
		LPMiner miner = LPMinerFactory.createLPMiner(log, configuration);
		miner.run();
		Pair<Petrinet, Marking> netAndMarking = miner.synthesizeNet();

		Petrinet net = netAndMarking.getFirst();
		Connection iMarking = new InitialMarkingConnection(netAndMarking.getFirst(),
				netAndMarking.getSecond());
		context.getConnectionManager().addConnection(iMarking);

		MarkedPetrinet markednet = null;
		if (isDummyEndAdded) {
			Marking finalMarking = ExpDiscoveryDummyEndUtil.removeDummyEndTransitionAndReturnFinalMarking(net);
			context.getConnectionManager().addConnection(
					new FinalMarkingConnection(net, finalMarking));

			markednet = new MarkedPetrinet(net, netAndMarking.getSecond(), finalMarking);
		} else {
			markednet = new MarkedPetrinet(net, netAndMarking.getSecond());
		}

		return markednet;
	}





	public CausalActivityGraph createCausalActivityGraph(PluginContext context, XLog log,
			XEventClassifier classifier) {
		MatrixMiner miner = new HAFMiniMatrixMiner();
		MatrixMinerParameters minerParameters = new MatrixMinerParameters(log);
		minerParameters.setClassifier(classifier);
		CausalActivityMatrix matrix = miner.mineMatrix(context, log, minerParameters);
		ConvertCausalActivityMatrixToCausalActivityGraphPlugin creator = new ConvertCausalActivityMatrixToCausalActivityGraphPlugin();
		ConvertCausalActivityMatrixToCausalActivityGraphParameters creatorParameters = new ConvertCausalActivityMatrixToCausalActivityGraphParameters();
		creatorParameters.setZeroValue(miner.getZeroValue());
		creatorParameters.setConcurrencyRatio(miner.getConcurrencyRatio());
		creatorParameters.setIncludeThreshold(miner.getIncludeThreshold());
		CausalActivityGraph graph = creator.run(context, matrix, creatorParameters);
//		DiscoveryStrategy strategy = new DiscoveryStrategy(getMiningStrategy());
//		LPFilter filter = new LPFilter(getFilter(), getFilter().getDefaultThreshold());
//		LPMinerConfiguration configuration = LPMinerConfigurationFactory
//				.customConfiguration(eventLog, getEngineType(),
//						parameters.getClassifier(), strategy, getConstraints(),
//						getObjectiveType(), getVariableType(), filter, false);
//		configuration.getDiscoveryStrategy().setCausalActivityGraph(graph);
//		LPMiner lpMiner = LPMinerFactory.createLPMiner(eventLog, configuration, null);
//		lpMiner.run();
		return graph;

	}

	
	public String toString(){
		return "ILP";
	}
}
