package org.processmining.xlu.experiment.discovery;

import java.util.Iterator;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.Xor;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.conversion.ProcessTree2Petrinet;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.InvalidProcessTreeException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.NotYetImplementedException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.PetrinetWithMarkings;
import org.processmining.xlu.relabel.plugin.matching.ReplacerIMLocal;

import gnu.trove.set.hash.THashSet;

public class ExpDiscoveryInductiveMiner implements ExpDiscoveryAlgorithm {

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log) {

		return minePetrinet(context, log, new XEventNameClassifier());
	}

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log,
			XEventClassifier classifier) {
		ProcessTree tree = mineTree(log, classifier);

		PetrinetWithMarkings pn = null;
		try {
			pn = ProcessTree2Petrinet.convert(tree);
		} catch (NotYetImplementedException e) {
			e.printStackTrace();
		} catch (InvalidProcessTreeException e) {
			e.printStackTrace();
		}
		Object[] nets = new Object[] { pn.petrinet, pn.initialMarking, pn.finalMarking };
		if (context != null) {
			context.addConnection(
					new InitialMarkingConnection((Petrinet) nets[0], (Marking) nets[1]));
			context.addConnection(
					new FinalMarkingConnection((Petrinet) nets[0], (Marking) nets[2]));
		}
		return new MarkedPetrinet((Petrinet) nets[0], (Marking) nets[1],
				(Marking) nets[2]);
	}

	public ProcessTree mineTree(XLog log, XEventClassifier classifier) {
		MiningParametersIM param = new MiningParametersIM();
		param.setClassifier(classifier);
		ProcessTree tree = IMProcessTree.mineProcessTree(log, param);

		ReplacerIMLocal.changeTransitionLabels(tree);
		postProcessing(tree.getRoot());
		return tree;
	}

	public static void postProcessing(Node node) {
		if (node instanceof Block) {
			if (node instanceof Xor) {
				Set<String> distinctNodes = new THashSet<>();

				Iterator<Node> i = ((Block) node).getChildren().iterator();
				while (i.hasNext()) {
					Node c = i.next(); // must be called before you can call i.remove()
					// Do something
					if (distinctNodes.contains(c.toString())) {
						for (Edge e : c.getIncomingEdges()) {
							((Block) node).removeOutgoingEdge(e);
						}
//						i.remove();
					} else {
						distinctNodes.add(c.toString());
					}
				}
			}

			for (Node c : ((Block) node).getChildren()) {
				postProcessing(c);
			}
		}
	}

	public String toString() {
		return "IM";
	}

}
