package org.processmining.xlu.experiment.discovery;

import java.util.ArrayList;
import java.util.List;

import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.semantics.petrinet.Marking;

public class MarkedPetrinet {

	private Petrinet petrinet;
	private Marking initialMarking;
	private List<Marking> finalMarkings;

	public MarkedPetrinet(Petrinet petrinet, Marking initialMarking) {
		this.initialMarking = initialMarking;
		this.petrinet = petrinet;
		this.finalMarkings = new ArrayList<Marking>();
	}

	public MarkedPetrinet(Petrinet petrinet, Marking initialMarking, Marking[] finalMarkings) {
		this(petrinet, initialMarking);
		for (Marking m : finalMarkings) {
			this.finalMarkings.add(m);
		}
	}

	public MarkedPetrinet(Petrinet petrinet, Marking initMarking, Marking finalMarking) {
		this(petrinet, initMarking);
		this.finalMarkings.add(finalMarking);
	}

	public Petrinet getPetrinet() {
		return petrinet;
	}

	public Marking getInitialMarking() {
		return initialMarking;
	}

	public Marking getFinalMarking() {
		if (!finalMarkings.isEmpty()) {
			return finalMarkings.get(0);
		}
		return null;
	}

	public List<Marking> getFinalMarkings() {
		return finalMarkings;
	}

	public void setFinalMarkings(Marking[] finalMarkings) {
		for (Marking m : finalMarkings) {
			this.finalMarkings.add(m);
		}
	}

	public void loadFinalMarking() {
		Marking m = new Marking();
		for (Place p : petrinet.getPlaces()) {
			if (this.petrinet.getOutEdges(p).isEmpty()) {
				m.add(p);
			}
		}
		this.finalMarkings.add(m);
	}

}
