package org.processmining.xlu.logrepair.plugins;

import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.model.XTrace;

public class DAnnotatedTrace {

	/*
	 * 0 = no deviation 1 = deviating dependency 2 = deviating event
	 */
	private int[] annotation;
	private XTrace trace;
	private int traceIndex;

	private List<Interval> deviatingIntervals = new ArrayList<>();

	public static class Interval {
		public int lowerbound;
		public int upperbound;
		
		public Interval(int lowerbound, int upperbound){
			this.lowerbound = lowerbound;
			this.upperbound = upperbound;
		}
	}
	
	public static class TraceInterval extends Interval {
		public int traceIndex;
		public TraceInterval( int traceIndex, int lowerbound, int upperbound) {
			super(lowerbound, upperbound);
			this.traceIndex = traceIndex;
		}
	}
	
	
	
	public DAnnotatedTrace(Integer traceIndex, XTrace t) {
//		this.trace = t;
		this.traceIndex = traceIndex;
		this.annotation = new int[t.size()];
	}

	public void setEventDeviating(int eventIndex) {
		this.annotation[eventIndex] = 2;

		// Store and update deviating interval.
		updateDeviatingIntervals(eventIndex);

	}

	private void updateDeviatingIntervals(int eventIndex) {
		int coveredByInterval = -1;
		int mergeInterval = -1;
		for (int i = 0; i < deviatingIntervals.size(); i++) {
			Interval deviatingInterval = deviatingIntervals.get(i);

			if (getLowerBound(deviatingInterval) - 1 <= eventIndex
					&& eventIndex <= getUpperBound(deviatingInterval) + 1) {
				if (eventIndex + 1 == getLowerBound(deviatingInterval)) {
					deviatingInterval.lowerbound = eventIndex;
				} else if (getUpperBound(deviatingInterval) == eventIndex - 1) {
					deviatingInterval.upperbound = eventIndex;
				} //else deviatingInterval[0] <= eventIndex && eventIndex <= deviatingInterval[1]){

				if (coveredByInterval < 0) {
					coveredByInterval = i;
				} else {
					mergeInterval = i;
				}
			}
		}

		if (coveredByInterval < 0) {
			deviatingIntervals.add(new Interval(eventIndex, eventIndex));
		}
		if (coveredByInterval >= 0 && mergeInterval >= 0) {
			Interval interval = deviatingIntervals.get(coveredByInterval);
			Interval intervalToBeRemoved = deviatingIntervals.get(mergeInterval);
			if (getUpperBound(interval) >= getLowerBound(intervalToBeRemoved)
					&& getLowerBound(interval) <= getUpperBound(intervalToBeRemoved)) {
				interval.upperbound = getUpperBound(intervalToBeRemoved);
			} else {
				interval.lowerbound = getLowerBound(intervalToBeRemoved);
			}
		}

	}

	private int getUpperBound(Interval deviatingInterval) {
		return deviatingInterval.upperbound;
	}

	private int getLowerBound(Interval deviatingInterval) {
		return deviatingInterval.lowerbound;
	}

	public void setDependencyDeviating(int eventSourceIndex, int eventTargetIndex) {
		//TODO: check already event deviate?
		//TODO: check eventsourceIndex + 1 = eventTargetIndex
		this.annotation[eventSourceIndex] = 1;
		this.annotation[eventTargetIndex] = 1;

		updateDeviatingIntervals(eventSourceIndex);
		updateDeviatingIntervals(eventTargetIndex);

	}

	public List<Interval> getDeviatingIntervals() {
		return this.deviatingIntervals;
		
	}

	public int getTraceIndex() {
		return this.traceIndex;
	}

}
