package org.processmining.xlu.matching.config;

import org.deckfour.xes.classification.XEventNameClassifier;
import org.processmining.partialorder.ptrace.param.PTraceParameter;
import org.processmining.partialorder.ptrace.param.PTraceParameter.PTraceType;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.plugins.alg.ENodeContextComparator;
import org.processmining.xlu.matching.plugins.alg.HeuristicStrategy;
import org.processmining.xlu.matching.plugins.alg.conf.impl.EGMatchConfigurationImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostHeuristicFunctionComplexImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAlgBackTrackingPruningImpl;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAlgGreedyImp;

public class EGMatchConfigurationFactory {

	public static EGMatchConfiguration createDataBasedDefaultConfig() {
		EGMatchConfiguration config = new EGMatchConfigurationImpl();
		config.setClassifier(new XEventNameClassifier());
		config.setPTraceParameter(new PTraceParameter(PTraceType.Data_Dependency));
		CostStrategyConfiguration costConfig = config.getCostStrategy()
				.getCostConfiguration();
		costConfig.setWeightBaseCostForNoMatch(1);
		costConfig.setWeightGlobalStructure(1);
		costConfig.setWeightMatched(1);
		costConfig.setWeightNoMatch(1);
		config.setDist(1);
		config.setMatchingAlgorithm(new MatchingAlgBackTrackingPruningImpl());
		return config;
	}

	public static EGMatchConfiguration createSeqBasedDefaultConfig() {
		EGMatchConfiguration config = new EGMatchConfigurationImpl();
		config.setClassifier(new XEventNameClassifier());
		config.setPTraceParameter(new PTraceParameter(PTraceType.Sequential_Dependency));
		CostStrategyConfiguration costConfig = config.getCostStrategy()
				.getCostConfiguration();
		costConfig.setWeightBaseCostForNoMatch(3);
		costConfig.setWeightGlobalStructure(1);
		costConfig.setWeightMatched(1);
		costConfig.setWeightNoMatch(3);
		config.setDist(1);
		config.setMatchingAlgorithm(new MatchingAlgBackTrackingPruningImpl());
		return config;
	}

	public static EGMatchConfiguration createTimeBasedPOGreedyConfig() {
		EGMatchConfiguration config = new EGMatchConfigurationImpl();
		config.setClassifier(new XEventNameClassifier());
		config.setPTraceParameter(new PTraceParameter(
				PTraceType.Non_Equal_Timestamp_Dependency));
		CostStrategyConfiguration costConfig = config.getCostStrategy()
				.getCostConfiguration();
		costConfig.setWeightBaseCostForNoMatch(3);
		costConfig.setWeightGlobalStructure(1);
		costConfig.setWeightMatched(1);
		costConfig.setWeightNoMatch(3);
		config.setDist(1);

		MatchingAlgGreedyImp matchingAlg = new MatchingAlgGreedyImp();
		matchingAlg.setNodeComparator(new ENodeContextComparator());
		HeuristicStrategy heuristicFunction = new CostHeuristicFunctionComplexImpl();
		matchingAlg.setHeuristicfunction(heuristicFunction);
		config.setMatchingAlgorithm(matchingAlg);
		return config;
	}

	public static EGMatchConfiguration createSeqBasedGreedyConfig() {
		EGMatchConfiguration config = new EGMatchConfigurationImpl();
		config.setClassifier(new XEventNameClassifier());
		config.setPTraceParameter(new PTraceParameter(PTraceType.Sequential_Dependency));
		CostStrategyConfiguration costConfig = config.getCostStrategy()
				.getCostConfiguration();
//		costConfig.setWeightBaseCostForNoMatch(3);
		costConfig.setWeightBaseCostForNoMatch(1);
		costConfig.setWeightGlobalStructure(1);
		costConfig.setWeightMatched(3);
//		costConfig.setWeightNoMatch(3);
		costConfig.setWeightNoMatch(3);
		config.setDist(1);

		MatchingAlgGreedyImp matchingAlg = new MatchingAlgGreedyImp();
		matchingAlg.setNodeComparator(new ENodeContextComparator());
		HeuristicStrategy heuristicFunction = new CostHeuristicFunctionComplexImpl();
		matchingAlg.setHeuristicfunction(heuristicFunction);
		config.setMatchingAlgorithm(matchingAlg);
		return config;
	}

	public static EGMatchConfiguration createSeqBasedGreedyMatchAsMuchAsPossibleConfig() {
		EGMatchConfiguration config = createSeqBasedGreedyConfig();
		config.getCostStrategy().getCostConfiguration().setWeightMatched(1);
		config.getCostStrategy().getCostConfiguration().setWeightNoMatch(10);
		return config;
	}
	
	public static EGMatchConfiguration createHOSPITALBasedGreedyMatchAsMuchAsPossibleConfig() {
		EGMatchConfiguration config = new EGMatchConfigurationImpl();
		config.setClassifier(new XEventNameClassifier());
		config.setPTraceParameter(new PTraceParameter(
				PTraceType.Non_Same_Day_Dependency));
		CostStrategyConfiguration costConfig = config.getCostStrategy()
				.getCostConfiguration();
		costConfig.setWeightMatched(5);
		costConfig.setWeightBaseCostForNoMatch(3);
		costConfig.setWeightGlobalStructure(1);
		costConfig.setWeightNoMatch(3);
		config.setDist(1);

		MatchingAlgGreedyImp matchingAlg = new MatchingAlgGreedyImp();
		matchingAlg.setNodeComparator(new ENodeContextComparator());
		HeuristicStrategy heuristicFunction = new CostHeuristicFunctionComplexImpl();
		matchingAlg.setHeuristicfunction(heuristicFunction);
		config.setMatchingAlgorithm(matchingAlg);
		return config;
	}

}
