package org.processmining.xlu.matching.config.alg;

import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.xlu.matching.config.gui.MappingSelectorDialog;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.RepEGraphMappingAware;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.plugins.EGFusionPlugin;
import org.processmining.xlu.matching.plugins.EGMatchPlugin;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.CostFunctionImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyWeightsImpl;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAllAlgBackTrackingBFImpl;
import org.processmining.xlu.matching.vis.EGraphToPGraphConverterBinaryImpl;
import org.processmining.xlu.pgraph.vis.converter.PGraphConverter;

public class EGMatchUserOrientedConfigAlg {

	private UIPluginContext context;

//	private List<EGMatchConfigDialog> dialogs = new ArrayList<EGMatchConfigDialog>();

	private XLog log;

	private EGMatchConfiguration conf;
	
	private final Map<CostStrategyConfiguration, Integer> rank;

	private ArrayList<RepEGraph> graphlist;
	PGraphConverter<RepEGraph> converter;

	private CostStrategy costfunction = new CostFunctionImpl();
	int numSampleCases = 5;
	int maxTraceSize = 10;
	
	
	public EGMatchUserOrientedConfigAlg(UIPluginContext context, XLog log, final EGMatchConfiguration config) {
		// 1. Randomly select traces
		this.log = log;
		this.conf = config;
		this.context = context;

		// Set algorithm 
		MatchingAllAlgBackTrackingBFImpl alg = new MatchingAllAlgBackTrackingBFImpl();
		alg.setComputeOnlyAllOptMappings(false);
		conf.setMatchingAlgorithm(alg);

		// 2. Generate Configurations according to space
		// 4. For each configurations, assign rank 0. 
		rank = getConfigurationSpace();

		// 3. Generates mappings
		

		
		XLog sampleLog = RandomLogSampleUtil.createRandomSampleLog(log, numSampleCases, maxTraceSize);
		
		EGMatchPlugin matchPlugin = new EGMatchPlugin();
		EGraphsWithMappings resMappings = matchPlugin.replayLog(context, sampleLog, conf);
		EGFusionPlugin plugin = new EGFusionPlugin();
		RepEGraphs graphs = plugin.fuseBinaryGraphs(context, resMappings);
		PLog plog = graphs.getpLog();
		
		graphlist = new ArrayList<>(graphs.getGraphs());
		converter = new EGraphToPGraphConverterBinaryImpl();
		// 5. For each config and each mapping, compute cost

		// 6. show mappings. 

	}
	
	private MappingSelectorDialog getViews(int traceI, int traceJ){
		Map<PartialOrderGraph, RepEGraph> views = new THashMap<PartialOrderGraph, RepEGraph>();
		
		Set<RepEGraph> remove = new THashSet<>();
		for (int i = graphlist.size() - 1; i >= 0; i-- ) {
			RepEGraph g = graphlist.get(i);
			List<Integer> ids = g.getRepresentedGraphIds();
			
			if (ids.contains(traceI) && ids.contains(traceJ)) {
				PartialOrderGraph visgraph = converter.convert(g);
//				PGraphVisView<RepEGraph> view = converter.createPGraphVisView(visgraph, null);
				
				views.put(visgraph, g);
				remove.add(g);
			}
			if(views.size()==3){
				break;
			}
		}
		graphlist.removeAll(remove);
		
		MappingSelectorDialog panel = new MappingSelectorDialog(traceI, traceJ, views, rank);
		return panel;
	}

	public InteractionResult showWizard() {
		// get the first dialog 
		int traceI, traceJ;
		traceI = traceJ = 0;

		MappingSelectorDialog currentDialog;
		while (traceI < this.numSampleCases && traceJ < this.numSampleCases) {
			// Show the first dialog
//			EGMatchConfigDialog currentDialog = dialogs.get(step);
			
			currentDialog = getViews(traceI, traceJ);
			
			InteractionResult res = context.showWizard(currentDialog.getTitle(), traceI == 0 && traceJ == 0, false,
					currentDialog.showDialog());

			// Naviagate the next or previous dialog
			switch (res) {
				case NEXT :
					RepEGraph g = currentDialog.getSelectedRepEGraph();
					if(g != null){
						updateRanking(g, currentDialog.getNotSelectedRepEGraphs());
						
						if(traceJ < numSampleCases - 1){
							traceJ++;
						} else {
							traceJ = 0;
							traceI++;
						}
					} 					
					break;
				case PREV :

					break;
				default :
					return res;
			}
		}
		
		InteractionResult res = context.showWizard("Done", false, true,
				new JPanel());
		
		return res;
	}
	
	public void updateRanking(RepEGraph g, Collection<RepEGraph> others) {
		//		if(g instanceof RepEGraphMappingAware){
		MatchMapping mapping = ((RepEGraphMappingAware) g).getMapping();
		Set<CostStrategyConfiguration> configs = new THashSet<>(rank.keySet());
		for (CostStrategyConfiguration c : configs) {
			costfunction.setCostConfiguration(c);
			int cost = costfunction.getTotalCost(mapping.getSourceGraph(), mapping.getTargetGraph(), mapping);
			boolean upgrade = true;
			for (RepEGraph other : others) {


				MatchMapping otherMapping = ((RepEGraphMappingAware) other).getMapping();
				int otherCost = costfunction .getTotalCost(otherMapping.getSourceGraph(), otherMapping.getTargetGraph(),
						otherMapping);
				if (otherCost <= cost) {
					upgrade = false;
					break;
				}

			}
			int ranklevel = this.rank.get(c);
			if (upgrade) {
				this.rank.put(c, ++ranklevel);
				System.out.println("Config " + c.getName() + " rank up " + ranklevel);
			} else {
				this.rank.put(c, --ranklevel);
				System.out.println("Config " + c.getName() + " rank down " + ranklevel);
			}
		}

		System.out.println(mapping);
	}

	private Map<CostStrategyConfiguration, Integer> getConfigurationSpace() {
		Map<CostStrategyConfiguration, Integer> mapConfig2Rank = new HashMap<CostStrategyConfiguration, Integer>();
		for (int wMatch = 1; wMatch < 3; wMatch++) {
			for (int wNoMatch = 1; wNoMatch < 3; wNoMatch++) {
				global: for (int wGlobal = 1; wGlobal < 3; wGlobal++) {
					for (int wBase = 1; wBase < 6; wBase += 2) {
						CostStrategyConfiguration conf = new CostStrategyWeightsImpl();
						conf.setWeightBaseCostForNoMatch(wBase);
						conf.setWeightGlobalStructure(wGlobal);
						conf.setWeightMatched(wMatch);
						conf.setWeightNoMatch(wNoMatch);
						mapConfig2Rank.put(conf, 0);
					}
				}
			}
		}
		return mapConfig2Rank;
	}

	public EGMatchConfiguration getConfiguration() {
		return this.conf;
	}



}
