package org.processmining.xlu.matching.config.alg;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.Random;

import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;


public class RandomLogSampleUtil {
	
	public static XLog createRandomSampleLog(XLog log, int numTraces) {
		return createRandomSampleLog(log, numTraces, Integer.MAX_VALUE);
	}
	
	public static XLog createRandomSampleLog(XLog log, int numTraces, int maxTraceSize) {
		Random rnd = new Random();
		TIntSet traceIndexSet = new TIntHashSet(numTraces);
		// Prevent infinite loop, sample size has to be smaller than log size
		numTraces = Math.min(log.size(), numTraces);
		for (int i=0; i < numTraces; i++) {			
			boolean isAdded = false;
			while (!isAdded) {
				int index = rnd.nextInt(log.size());
				isAdded = log.get(index).size() <= maxTraceSize && traceIndexSet.add(index);
			}
		}
		return extractTracesByIndex(log, traceIndexSet.toArray());
	}
	
	
	public static XLog extractTracesByIndex(XLog log, int[] traceIndexToKeep) {
		XLog newLog = XFactoryRegistry.instance().currentDefault().createLog(log.getAttributes());
	
		
		for (int traceIndex: traceIndexToKeep) {
			if (traceIndex >= 0) {
				newLog.add(log.get(traceIndex));	
			}
		}
		
		return newLog;		
	}


	

}
