package org.processmining.xlu.matching.config.gui;

import java.util.ArrayList;
import java.util.List;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;

public class EGMatchConfigWizard {

	protected UIPluginContext context;
	protected EGMatchConfiguration config;
	protected XLog log;

	// This list assumes a sequential navigation through the dialogs. 
	protected List<EGMatchConfigDialog> dialogs = new ArrayList<EGMatchConfigDialog>();

	public EGMatchConfigWizard(UIPluginContext context, EGMatchConfiguration conf, XLog log) {
		this.context = context;
		this.config = conf;
		this.log = log;
		// TODO Make dialogs and set dialogs...

		initalizeDialogs();
	}

	protected void initalizeDialogs() {
		EGMatchConfigDialog panel = new XEventClassSelectorPanel(config, log);
		this.dialogs.add(panel);

		panel = new PTraceParameterMatchDialog(context, config);
		this.dialogs.add(panel);

		panel = new CostConfigSelectorPanel(config);
		this.dialogs.add(panel);

		panel = new EGMatchingAlgSelectorDialog(config);
		this.dialogs.add(panel);

	}

	public InteractionResult showWizard() {
		// get the first dialog 
		int step = 0;
		while (step < dialogs.size()) {
			// Show the first dialog
			EGMatchConfigDialog currentDialog = dialogs.get(step);
			InteractionResult res = context.showWizard(currentDialog.getTitle(), step == 0, step == dialogs.size() - 1,
					currentDialog.showDialog());

			// Naviagate the next or previous dialog
			switch (res) {
				case NEXT :
					step++;
					break;
				case PREV :
					step--;
					break;
				default :
					return res;
			}
		}
		return null;
	}

	public EGMatchConfiguration getConfiguration() {
		if (config.getClasses() == null || config.getClasses().size() == 0) {
			XEventClasses classes = XEventClasses.deriveEventClasses(config.getClassifier(), log);
			config.setClasses(classes);
		}
		return config;
	}

}
