package org.processmining.xlu.matching.merge.impl;

import java.util.HashMap;
import java.util.Map;

import org.processmining.xlu.matching.merge.FusionAlg;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.impl.RepEGraphBinaryMappingAwareImpl;

public class FusionAlgBinaryImpl implements FusionAlg {

	public RepEGraph mergeEGraph(EGraph source, EGraph target, MatchMapping mapping) {
		EGraphFactory factory = EGraphFactory.getInstance();

		RepEGraphBinaryMappingAwareImpl newGraph = new RepEGraphBinaryMappingAwareImpl();

		/* Update nodes */
		Map<Integer, Integer> mapOrg2NewId = new HashMap<Integer, Integer>();
		newGraph.addRepresentedGraph(source.getId());

		for (ENode sourceNode : source.getNodes()) {
			ENode newNode = factory.createNode(newGraph, sourceNode);
			newGraph.updateNodeRepresentedNode(newNode.getId(), sourceNode.getGraphId(), sourceNode.getId());
			mapOrg2NewId.put(sourceNode.getId(), newNode.getId());
		}

		newGraph.addRepresentedGraph(target.getId());
		Map<Integer, Integer> mapOrgTarget2NewId = new HashMap<Integer, Integer>();
		for (ENode targetNode : target.getNodes()) {
			Integer sourceNodeId = mapping.getSourceNodeIdMappedTo(targetNode.getId());
			int newId;
			if (sourceNodeId != null) {
				newId = mapOrg2NewId.get(sourceNodeId);
			} else {
				ENode newNode = factory.createNode(newGraph, targetNode);
				newId = newNode.getId();
			}
			mapOrgTarget2NewId.put(targetNode.getId(), newId);
			newGraph.updateNodeRepresentedNode(newId, targetNode.getGraphId(), targetNode.getId());
		}

		for (EEdge e : source.getEdges()) {
			int newSourceId = mapOrg2NewId.get(e.getSourceId());
			int newTargetId = mapOrg2NewId.get(e.getTargetId());
			EEdge newEdge = factory.createEdge(newGraph, newSourceId, newTargetId);
			newGraph.addEdge(newEdge);
			newGraph.updateEdgeRepresentedEdge(newEdge.getId(), e.getGraphId(), e.getId());
		}

		for (EEdge e : target.getEdges()) {
			int newSourceId = mapOrgTarget2NewId.get(e.getSourceId());
			int newTargetId = mapOrgTarget2NewId.get(e.getTargetId());
			EEdge newEdge = newGraph.getEdge(newSourceId, newTargetId);
			if (newEdge == null) {
				newEdge = factory.createEdge(newGraph, newSourceId, newTargetId);
				newGraph.addEdge(newEdge);
			}
			newGraph.updateEdgeRepresentedEdge(newEdge.getId(), e.getGraphId(), e.getId());
		}

		newGraph.setMapping(mapping);
		
		return newGraph;
	}


	

	

}
