package org.processmining.xlu.matching.merge.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.xlu.matching.merge.FusionAlg;
import org.processmining.xlu.matching.merge.FusionStrategy;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.collection.EGraphs;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyWeightsImpl;
import org.processmining.xlu.matching.zexperiment.model.FusionBinaryResult;
public class FusionStrategyBinaryImpl implements FusionStrategy {
	
	private PLog plog;
	private EGMatchConfiguration conf;
//	private int maxCost;
	
	private List<CostStrategyConfiguration> configs = new ArrayList<>();
	
	FusionBinaryResult data = new FusionBinaryResult();

	/**
	 * 
	 * @param plog
	 * @param conf
	 * @param maxCost: 	if -2 then all mappings, 
	 * 					if -1 then all optimal mappings,
	 * 					if 0 then only an optimal mapping, 
	 * 					if > 0 then only mappings with cost less than.  
	 */
	public FusionStrategyBinaryImpl(PLog plog, EGMatchConfiguration conf, int maxCost){
		this.plog = plog;
		this.conf = conf;
//		this.maxCost = maxCost;
		
		CostStrategyConfiguration config1 = new CostStrategyWeightsImpl();
		config1.setName("Config 1");
//		config1.setWeightMatched(1);
//		config1.setWeightNoMatch(0);
//		config1.setWeightGlobalStructure(0);
		configs.add(config1);
		
		CostStrategyConfiguration config2 = new CostStrategyWeightsImpl();
		config2.setName("Config 2");
		config2.setWeightMatched(1);
		config2.setWeightNoMatch(1);
		config2.setWeightGlobalStructure(0);
		configs.add(config2);
		
		CostStrategyConfiguration config3 = new CostStrategyWeightsImpl();
		config3.setName("Config 3");
		config3.setWeightMatched(1);
		config3.setWeightNoMatch(1);
		config3.setWeightGlobalStructure(1);
		configs.add(config3);
	}

	
	public RepEGraphs fuseEGraphs(EGraphs graphSet, MatchMappings mappings) {
//		List<String> renamedLabels = Arrays.asList(NewEGMatchTestPluginMUDataSetImpl.MU_RENAME_TEST);// TODO remove & only for test
		
		List<String> renamedLabels = new ArrayList<>();
		EGraphFactory factory = EGraphFactory.getInstance();
		XLog log = plog.getXLog();
		CostStrategy costfunction = conf.getCostStrategy();
		List<EGraph> graphs = graphSet.getGraphs();
		RepEGraphs mergedGraphs = new RepEGraphs();
		

		// 
		FusionAlg merge = new FusionAlgBinaryImpl();
		

		long s = System.currentTimeMillis();
		
		for (int i = 0; i < graphs.size(); i++) {
			EGraph g1 = graphs.get(i);
			for (int j = i + 1; j < graphs.size(); j++) {
				EGraph g2 = graphs.get(j);
//				MatchMapping mapping = mapGraphsToMapping.getMapping(i, j);
				//TODO: maxCost
				
				List<MatchMapping> mappingList = mappings.getAllMappings(i, j);
				// merge graphs
				for (MatchMapping mapping : mappingList) {
					RepEGraph representiveGraph = merge.mergeEGraph(g1, g2, mapping);
					mergedGraphs.add(representiveGraph);
					
					for(CostStrategyConfiguration config : this.configs){
						costfunction.setCostConfiguration(config);
						int cost = costfunction.getTotalCost(g1, g2, mapping);
						representiveGraph.setCost(config.getName(), cost);
					}

					// Cost properties.
					int sumCost = 0;
					Map<Integer, Integer> mapNode2Cost = new HashMap<Integer, Integer>();
					Map<Integer, Boolean> mapNode2Corrent = new HashMap<Integer, Boolean>();
					for (Entry<Integer, Map<Integer, Integer>> entry : representiveGraph.getNodeRepresentedNodes()
							.entrySet()) {
						int newNodeId = entry.getKey();
						Map<Integer, Integer> mappedOrgGraphToNode = entry.getValue();
						Integer sourceNodeId = mappedOrgGraphToNode.get(g1.getId());
						Integer targetNodeId = mappedOrgGraphToNode.get(g2.getId());

//						int cost = 0;

						if (sourceNodeId != null && targetNodeId != null) {

							// START adhoc Quality measurements
							if (!factory.isArtifical(g1.getNodeById(sourceNodeId))
									&& !factory.isArtifical(g2.getNodeById(targetNodeId))) {
//								cost = costfunction
//										.getCost(g1.getNode(sourceNodeId), g2.getNode(targetNodeId), mapping);

								XEvent source = log.get(i).get(sourceNodeId);
								XEvent target = log.get(j).get(targetNodeId);

								String sourceLabel = XConceptExtension.instance().extractName(source);
								String targetLabel = XConceptExtension.instance().extractName(target);
								boolean vclabel = renamedLabels.contains(sourceLabel)
										|| renamedLabels.contains(targetLabel);
								boolean equal = sourceLabel.equals(targetLabel);
								if (vclabel) {
									data.countTotRenamedVC++;
									if (equal) {
										data.countTotRenamedVCpos++;
										data.truePositive++;
									}
								} else if (equal) {
									data.truePositive++;
								}
								data.countTot++;
								mapNode2Corrent.put(newNodeId, equal);
								// END adhoc Quality measurements 
							}
						} else if (targetNodeId == null) { // sourceNodeId != null
//							cost = costfunction.getCostNoMapping(g1.getNode(sourceNodeId), mapping);
						} else if (sourceNodeId == null) {
//							cost = costfunction.getCostNoMapping(g2.getNode(targetNodeId), mapping);
						}
//						mapNode2Cost.put(newNodeId, cost);
//						sumCost += cost;

					}
					representiveGraph.setNodeToCost(mapNode2Cost);
					representiveGraph.setCost(sumCost);
					representiveGraph.setNodeToCorrect(mapNode2Corrent);

					// fill in distance matrix
					if (sumCost != mapping.getCost()) {
						System.err.println("Cost of mapping not equal to cost of merge! " + mapping.getCost() + " / "
								+ sumCost);
					}

				} //End for mapping
			}//End for j
		}// End for i
		System.out.format("%d / %d renamed mapping positive. In total: %d / %d positive %n", data.countTotRenamedVCpos,
				data.countTotRenamedVC, data.truePositive, data.countTot);
		long e = System.currentTimeMillis();
		double estimatedTime = (e - s) / 1000.0 / 60.0;
		System.out.format("Merging graphs took %f min. %n", estimatedTime);

		//		} // for k

		/* test for hierarchical-clustering-1.0.3.jar */
		//		String[] names = new String[log.size()];
		//		for (int i = 0; i < log.size(); i++) {
		//			names[i] = XConceptExtension.instance().extractName(log.get(i));
		//		}
		//		ClusteringAlgorithm alg = new DefaultClusteringAlgorithm();
		//		double[][] matrix = new double[log.size()][log.size()];
		//		for( RepEGraph g : mergedGraphs){
		//			matrix[g.getRepresentedGraphIds().get(0)][g.getRepresentedGraphIds().get(1)] = g.getCost();
		//			matrix[g.getRepresentedGraphIds().get(1)][g.getRepresentedGraphIds().get(0)] = g.getCost();
		//		}
		//		Cluster cluster = alg.performClustering(matrix, names, new AverageLinkageStrategy());
		//		DendrogramPanel dp = new DendrogramPanel();
		//		dp.setModel(cluster);

		//		BufferedWriter out = null;
		//		try {
		//			FileWriter fstream = new FileWriter("res/out.txt", true); //true tells to append data.
		//			out = new BufferedWriter(fstream);
		//			for (int i = 0; i < matrix.length; i++) {
		//				//		    	out.write(i + ";");
		//				for (int j = 0; j < matrix.length; j++) {
		//					out.write(matrix[i][j] + ";");
		//				}
		//				out.newLine();
		//			}
		//
		//		} catch (IOException e1) {
		//			System.err.println("Error: " + e1.getMessage());
		//		} finally {
		//			if (out != null) {
		//				try {
		//					out.close();
		//				} catch (IOException e1) {
		//					// TODO Auto-generated catch block
		//					e1.printStackTrace();
		//				}
		//			}
		//		}
		return mergedGraphs;
	}

}
