package org.processmining.xlu.matching.merge.impl;

import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.model.XLog;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.xlu.matching.merge.FusionAlg;
import org.processmining.xlu.matching.merge.FusionStrategy;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.collection.EGraphs;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.models.impl.MatchMappingsImpl;
import org.processmining.xlu.matching.zexperiment.model.FusionBinaryResult;
import org.processmining.xlu.matching.zexperiment.verification.EGMatchTestVerificationMethod;

public class FusionStrategyBinaryVerificationImpl implements FusionStrategy {

	private PLog plog;
	//	private EGMatchConfiguration conf;

	FusionBinaryResult data;
	EGMatchTestVerificationMethod verify;

	private boolean doRandomMapping = true;
	
	//Feb22 adhoc test 
	MatchMappings qualityOpts;

	public double estimatedTime;

	public MatchMappings getQualityOpts() {
		return qualityOpts;
	}



	public FusionBinaryResult getData() {
		return data;
	}

	public FusionStrategyBinaryVerificationImpl(PLog plog, boolean doRandomOpt, EGMatchTestVerificationMethod method) {
		this.plog = plog;
		data = new FusionBinaryResult();
		this.doRandomMapping = doRandomOpt;
		this.verify  = method;
	}

	public RepEGraphs fuseEGraphs(EGraphs graphSet, MatchMappings mappings) {
		XLog log = plog.getXLog();
		List<EGraph> graphs = graphSet.getGraphs();
		RepEGraphs mergedGraphs = new RepEGraphs();
		FusionAlg merge = new FusionAlgBinaryImpl();
		
		List<MatchMapping> qualityOptMappings = new ArrayList<MatchMapping>(); //TODO REMOVE
		
		long s = System.currentTimeMillis();

		for (int i = 0; i < graphs.size(); i++) {
			EGraph g1 = graphs.get(i);
			//			System.out.println("Started merge " + i + " / " + graphs.size());
			for (int j = i + 1; j < graphs.size(); j++) {
				EGraph g2 = graphs.get(j);

				List<MatchMapping> mappingList = mappings.getAllMappings(i, j);

				FusionBinaryResult optMappingQuality = new FusionBinaryResult(0, Integer.MAX_VALUE);
				int size = mappingList.size();
				if (selectRandomMapping()) {
					size = 1;
				}
				MatchMapping qualityOpt = null;
				for (int l = 0; l < size; l++) {

					MatchMapping mapping = mappingList.get(l);
					// merge graphs	
					RepEGraph representiveGraph = merge.mergeEGraph(g1, g2, mapping);

					FusionBinaryResult mappingQuality = new FusionBinaryResult();
					verify.computeCorrectness(representiveGraph, g1, g2, mapping, log, mappingQuality);

//					if (mappingQuality.falsePositive > 0) {
						mergedGraphs.add(representiveGraph);
//					}

					if (mappingQuality.falsePositive < optMappingQuality.falsePositive) {
						optMappingQuality.falsePositive = mappingQuality.falsePositive;
						optMappingQuality.truePositive = mappingQuality.truePositive;
						qualityOpt = mapping;
					}

				} //End for mapping

				this.data.falsePositive += optMappingQuality.falsePositive;
				this.data.truePositive += optMappingQuality.truePositive;
				//TODO!!!!!!!!!!!!!!!!!!!!!REMOVE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				qualityOptMappings.add(qualityOpt);

			}//End for j
		}// End for i
		System.out.format("In total: %d / %d positive %n", data.truePositive, data.falsePositive);
		long e = System.currentTimeMillis();
		estimatedTime = (e - s) / 1000.0 / 60.0;
		System.out.format("Merging graphs took %f min. %n", estimatedTime);
		
		this.qualityOpts = new MatchMappingsImpl(qualityOptMappings);

		return mergedGraphs;
		//		return null;
	}

	private boolean selectRandomMapping() {
		return doRandomMapping;
	}





}
