package org.processmining.xlu.matching.merge.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.collections15.map.HashedMap;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.xlu.matching.merge.FusionStrategy;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.matching.models.collection.EGraphs;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.models.impl.RepEGraphImpl;

public class FusionStrategyIncrBinaryImpl implements FusionStrategy {
	XLog log;
	private PluginContext context;
	public double estimatedTime;

	public FusionStrategyIncrBinaryImpl(PluginContext context, XLog log) {
		this.log = log;
		this.context = context;
	}

	public RepEGraphs fuseEGraphs(EGraphs graphs, MatchMappings mappings) {
		//		FusionAlg alg = new FusionAlgBinaryImpl();
		long s = System.currentTimeMillis();

		SortedSet<MatchMapping> mappingList = new TreeSet<MatchMapping>(mappings.getMappingComparator());
		mappingList.addAll(mappings.getOptimalMappings());
		//		Collections.sort(mappingList, mappings.getMappingComparator());

		//		List<RepEGraph> mergedGraphs = new ArrayList<RepEGraph>();
		RepEGraphs mergedGraphs = new RepEGraphs();
		Map<Integer, Tree<RepEGraph>> mapEG2REG = new HashedMap<Integer, Tree<RepEGraph>>();
		for (EGraph eg : graphs.getGraphs()) {
			RepEGraph reg = createRepEGraph(eg, new HashedMap<Integer, Integer>());
			mapEG2REG.put(eg.getId(), new Tree<RepEGraph>(reg));
			mergedGraphs.add(reg);
		}

		//		Map<Integer, > hierarchy = new HashMap<Integer, Tree<RepEGraph>>();

		while (mappingList.size() > 0) {
			
			//			System.out.println(mappingList.size() + " remaining mappings %n" );
			MatchMapping mapping = mappingList.first();
			Tree<RepEGraph> sourceTree = mapEG2REG.get(mapping.getSourceGraph().getId());
			RepEGraph sourceREG = sourceTree.getObject();
			Tree<RepEGraph> targetTree = mapEG2REG.get(mapping.getTargetGraph().getId());
			RepEGraph targetREG = targetTree.getObject();
			List<MatchMapping> subMappings = getAndRemoveMappingsBetween(sourceREG.getRepresentedGraphIds(),
					targetREG.getRepresentedGraphIds(), mappingList, mappings);

			
			if (subMappings.size() > 0) {
				RepEGraph fusedREG = fuseRepEGraph(sourceREG, targetREG, subMappings);
				if (fusedREG != null) {

					int avgCost = getAverageCost(subMappings);
					fusedREG.setCost(avgCost);
					
					// update MapEG2REG
					Tree<RepEGraph> parentTree = new Tree<RepEGraph>(fusedREG);
					
					// sourceTree.setParent(parentTree);
					// targetTree.setParent(parentTree);
					parentTree.add(sourceTree);
					parentTree.add(targetTree);
					for (Integer i : fusedREG.getRepresentedGraphIds()) {
						mapEG2REG.put(i, parentTree);
					}
					mergedGraphs.add(fusedREG);
				}
			}
		}
		Set<Tree<RepEGraph>> children = new HashSet<>(mapEG2REG.values());
		Tree<RepEGraph> root = new Tree<RepEGraph>(null);
		for (Tree<RepEGraph> c : children) {
			root.add(c);
		}
		mergedGraphs.setHierarchy(root);

		//		RepEGraphs res = new RepEGraphs(mergedGraphs);

		//		// I. Fuse graphs with mapping cost 0
		//		List<EGraph> graphList = new ArrayList<EGraph>(graphs.getGraphs());
		//		List<EGraph> toFuseGraphs = new ArrayList<EGraph>();
		//		while (graphList.size() > 0) {
		//			EGraph source = graphList.remove(0);
		//			Map<Integer, Integer> mapOrg2NewId = new HashMap<Integer, Integer>();
		//			RepEGraph newGraph = createRepEGraph(source, mapOrg2NewId);
		//
		//			List<EGraph> matchedSet = new ArrayList<EGraph>();
		//			for (EGraph toMatchGraph : graphList) {
		//				MatchMapping mapping = mappings.getOptMapping(source.getId(), toMatchGraph.getId());
		//				if (mapping.getCost() == 0) {
		//					matchedSet.add(toMatchGraph);
		//					updateRepEGraph(newGraph, toMatchGraph, mapping, mapOrg2NewId);
		//				}
		//			}
		//			if (matchedSet.size() > 0) {
		//				graphList.removeAll(matchedSet);
		//			} else {
		//				toFuseGraphs.add(source);
		//			}
		//		}
		//
		//		// Fuse graphs
		//		for(toFuseGraphs)
		long e = System.currentTimeMillis();
		estimatedTime = (e - s) / 1000.0 / 60.0;
		System.out.format("Merging graphs took %f min. %n", estimatedTime);
		return mergedGraphs;
	}

	private int getAverageCost(List<MatchMapping> subMappings) {
		int sum = 0;
		for(MatchMapping m : subMappings){
			sum += m.getCost();
		}
		sum = sum / subMappings.size();
		return sum;
	}

	private List<MatchMapping> getAndRemoveMappingsBetween(List<Integer> sIds, List<Integer> tIds,
			SortedSet<MatchMapping> mappingList, MatchMappings mappings) {
		List<MatchMapping> res = new ArrayList<MatchMapping>();
		//		for (MatchMapping m : mappingList) {
		//			int sourceId = m.getSourceGraph().getId();
		//			int targetId = m.getTargetGraph().getId();
		//			if (sIds.contains(sourceId) && tIds.contains(targetId)) {
		//				res.add(m);
		//			} else if (sIds.contains(targetId) && tIds.contains(sourceId)) {
		//				res.add(m);
		//			}
		//
		//		}
		for (Integer s : sIds) {
			for (Integer t : tIds) {
				MatchMapping m = s < t ? mappings.getAnOptimalMapping(s, t) : mappings.getAnOptimalMapping(t, s);
				res.add(m);
				boolean r = mappingList.remove(m);
				if (!r) {
					//					System.out.println("Not removed");
				}
			}
		}
		//		mappingList.removeAll(res);
		return res;
	}

	protected RepEGraph fuseRepEGraph(RepEGraph sourceREG, RepEGraph targetREG, List<MatchMapping> mappings) {
		Map<Integer, Integer> mapSourceOrg2NewId = new HashedMap<Integer, Integer>();
		Map<Integer, Integer> mapTargetOrg2NewId = new HashedMap<Integer, Integer>();
		//		List<ENode> targetRENodes = new ArrayList<ENode>(targetREG.getNodes());

		RepEGraph fusedGraph = createRepEGraphAndAddSourceNodesEdges(sourceREG, mapSourceOrg2NewId);

		boolean valid = true;

		MatchMapping mappingREG = new MatchMappingImpl(sourceREG, targetREG);
		for (MatchMapping mapping : mappings) {
			if(context != null){
				context.getProgress().inc();
			}

			boolean succeed = updateHigherLevelMapping(mappingREG, sourceREG, targetREG, mapping);
			if (!succeed) { // there is conflic
				return null;
			}

		}

		for (Integer i : targetREG.getRepresentedGraphIds()) {
			fusedGraph.addRepresentedGraph(i);
		}
		for (ENode targetRENode : targetREG.getNodes()) {
			int targetRENodeId = targetRENode.getId();
			Integer sourceRENodeId = mappingREG.getSourceNodeIdMappedTo(targetRENodeId);
			int newNodeId;
			if (sourceRENodeId != null) {
				newNodeId = mapSourceOrg2NewId.get(sourceRENodeId);
			} else {
				ENode newNode = EGraphFactory.getInstance().createNode(fusedGraph, targetRENode);
				newNodeId = newNode.getId();
			}
			mapTargetOrg2NewId.put(targetRENodeId, newNodeId);
			for (Entry<Integer, Integer> entry : targetREG.getNodeRepresentedNodes().get(targetRENodeId).entrySet()) {
				fusedGraph.updateNodeRepresentedNode(newNodeId, entry.getKey(), entry.getValue());
			}
		}

		for (EEdge e : targetREG.getEdges()) {
			int newSourceId = mapTargetOrg2NewId.get(e.getSourceId());
			int newTargetId = mapTargetOrg2NewId.get(e.getTargetId());
			EEdge newEdge = EGraphFactory.getInstance().createEdge(fusedGraph, newSourceId, newTargetId);
			for (Entry<Integer, Integer> entry : targetREG.getEdgeRepresentedEdges().get(e.getId()).entrySet()) {
				fusedGraph.updateEdgeRepresentedEdge(newEdge.getId(), entry.getKey(), entry.getValue());
			}
		}

		// Implementation Testing
		//		debugTEST(sourceREG, targetREG, fusedGraph);
		//End implementation test
		//		System.out.println("Valid merge");
		return fusedGraph;
	}

	protected boolean updateHigherLevelMapping(MatchMapping mappingREG, RepEGraph sourceREG, RepEGraph targetREG,
			MatchMapping mapping) {
		int sGraphId = mapping.getSourceGraph().getId();
		int tGraphId = mapping.getTargetGraph().getId();
		if (sourceREG.getRepresentedGraphIds().contains(sGraphId)) {
			// skip
		} else if (targetREG.getRepresentedGraphIds().contains(sGraphId)) {
			// Mapping source = targetREG and target = SourceREG
			// Thus, Inverse the mapping
			MatchMapping inversed = new MatchMappingImpl(mapping.getSourceGraph(), mapping.getTargetGraph());
			inversed.clone(mapping);
			inversed.inverse();
			mapping = inversed;
			sGraphId = mapping.getSourceGraph().getId();
			tGraphId = mapping.getTargetGraph().getId();
		} else {
			System.err.println("Mapping invalid");
		}

		// Case 1
		for (Entry<Integer, Integer> match : mapping.getMatchs()) {
			ENode sourceInSourceREG = sourceREG.getNodeRepresenting(sGraphId, match.getKey());
			ENode targetInTargetREG = targetREG.getNodeRepresenting(tGraphId, match.getValue());

			if (!areValidMatch(sourceInSourceREG, sourceREG, targetInTargetREG, targetREG)) {
				// allow inheritance classes to implement differently...
//				System.out.println("Skipping some nodes in a mapping");
				continue;
			}

			Integer mappedTargetId = mappingREG.getTargetNodeIdMappedTo(sourceInSourceREG.getId());
			Integer mappedSourceId = mappingREG.getSourceNodeIdMappedTo(targetInTargetREG.getId());
			if (mappedTargetId != null && mappedSourceId != null) {
				if (mappedTargetId == targetInTargetREG.getId() && mappedSourceId == sourceInSourceREG.getId()) {

					// Good
				} else {
					// Bad
					return false;
				}
			} else if (mappedSourceId != null || mappedTargetId != null) {
				// one of the paired is matched to something else...
				return false;
			} else {
				mappingREG.addMatch(sourceInSourceREG, targetInTargetREG);
				//					mappingREG.removeMatch(sourceInSourceREG, null);
				//					mappingREG.removeMatch(null, targetInTargetREG);
			}
		}

		// Case 2 && 3 result in nodes not mapped...
		return true;
	}

	protected boolean areValidMatch(ENode sourceInSourceREG, RepEGraph sourceREG, ENode targetInTargetREG, RepEGraph targetREG) {
		return true;
	}

	//	private void debugTEST(RepEGraph sourceREG, RepEGraph targetREG, RepEGraph fusedGraph) {
	//		int countEvent = 0;
	//		int countNoise = 0;
	//		int countFusedEvent = 0;
	//		int countFusedNoise = 0;
	//		Map<Integer, Integer> org = new HashedMap<Integer, Integer>();
	//		//		Map<Integer, Integer> fused = new HashedMap<Integer, Integer>();
	//		for (Entry<Integer, Map<Integer, Integer>> map : sourceREG.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//				System.out.println("SourceREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID "
	//						+ entry.getValue());
	//				//				org.put(entry.getKey(), entry.getValue());
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countNoise++;
	//				} else {
	//					countEvent++;
	//				}
	//			}
	//		}
	//		for (Entry<Integer, Map<Integer, Integer>> map : targetREG.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//
	//				System.out.println("TargetREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID "
	//						+ entry.getValue());
	//				//				org.put(entry.getKey(), entry.getValue());
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countNoise++;
	//				} else {
	//					countEvent++;
	//				}
	//			}
	//		}
	//		for (Entry<Integer, Map<Integer, Integer>> map : fusedGraph.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//
	//				System.out.println("FusedREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID "
	//						+ entry.getValue());
	//				//				if(org.get(entry.getKey())!= entry.getValue()){
	//				//					System.out.println("NNNNNNNN");
	//				//				}
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countFusedNoise++;
	//				} else {
	//					countFusedEvent++;
	//				}
	//			}
	//		}
	//		if (countFusedEvent != countEvent || countFusedNoise != countNoise) {
	//			System.out.println("AAAAAAAAAAAAAAA");
	//		}
	//		
	//	}

	//	private RepEGraph fuseRepEGraph(RepEGraph sourceREG, RepEGraph targetREG, List<MatchMapping> mappings) {
	//		Map<Integer, Integer> mapSourceOrg2NewId = new HashedMap<Integer, Integer>();
	//		Map<Integer, Integer> mapTargetOrg2NewId = new HashedMap<Integer, Integer>();
	//		List<ENode> targetRENodes = new ArrayList<ENode>(targetREG.getNodes());
	//		
	//		
	//		
	//		RepEGraph fusedGraph = createRepEGraphAndAddSourceNodesEdges(sourceREG, mapSourceOrg2NewId);
	//		
	//
	//		boolean valid = true;
	//		for (ENode sourceRENode : sourceREG.getNodes()) {
	//			// Get the original EGraphs and ENodes which the sourceRENode is representing
	//			Map<Integer, Integer> sourceRENode_MapGraph2ENode = sourceREG.getNodeRepresentedNodes().get(
	//					sourceRENode.getId());
	//
	//			// There is a targetRENode mapped to sourceRENode, 
	//			// Get the original EGraphs and ENodes which the targetRENode is representing
	//			ENode targetRENode = null;
	//			Map<Integer, Integer> mappedTargetRENode_MapGraph2ENode = new HashMap<Integer, Integer>();
	//			Set<Integer> mappedTargedIds = new HashSet<Integer>();
	//			boolean mappedToNothing = false;
	//			for (MatchMapping mapping : mappings) {
	//				// Mapping between two EGraphs represented by sourceREG and targetREG
	//				int sGraphId = mapping.getSourceGraph().getId();
	//				int tGraphId = mapping.getTargetGraph().getId();
	//				if (sourceREG.getRepresentedGraphIds().contains(sGraphId)) {
	//					// skip
	//				} else if (targetREG.getRepresentedGraphIds().contains(sGraphId)) {
	//					// Mapping source = targetREG and target = SourceREG
	//					// Thus, Inverse the mapping
	//					MatchMapping inversed = new MatchMappingImpl(mapping.getTargetGraph(), mapping.getSourceGraph());
	//					inversed.clone(mapping);
	//					inversed.inverse();
	//					mapping = inversed;
	//					sGraphId = mapping.getSourceGraph().getId();
	//					tGraphId = mapping.getTargetGraph().getId();
	//				} else {
	//					System.err.println("Mapping invalid");
	//				}
	//
	//				// sourceRENode is representing the node sNodeId in the graph sGraphId
	//				Integer sNodeId = sourceRENode_MapGraph2ENode.get(sGraphId);
	//
	//				if (sNodeId == null) {
	//					continue;
	//				}
	//
	//				// and to get the ENode to which sNodeId is mapped
	//				Integer tNodeId = mapping.getTargetNodeIdMappedTo(sNodeId);
	//
	//				if (tNodeId != null) {
	//					if (targetRENode == null) { // initialize targetRENode
	//						// Get the targetRENode that is representing tNodeId in targetREG
	//						targetRENode = targetREG.getNodeRepresenting(tGraphId, tNodeId);
	//						// Obtain the ENodes and EGraphs this targetRENode is representing
	//						mappedTargetRENode_MapGraph2ENode = targetREG.getNodeRepresentedNodes().get(
	//								targetRENode.getId());
	//						mappedTargedIds = new HashSet<Integer>(mappedTargetRENode_MapGraph2ENode.keySet());
	//					}
	//				} else {
	//					mappedToNothing = true;
	//				}
	//
	//				// for each EGraph and ENode represented by sourceRENode, there is a mapping that mappings it to a EGraph and ENode that is represented by targetRENode.
	//				valid &= mappedTargetRENode_MapGraph2ENode.get(tGraphId) == tNodeId;
	//				mappedTargedIds.remove(tGraphId);
	//				if (!valid || (targetRENode!= null && mappedToNothing)) {
	//					//					System.out.println("Invalid node");
	//					return null;
	//				}
	//			}// End mapping. 
	//			
	//			if(mappedTargedIds.size() != 0){
	//				System.out.println("After all mapping, target id left?");
	//				return null;
	//			} 
	//			if (targetRENode != null) {
	//				
	//				targetRENodes.remove(targetRENode);
	//				int newNodeId = mapSourceOrg2NewId.get(sourceRENode.getId());
	//				mapTargetOrg2NewId.put(targetRENode.getId(), newNodeId);
	//				for (Entry<Integer, Integer> entry : mappedTargetRENode_MapGraph2ENode.entrySet()) {
	//					fusedGraph.updateNodeRepresentedNode(newNodeId, entry.getKey(), entry.getValue());
	//				}
	//			}
	//
	//			// valid sourceRENode => Fuse the two nodes update FusedGraph.
	//			//			System.out.println("Valid node");
	//
	//		}
	//
	//		
	//		for (Integer i : targetREG.getRepresentedGraphIds()) {
	//			fusedGraph.addRepresentedGraph(i);
	//		}
	//		for (ENode targetNode : targetRENodes) {
	//			ENode newNode = EGraphFactory.getInstance().createNode(fusedGraph, targetNode);
	//			for (Entry<Integer, Integer> entry : targetREG.getNodeRepresentedNodes().get(targetNode.getId()).entrySet()) {
	//				fusedGraph.updateNodeRepresentedNode(newNode.getId(), entry.getKey(), entry.getValue());
	//			}
	//			mapTargetOrg2NewId.put(targetNode.getId(), newNode.getId());
	//		}
	//
	//		for (EEdge e : targetREG.getEdges()) {
	//			int newSourceId = mapTargetOrg2NewId.get(e.getSourceId());
	//			int newTargetId = mapTargetOrg2NewId.get(e.getTargetId());
	//			EEdge newEdge = EGraphFactory.getInstance().createEdge(fusedGraph, newSourceId, newTargetId);
	//			for (Entry<Integer, Integer> entry : targetREG.getEdgeRepresentedEdges().get(e.getId()).entrySet()) {
	//				fusedGraph.updateEdgeRepresentedEdge(newEdge.getId(), entry.getKey(), entry.getValue());
	//			}
	//		}
	//
	//		// Implementation Testing
	//		int countEvent = 0;
	//		int countNoise = 0;
	//		int countFusedEvent = 0;
	//		int countFusedNoise = 0;
	//		Map<Integer, Integer> org = new HashedMap<Integer, Integer>();
	//		//		Map<Integer, Integer> fused = new HashedMap<Integer, Integer>();
	//		for (Entry<Integer, Map<Integer, Integer>> map : sourceREG.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//				System.out.println("SourceREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID " + entry.getValue());
	//				//				org.put(entry.getKey(), entry.getValue());
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countNoise++;
	//				} else {
	//					countEvent++;
	//				}
	//			}
	//		}
	//		for (Entry<Integer, Map<Integer, Integer>> map : targetREG.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//
	//				System.out.println("TargetREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID " + entry.getValue());
	//				//				org.put(entry.getKey(), entry.getValue());
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countNoise++;
	//				} else {
	//					countEvent++;
	//				}
	//			}
	//		}
	//		for (Entry<Integer, Map<Integer, Integer>> map : fusedGraph.getNodeRepresentedNodes().entrySet()) {
	//			for (Entry<Integer, Integer> entry : map.getValue().entrySet()) {
	//
	//				System.out.println("FusedREG: Node " + map.getKey() + ", GID " + entry.getKey() + ", EID " + entry.getValue());
	//				//				if(org.get(entry.getKey())!= entry.getValue()){
	//				//					System.out.println("NNNNNNNN");
	//				//				}
	//				XTrace xTrace = log.get(entry.getKey());
	//				if (entry.getValue() >= xTrace.size()) {
	//					continue;
	//				}
	//				XEvent e = xTrace.get(entry.getValue());
	//				if (XExperimentNoiseExtension.isNoise(e)) {
	//					countFusedNoise++;
	//				} else {
	//					countFusedEvent++;
	//				}
	//			}
	//		}
	//		if (countFusedEvent != countEvent || countFusedNoise != countNoise) {
	//			System.out.println("AAAAAAAAAAAAAAA");
	//		}
	//		//End implementation test
	//		//		System.out.println("Valid merge");
	//		return fusedGraph;
	//	}

	//	private void updateRepEGraph(RepEGraph newGraph, EGraph target, MatchMapping mapping,
	//			Map<Integer, Integer> mapOrg2NewId) {
	//
	//		EGraphFactory factory = EGraphFactory.getInstance();
	//		newGraph.addRepresentedGraph(target.getId());
	//		Map<Integer, Integer> mapOrgTarget2NewId = new HashMap<Integer, Integer>();
	//		for (ENode targetNode : target.getNodes()) {
	//			Integer sourceNodeId = mapping.getSourceNodeIdMappedTo(targetNode.getId());
	//			int newId;
	//			if (sourceNodeId != null) {
	//				newId = mapOrg2NewId.get(sourceNodeId);
	//			} else {
	//				ENode newNode = factory.createNode(newGraph, targetNode);
	//				newId = newNode.getId();
	//			}
	//			mapOrgTarget2NewId.put(targetNode.getId(), newId);
	//			newGraph.updateNodeRepresentedNode(newId, targetNode.getGraphId(), targetNode.getId());
	//		}
	//		for (EEdge e : target.getEdges()) {
	//			int newSourceId = mapOrgTarget2NewId.get(e.getSourceId());
	//			int newTargetId = mapOrgTarget2NewId.get(e.getTargetId());
	//			EEdge newEdge = newGraph.getEdge(newSourceId, newTargetId);
	//			if (newEdge == null) {
	//				newEdge = factory.createEdge(newGraph, newSourceId, newTargetId);
	//				newGraph.addEdge(newEdge);
	//			}
	//			newGraph.updateEdgeRepresentedEdge(newEdge.getId(), e.getGraphId(), e.getId());
	//		}
	//
	//	}

	private RepEGraph createRepEGraph(EGraph source, Map<Integer, Integer> mapOrg2NewId) {
		RepEGraph newGraph = new RepEGraphImpl();
		EGraphFactory factory = EGraphFactory.getInstance();

		newGraph.addRepresentedGraph(source.getId());

		for (ENode sourceNode : source.getNodes()) {
			ENode newNode = factory.createNode(newGraph, sourceNode);
			newGraph.updateNodeRepresentedNode(newNode.getId(), sourceNode.getGraphId(), sourceNode.getId());
			mapOrg2NewId.put(sourceNode.getId(), newNode.getId());
		}
		for (EEdge e : source.getEdges()) {
			int newSourceId = mapOrg2NewId.get(e.getSourceId());
			int newTargetId = mapOrg2NewId.get(e.getTargetId());
			EEdge newEdge = factory.createEdge(newGraph, newSourceId, newTargetId);
			newGraph.addEdge(newEdge);
			newGraph.updateEdgeRepresentedEdge(newEdge.getId(), e.getGraphId(), e.getId());
		}
		return newGraph;
	}

	protected RepEGraph createRepEGraphAndAddSourceNodesEdges(RepEGraph source, Map<Integer, Integer> mapOrg2NewId) {
		RepEGraph newGraph = new RepEGraphImpl();
		EGraphFactory factory = EGraphFactory.getInstance();

		// update represented graphs
		for (Integer graphId : source.getRepresentedGraphIds()) {
			newGraph.addRepresentedGraph(graphId);
		}

		// update represented nodes
		for (ENode sourceNode : source.getNodes()) {
			ENode newNode = factory.createNode(newGraph, sourceNode);
			for (Entry<Integer, Integer> entry : source.getNodeRepresentedNodes().get(sourceNode.getId()).entrySet()) {
				newGraph.updateNodeRepresentedNode(newNode.getId(), entry.getKey(), entry.getValue());
			}
			mapOrg2NewId.put(sourceNode.getId(), newNode.getId());
		}

		// update represented edges
		for (EEdge e : source.getEdges()) {
			int newSourceId = mapOrg2NewId.get(e.getSourceId());
			int newTargetId = mapOrg2NewId.get(e.getTargetId());
			EEdge newEdge = factory.createEdge(newGraph, newSourceId, newTargetId);
			for (Entry<Integer, Integer> entry : source.getEdgeRepresentedEdges().get(e.getId()).entrySet()) {
				newGraph.updateEdgeRepresentedEdge(newEdge.getId(), entry.getKey(), entry.getValue());
			}
		}
		return newGraph;

	}

}
