package org.processmining.xlu.matching.models;

import java.util.List;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.partialorder.ptrace.model.PTrace;
import org.processmining.xlu.matching.models.impl.EEdgeImpl;
import org.processmining.xlu.matching.models.impl.EGraphImp;
import org.processmining.xlu.matching.models.impl.ENodeAwareImpl;
import org.processmining.xlu.matching.models.impl.ENodeDummyImpl;

public class EGraphFactory {
	
	public static final String Dummy_End_Label = "dEd";
	public static final String Dummy_Start_Label = "dSd";
	private static EGraphFactory instance = null;
	    private EGraphFactory() { }
	 
	    public static synchronized EGraphFactory getInstance() {
	        if (instance == null) {
	            instance = new EGraphFactory();
	        }
	 
	        return instance;
	    }

	public EGraph createEmptyExecutionGraph(PTrace ptrace) {
		return new EGraphImp(ptrace.getTraceIndex());
	}

//	public ENode createNode(EGraph eg, XEventClass c, int[] preds, int[] succs) {
//		String label = c.toString();
//		ENode node = new ENodeImpl(eg, label, preds, succs);
//		eg.addNode(node);
//		return node;
//	}

	public ENode createNode(EGraph graph, ENode node) {
		ENode newNode = new ENodeAwareImpl(graph, node.getLabel());
		graph.addNode(newNode);
		return newNode;
	}

	public EEdge createEdge(EGraph graph, int newSourceId, int newTargetId) {
		EEdge e = new EEdgeImpl(newSourceId, newTargetId);
		graph.addEdge(e);
		return e;
	}

	public ENode createArtificalStart(EGraph eg, int[] succs) {
		ENode start = new ENodeDummyImpl(eg, Dummy_Start_Label);
		eg.addNode(start);
		return start;
	}

	public ENode createArtificalEnd(EGraph eg, int[] preds) {
		ENode end = new ENodeDummyImpl(eg, Dummy_End_Label);
		eg.addNode(end);
		return end;
	}
	
	public ENode createArtificalStart(EGraph eg) {
		ENode start = new ENodeDummyImpl(eg, Dummy_Start_Label);
		eg.addNode(start);
		return start;
	}

	public ENode createArtificalEnd(EGraph eg) {
		ENode end = new ENodeDummyImpl(eg, Dummy_End_Label);
		eg.addNode(end);
		return end;
	}
	
	public boolean isArtificalStart(ENode node){
		return node instanceof ENodeDummyImpl || node.getLabel().equals(Dummy_Start_Label);
	}
	
	public boolean isArtificalEnd(ENode node){
		return node instanceof ENodeDummyImpl || node.getLabel().equals(Dummy_End_Label);
	}

	public boolean isArtifical(ENode node) {
		return isArtificalStart(node) || isArtificalEnd(node);
	}

//	public ENode createNode(EGraph eg, XEventClass c, int[] preds, int[] succs, int[] concs) {
//		String label = c.toString();
//		ENode node = new ENodeImpl(eg, label, preds, succs, concs);
//		eg.addNode(node);
//		return node;
//	}

	public ENode createNode(EGraph eg, XEventClass c, List<Integer> predIds, List<Integer> succIds, List<Integer> concIds) {
		String label = c.toString();
		ENode node = new ENodeAwareImpl(eg, label, predIds, succIds, concIds);
		eg.addNode(node);
		return node;
	}

	//TODO: might be removed
	public void setGraphDist(EGraph eg, int[][] distMatrix) {
		if(eg instanceof EGraphImp){
			((EGraphImp) eg).setGraphDist(distMatrix);
		}
	}

	public ENode createNode(EGraph eg, XEventClass c, List<Integer> predIds, List<Integer> succIds, List<Integer> concIds,
			int[] preds, int[] succs, int[] concs) {
		String label = c.toString();
		ENode node = new ENodeAwareImpl(eg, label, predIds, succIds, concIds, preds, succs, concs);
		eg.addNode(node);
		return node;
	}

}
