package org.processmining.xlu.matching.models.collection;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;


/**
 * 
 * 
 * 
 * @author xlu
 */
public class RepEGraphs implements Iterable<RepEGraph> {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8662324394671114526L;
	private XLog log;
	private PLog pLog;
	private XEventClasses classes;
//	private PGraphConverter<RepEGraph> converter;
	private REGraphType type;

	private List<RepEGraph> graphs;
	private Tree<RepEGraph> tree;
	private MatchMappings mappings;


	public RepEGraphs(XLog log, PLog pLog, XEventClasses classes) {
		this();
		this.log = log;
		this.pLog = pLog;
		this.classes = classes;
	}

	public RepEGraphs() {
		graphs = new ArrayList<RepEGraph>();
	}

	public Set<Integer> getInstanceGraphIds() {
		Set<Integer> ids = new HashSet<Integer>();
		for (RepEGraph g : this) {
			ids.addAll(g.getRepresentedGraphIds());
		}
		return ids;
	}

	public XLog getLog() {
		return log;
	}

	public void setLog(XLog log) {
		this.log = log;
	}

	public PLog getpLog() {
		return pLog;
	}

	public void setpLog(PLog pLog) {
		this.pLog = pLog;
	}

	public XEventClasses getClasses() {
		return classes;
	}

	public void setClasses(XEventClasses classes) {
		this.classes = classes;
	}

//	public PGraphConverter<RepEGraph> getConverter() {
//		return converter;
//	}

//	public void setConverter(PGraphConverter<RepEGraph> converter) {
//		this.converter = converter;
//	}

	public Iterator<RepEGraph> iterator() {
		return graphs.iterator();
	}

	/**
	 * Get Direct Children of Root.
	 * 
	 * @return
	 */
	public List<RepEGraph> getGraphs() {
		if (tree != null && tree.getChildCount() > 0) {
			List<RepEGraph> res = new ArrayList<RepEGraph>();
			for (int i = 0; i < tree.getChildCount(); i++) {
				RepEGraph g = ((Tree<RepEGraph>) tree.getChildAt(i)).getObject();
				res.add(g);
			}
			return res;
		}
		return graphs;
	}

	/**
	 * Return all graphs.
	 * 
	 * @return
	 */
	public List<RepEGraph> getAllGraphs() {
		return graphs;
	}

	public void add(RepEGraph fusedREG) {
		int id = graphs.size();
		graphs.add(0, fusedREG);
		fusedREG.setId(id);
	}

	public void setHierarchy(Tree<RepEGraph> root) {
		this.tree = root;

	}

	public Tree<RepEGraph> getHierarchy() {
		return tree;
	}

	public RepEGraph getRepEGraph(int graphId) {
		for (RepEGraph g : this) {
			if (g.getId() == graphId) {
				return g;
			}
		}
		return null;
	}

	public void setMappings(MatchMappings mappings) {
		this.mappings = mappings;
	}

	public MatchMappings getMappings() {
		return mappings;
	}

	/**
	 * Get the type of graph which has influence on how the graphs are
	 * visualized
	 * 
	 * @return
	 */
	public REGraphType getType() {
		return type;
	}

	public void setType(REGraphType type) {
		this.type = type;
	}

}
