package org.processmining.xlu.matching.models.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections15.map.HashedMap;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;

import com.google.common.base.Joiner;

public class EGraphImp implements EGraph {

	int id;
	protected List<ENode> nodes;
	protected List<EEdge> edges;

	/* Indexing structure */
	private Map<String, List<Integer>> mapLabel2Node; // instead of complex index structure, we try the simple first
	private int[][] distMatrix; 
	//	protected NodeIndexing index; 

	/* properties */
	private int traceIndex;

	protected EGraphImp() {
		nodes = new ArrayList<>();
		edges = new ArrayList<>();
		this.mapLabel2Node = new HashedMap<String, List<Integer>>();
	}

	public EGraphImp(int id) {
		this();
		this.id = id;
	}
	
	public void removeNode(ENode node){
		
		List<EEdge> ins = new ArrayList<EEdge>();
		List<EEdge> outs = new ArrayList<EEdge>();
		for(EEdge e : edges){
			if(e.getSourceId() == node.getId()){
				outs.add(e);
			} else if(e.getTargetId() == node.getId()){
				ins.add(e);
			}
		}
		for(EEdge in : ins){
			ENode source = this.getNodeById(in.getSourceId());
			for(EEdge out : outs){
				ENode target = this.getNodeById(out.getTargetId());
				
				EEdge e = new EEdgeImpl(source.getId(), target.getId());
				this.addEdge(e);
			}
		}
		
		nodes.remove(node);
		edges.removeAll(ins);
		edges.removeAll(outs);
	}

	public int addNode(ENode node) {
		nodes.add(node); // add node to the node collection
		int id = nodes.size() - 1;
		node.setId(id); // TODO: might need fix
		node.setGraphId(this.getId());
		indexNode(id, node);
		return id;
	}

	private void indexNode(int id, ENode node) {
		List<Integer> indices = this.mapLabel2Node.get(node.getLabel()); // update label indexing.
		if (indices == null) {
			indices = new ArrayList<Integer>();
		}
		indices.add(id);
		this.mapLabel2Node.put(node.getLabel(), indices);
	}

	public List<ENode> getNodesWithLabel(ENode sourceNode) {
		List<Integer> indices = this.mapLabel2Node.get(sourceNode.getLabel());
		List<ENode> res = new ArrayList<ENode>();
		for (Integer i : indices) {
			res.add(this.getNodeById(i));
		}

		return res;
	}

	public List<ENode> getNodes() {
		return this.nodes;
	}

	public ENode getNodeById(int id) {
		//TODO
		for(int i = 0; i<nodes.size(); i++ ){
			if(nodes.get(i).getId() == id){
				return nodes.get(i);
			}
		}
		return null;
	}

	public int getSize() {
		return this.nodes.size();
	}

	public int addEdge(EEdge edge) {
		edges.add(edge);
		int index = edges.size() - 1;
		edge.setId(index);
		edge.setGraphId(this.getId());
		return index;
	}

	public List<EEdge> getEdges() {
		return new ArrayList<EEdge>(this.edges);
	}

	public int getId() {
		return id;
	}

	public EEdge getEdge(int newSourceId, int newTargetId) {
		for (EEdge e : getEdges()) {
			if (e.getSourceId() == newSourceId && e.getTargetId() == newTargetId) {
				return e;
			}
		}
		return null;
	}

	public void setGraphDist(int[][] distMatrix) {
		this.distMatrix = distMatrix;
	}

	public int getDist(int sourceId, int targetId) {
		if (sourceId < distMatrix.length && targetId < distMatrix[sourceId].length) {
			return distMatrix[sourceId][targetId];
		}
		return Integer.MAX_VALUE;
	}

	public void setId(int id) {
		this.id = id;
		for(ENode n : this.nodes){
			n.setGraphId(id);
		}
	}
	
	public String toString(){
		String l = "E = " + Joiner.on(",").join(nodes);
		l += "R = " + Joiner.on(",").join(edges);
		return l;
	}

}
