package org.processmining.xlu.matching.models.impl;

import gnu.trove.set.hash.THashSet;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;

public class MatchMappingImpl implements MatchMapping {

	BiMap<Integer, Integer> bimap = HashBiMap.create();
	Set<Integer> sourcesToNull = new THashSet<Integer>();
	Set<Integer> targetsToNull = new THashSet<Integer>();

	EGraph source = null;
	EGraph target = null;

	private int cost;

	public void inverse() {
		BiMap<Integer, Integer> map = bimap.inverse();
		bimap = map;
		Set<Integer> tempSourcesToNull = sourcesToNull;
		this.sourcesToNull = targetsToNull;
		this.targetsToNull = tempSourcesToNull;
		EGraph tempSource = source;
		this.source = target;
		this.target = tempSource;
	}

	public MatchMappingImpl(EGraph source, EGraph target) {
		this.source = source;
		this.target = target;
	}

	public void addMatch(ENode sourceNode, ENode targetNode) {
		if (sourceNode == null) {
			targetsToNull.add(targetNode.getId());
			//			setTargetGraph(targetNode.getGraph());
		} else if (targetNode == null) {
			sourcesToNull.add(sourceNode.getId());
			//			setSourceGraph(sourceNode.getGraph());
		} else {
			bimap.put(sourceNode.getId(), targetNode.getId());
			//			setSourceGraph(sourceNode.getGraph());
			//			setTargetGraph(targetNode.getGraph());
		}
	}

	public void setTargetGraph(EGraph graph) {
		if (target == null)
			this.target = graph;
	}

	public void setSourceGraph(EGraph graph) {
		if (source == null)
			this.source = graph;
	}

	public Collection<Integer> getRang() {
		return bimap.values();
	}

	public Integer getSourceNodeIdMappedTo(int targetNodeId) {
		if (this.targetsToNull.contains(targetNodeId)) {
			return null;
		}
		return bimap.inverse().get(targetNodeId);
	}

	public void addMatchByIds(int newId, int targetNodeId) {
		bimap.put(newId, targetNodeId);
	}

	public void removeMatch(ENode source, ENode target) {
		if (source == null) {
			this.targetsToNull.remove(target.getId());
		} else if (target == null) {
			this.sourcesToNull.remove(source.getId());
		} else {
			bimap.remove(source.getId());
		}
	}

	public void clone(MatchMapping mapping) {
		setBimap(mapping.getMap());
		setSourcesToNull(mapping.getSourcesToNull());
		setTargetsToNull(mapping.getTargetToNull());
		setCost(mapping.getCost());
	}

	public Set<Entry<Integer, Integer>> getMatchs() {
		return bimap.entrySet();
	}

	public Set<Integer> getSourcesToNull() {
		return this.sourcesToNull;
	}

	public Set<Integer> getTargetToNull() {
		return this.targetsToNull;
	}

	private void setTargetsToNull(Set<Integer> targetsToNull) {
		this.targetsToNull = new HashSet<Integer>(targetsToNull);

	}

	private void setSourcesToNull(Set<Integer> sourcesToNull) {
		this.sourcesToNull = new HashSet<Integer>(sourcesToNull);
	}

	private void setBimap(Map<Integer, Integer> map) {
		this.bimap = HashBiMap.create(map);
	}

	public Map<Integer, Integer> getMap() {
		return this.bimap;
	}

	public int getCost() {
		return this.cost;
	}

	public void setCost(int cost) {
		this.cost = cost;

	}

	public Integer getTargetNodeIdMappedTo(int sourceNodeId) {
		if (this.sourcesToNull.contains(sourceNodeId)) {
			return null;
		}
		return bimap.get(sourceNodeId);
	}

	public EGraph getSourceGraph() {
		return this.source;
	}

	public EGraph getTargetGraph() {
		return this.target;
	}

	public boolean inDomain(int sourceNodeId) {
		return 	this.bimap.containsKey(sourceNodeId) || this.sourcesToNull.contains(sourceNodeId);
	}

	public boolean inRange(int targetNodeId) {
		return this.bimap.containsValue(targetNodeId) || this.targetsToNull.contains(targetNodeId);
	}


}
