package org.processmining.xlu.matching.models.impl;

import gnu.trove.map.hash.THashMap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.collection.MatchMappings;

public class MatchMappingsImpl implements MatchMappings {

	Map<Integer, Map<Integer, List<MatchMapping>>> mappings;
	Map<Integer, Map<Integer, Boolean>> sorted;

	public MatchMappingsImpl() {
		mappings = new THashMap<Integer, Map<Integer, List<MatchMapping>>>();
		sorted = new THashMap<Integer, Map<Integer, Boolean>>();
	}

	public MatchMappingsImpl(List<MatchMapping> mappings) {
		this();
		for (MatchMapping m : mappings) {
			int sourceId = m.getSourceGraph().getId();
			int targetId = m.getTargetGraph().getId();
			List<MatchMapping> list = initMap(sourceId, targetId);
			list.add(m);
		}
	}

	@Override
	public List<MatchMapping> getAllMappings() {
		List<MatchMapping> maplist = new ArrayList<MatchMapping>();
		for (Map<Integer, List<MatchMapping>> entry : mappings.values()) {
			for (List<MatchMapping> mappings : entry.values()) {
				maplist.addAll(mappings);
			}
		}
		return maplist;
	}

	@Override
	public List<MatchMapping> getOptimalMappings() {
		List<MatchMapping> maplist = new ArrayList<MatchMapping>();
		for (Entry<Integer, Map<Integer, List<MatchMapping>>> entry : mappings.entrySet()) {
			int i = entry.getKey();
			for (Entry<Integer, List<MatchMapping>> mappings : entry.getValue().entrySet()) {
				int j = mappings.getKey();
				if (!this.sorted.get(i).get(j)) {
					Collections.sort(mappings.getValue(), getMappingComparator());
					this.sorted.get(i).put(j, true);
				}
				if (mappings.getValue().size() > 0) {
					maplist.add(mappings.getValue().get(0));
				}
			}
		}
		return maplist;
	}

	@Override
	public List<MatchMapping> getAllMappings(int i, int j) {
		Map<Integer, List<MatchMapping>> mapI = mappings.get(i);
		if (mapI == null) {
			int tempi = i;
			i = j;
			j = tempi;
			mapI = mappings.get(i);
			if (mapI == null) {
				return new ArrayList<MatchMapping>();
			}
		}
		List<MatchMapping> list = mapI.get(j);
		if (list == null) {
			list = new ArrayList<MatchMapping>();
		}
		return list;
	}

	@Override
	public MatchMapping getAnOptimalMapping(int i, int j) {
		int source = i;
		int target = j;
		boolean inverse = false;
		if (i > j) {
			source = j;
			target = i;
			inverse = true;
		}
		List<MatchMapping> list = getAllMappings(source, target);

		MatchMapping mapping = null;

		if (list == null || list.isEmpty()) {
			return mapping;
		}
		
		if (list.size() == 1) {
			mapping = list.get(0);
		} else {
			// more than one optimal, sort
			Map<Integer, Boolean> map = sorted.get(source);
			if (!map.get(target)) {
				Collections.sort(list, getMappingComparator());
				map.put(target, true);
			}

			mapping = list.get(0);
		}

		if (inverse) {
			MatchMapping inversed = new MatchMappingImpl(mapping.getSourceGraph(), mapping.getTargetGraph());
			inversed.clone(mapping);
			inversed.inverse();
			return inversed;
		}
		
		return mapping;
	}

	public Iterator<MatchMapping> iterator() {
		return getAllMappings().iterator();
	}

	public void add(int i, int j, List<MatchMapping> mappings) {

		List<MatchMapping> list = initMap(i, j);
		list.addAll(mappings);

	}

	private List<MatchMapping> initMap(int i, int j) {
		int sourceId = i;
		int targetId = j;
		// init if no mapping yet
		Map<Integer, List<MatchMapping>> map = this.mappings.get(sourceId);
		if (map == null) {
			map = new THashMap<Integer, List<MatchMapping>>();
			this.sorted.put(i, new THashMap<Integer, Boolean>());
		}
		List<MatchMapping> list = map.get(targetId);
		if (list == null) {
			list = new ArrayList<MatchMapping>();
		}
		map.put(targetId, list);
		this.mappings.put(sourceId, map);
		sorted.get(i).put(j, false);
		return list;
	}

	public Comparator<MatchMapping> getMappingComparator() {
		return new MappingCostComparator();
	}

	/**
	 * Two mappings between the same two graphs with the same cost is considered
	 * the same...
	 */
	public class MappingCostComparator implements Comparator<MatchMapping> {
		@Override
		public int compare(MatchMapping m1, MatchMapping m2) {
			int cost = m1.getCost() - m2.getCost();
			if (cost != 0) {
				return cost;
			} else {
				int i = m1.getSourceGraph().getId() - m2.getSourceGraph().getId();
				if (i != 0) {
					return i;
				} else {
					int j = m1.getTargetGraph().getId() - m2.getTargetGraph().getId();
					return j;
				}
			}
		}
	}

	public void add(int i, int j, MatchMapping mapping) {
		List<MatchMapping> list = initMap(i, j);
		list.add(mapping);

	}

}
