package org.processmining.xlu.matching.models.impl;

import gnu.trove.map.hash.THashMap;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.collections15.map.HashedMap;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;

import com.google.common.base.Joiner;

public class RepEGraphImpl extends EGraphImp implements RepEGraph {

	/* Objects */
	List<Integer> mappedGraphs;
	Map<Integer, Map<Integer, Integer>> mapNodeToGraphToNode;
	Map<Integer, Map<Integer, Integer>> mapEdgeToGraphToEdge;

	public RepEGraphImpl() {
		super();
		mappedGraphs = new ArrayList<Integer>();
		mapNodeToGraphToNode = new HashedMap<Integer, Map<Integer, Integer>>();
		mapEdgeToGraphToEdge = new HashedMap<Integer, Map<Integer, Integer>>();
	}

	public void addRepresentedGraph(int graphId) {
		this.mappedGraphs.add(graphId);
	}

	@Override
	public void removeNode(ENode node) {

		List<EEdge> ins = new ArrayList<EEdge>();
		List<EEdge> outs = new ArrayList<EEdge>();
		for (EEdge e : edges) {
			if (e.getSourceId() == node.getId()) {
				outs.add(e);
			} else if (e.getTargetId() == node.getId()) {
				ins.add(e);
			}
		}
		for (EEdge in : ins) {
			ENode source = this.getNodeById(in.getSourceId());
			Map<Integer, Integer> inMap = this.mapEdgeToGraphToEdge.get(in.getId());

			for (EEdge out : outs) {
				ENode target = this.getNodeById(out.getTargetId());
				Map<Integer, Integer> outMap = this.mapEdgeToGraphToEdge.get(out.getId());

				EEdge e = new EEdgeImpl(source.getId(), target.getId());
				int newid = this.addEdge(e);
				if (inMap != null && outMap != null) {
					for (Entry<Integer, Integer> entry : inMap.entrySet()) {
						Integer inGraph = entry.getKey();
						if (outMap.containsKey(inGraph)) {
							//TODO one Edge representing Two edges in the same graph
							this.updateEdgeRepresentedEdge(newid, inGraph, entry.getValue());
						}
					}
				}

			}
		}

		nodes.remove(node);
		this.mapNodeToGraphToNode.remove(node.getId());
		edges.removeAll(ins);
		for (EEdge i : ins) {
			this.mapEdgeToGraphToEdge.remove(i.getId());
		}
		edges.removeAll(outs);
		for (EEdge i : outs) {
			this.mapEdgeToGraphToEdge.remove(i.getId());
		}
	}

	public void updateNodeRepresentedNode(int nodeId, int orgGraphId, int orgNodeId) {
		Map<Integer, Integer> map = this.mapNodeToGraphToNode.get(nodeId);
		if (map == null) {
			map = new HashedMap<Integer, Integer>();

		}
		map.put(orgGraphId, orgNodeId);
		mapNodeToGraphToNode.put(nodeId, map);
	}

	public void updateEdgeRepresentedEdge(int edgeId, int orgGraphId, int orgEdgeId) {
		Map<Integer, Integer> map = this.mapEdgeToGraphToEdge.get(edgeId);
		if (map == null) {
			map = new HashedMap<Integer, Integer>();

		}
		map.put(orgGraphId, orgEdgeId);
		mapEdgeToGraphToEdge.put(edgeId, map);
	}

	public List<Integer> getRepresentedGraphIds() {
		return this.mappedGraphs;
	}

	public Map<Integer, Map<Integer, Integer>> getEdgeRepresentedEdges() {
		return this.mapEdgeToGraphToEdge;
	}

	public Map<Integer, Map<Integer, Integer>> getNodeRepresentedNodes() {
		return this.mapNodeToGraphToNode;
	}

	//TODO : these are ad-hoc properties implementation. To be generalized. ===================================
	int cost;
	Map<Integer, Integer> mapNodeToCost = new THashMap<Integer, Integer>();
	Map<Integer, Boolean> mapNodeToCorrect = new THashMap<Integer, Boolean>();
	Map<String, Integer> mapCFNameToCost = new THashMap<String, Integer>();

	public int getCost() {
		return cost;
	}

	public void setCost(int cost) {
		this.cost = cost;

	}

	public Map<Integer, Integer> getNodeToCost() {
		return this.mapNodeToCost;
	}

	public void setNodeToCost(Map<Integer, Integer> nodeToCost) {
		this.mapNodeToCost = nodeToCost;

	}

	public Map<Integer, Boolean> getNodeToMappedCorrectly() {
		return this.mapNodeToCorrect;
	}

	public void setNodeToCorrect(Map<Integer, Boolean> nodeToCorrect) {
		this.mapNodeToCorrect = nodeToCorrect;
	}

	public ENode getNodeRepresenting(int graphId, int nodeId) {
		for (ENode n : this.getNodes()) {
			Map<Integer, Integer> map = this.getNodeRepresentedNodes().get(n.getId());
			if (map.get(graphId) != null && map.get(graphId) == nodeId) {
				return n;
			}
		}
		return null;
	}

	public String toString() {
		return Joiner.on(",").join(this.getRepresentedGraphIds());
	}

	public int getCost(String costFunctionName) {
		if (mapCFNameToCost != null) {
			Integer cost = mapCFNameToCost.get(costFunctionName);
			if (cost != null) {
				return cost;
			}
		}
		return Integer.MAX_VALUE;
	}

	public void setCost(String costFunctionName, int cost) {
		if (mapCFNameToCost == null) {
			mapCFNameToCost = new HashedMap<String, Integer>();
		}
		mapCFNameToCost.put(costFunctionName, cost);
	}

	public Map<String, Integer> getCosts() {
		return this.mapCFNameToCost;
	}
}
