package org.processmining.xlu.matching.plugins;

import org.deckfour.xes.classification.XEventClasses;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.partialorder.ptrace.param.PTraceParameter;
import org.processmining.xlu.matching.merge.FusionStrategy;
import org.processmining.xlu.matching.merge.impl.FusionStrategyBinaryImpl;
import org.processmining.xlu.matching.merge.impl.FusionStrategyBinaryVerificationImpl;
import org.processmining.xlu.matching.merge.impl.FusionStrategyIncrBinaryImpl;
import org.processmining.xlu.matching.merge.impl.FusionStrategyIncrFilteringImpl;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.models.collection.REGraphType;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.zexperiment.EGFusionTest;
import org.processmining.xlu.matching.zexperiment.model.QualityTFPN;
import org.processmining.xlu.matching.zexperiment.verification.EGMatchTestVerificationMethodGeneratedImpl;

@Plugin(name = "Execution Graph Fusion", returnLabels = { "RepEGraphs" }, returnTypes = { RepEGraphs.class },
		parameterLabels = { "Event Log", "Parameters" }, categories = { PluginCategory.Discovery,
				PluginCategory.Analytics }, keywords = { "Clustering" },
		help = "Merging the execution graphs accordingly", userAccessible = true, level = PluginLevel.Local)
public class EGFusionPlugin {

	private PluginContext context;
	private PTraceParameter parameter;

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl",
			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Execution Graph Fusion - Binary BF", requiredParameterLabels = { 0 })
	public RepEGraphs fuseBinaryGraphs(PluginContext context, EGraphsWithMappings graphsWithMappings) {
		initContext(context, graphsWithMappings.getGraphs().size());
		XEventClasses classes = graphsWithMappings.getMatchConfiguration().getClasses();
		MatchMappings mapGraphsToMapping = graphsWithMappings.getMappings();
		EGMatchConfiguration conf = graphsWithMappings.getMatchConfiguration();

		//		FusionStrategy fusionStrategy = new FusionStrategyIncrBinaryImpl();
		FusionStrategy fusionStrategy = new FusionStrategyBinaryImpl(graphsWithMappings.getPlog(), conf, 0);

		RepEGraphs mergedGraphs = fusionStrategy.fuseEGraphs(graphsWithMappings, mapGraphsToMapping);
		mergedGraphs.setClasses(classes);
		setLogs(mergedGraphs, graphsWithMappings);
		//		mergedGraphs.setConverter(new EGraphToPGraphConverterImpl(graphsWithMappings.getPlog().getXLog(), classes,
		//				mergedGraphs));

		return mergedGraphs;
	}

	private void setLogs(RepEGraphs mergedGraphs, EGraphsWithMappings graphsWithMappings) {
		mergedGraphs.setLog(graphsWithMappings.getPlog().getXLog());
		mergedGraphs.setpLog(graphsWithMappings.getPlog());
		mergedGraphs.setMappings(graphsWithMappings.getMappings());
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl",
			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Execution Graph Fusion - Incremental", requiredParameterLabels = { 0 })
	public RepEGraphs fuseIncGraphs(PluginContext context, EGraphsWithMappings graphsWithMappings) {
		initContext(context, graphsWithMappings.getGraphs().size());
		
		XEventClasses classes = graphsWithMappings.getMatchConfiguration().getClasses();
		MatchMappings mapGraphsToMapping = graphsWithMappings.getMappings();
		EGMatchConfiguration conf = graphsWithMappings.getMatchConfiguration();

		//		RepEGraphs mergedGraphs = new RepEGraphs();

		FusionStrategy fusionStrategy = new FusionStrategyIncrBinaryImpl(context, graphsWithMappings.getPlog().getXLog());
		//		FusionStrategy fusionStrategy = new FusionStrategyBruteForceBinaryVerificationImpl(
		//		graphsWithMappings.getPlog(), conf, 0);

		RepEGraphs mergedGraphs = fusionStrategy.fuseEGraphs(graphsWithMappings, mapGraphsToMapping);

//		for (int threshold = 0; threshold < 100; threshold += 10) {
//			QualityTFPN data = EGFusionTest.computeFusionQualityNoise(graphsWithMappings
//					.getPlog().getXLog(), mergedGraphs.getHierarchy(), threshold);
//
//			EGFusionTest.printSystemResult(conf, threshold, data);
//
//		}
		mergedGraphs.setClasses(classes);
		mergedGraphs.setType(REGraphType.HIERARCHICAL);
		setLogs(mergedGraphs, graphsWithMappings);
		return mergedGraphs;
	}

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl",
//			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Execution Graph Fusion - Verify GenLog", requiredParameterLabels = { 0 })
	public RepEGraphs fuseVerifyIncGraphs(UIPluginContext context, EGraphsWithMappings graphsWithMappings) {
		initContext(context, graphsWithMappings.getGraphs().size());
		
		XEventClasses classes = graphsWithMappings.getMatchConfiguration().getClasses();
		MatchMappings mapGraphsToMapping = graphsWithMappings.getMappings();
		EGMatchConfiguration conf = graphsWithMappings.getMatchConfiguration();

		//		RepEGraphs mergedGraphs = new RepEGraphs();

		FusionStrategy fusionStrategy = new FusionStrategyBinaryVerificationImpl(graphsWithMappings.getPlog(), true,
				new EGMatchTestVerificationMethodGeneratedImpl());
		//		FusionStrategy fusionStrategy = new FusionStrategyBruteForceBinaryVerificationImpl(
		//				graphsWithMappings.getPlog(), conf, 0);

		RepEGraphs mergedGraphs = fusionStrategy.fuseEGraphs(graphsWithMappings, mapGraphsToMapping);
		mergedGraphs.setClasses(classes);
		setLogs(mergedGraphs, graphsWithMappings);
		return mergedGraphs;
	}

	
	/* ----------------------------------------------------------------
	 *
	 *   Implemented but never tested.
	 *
	 * ---------------------------------------------------------------- */
	 
//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl",
//			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Execution Graph Fusion - Incremental Filter", requiredParameterLabels = { 0 })
	public RepEGraphs fuseIncGraphsIncFilter(PluginContext context, EGraphsWithMappings graphsWithMappings) {
		initContext(context, graphsWithMappings.getGraphs().size());
		
		XEventClasses classes = graphsWithMappings.getMatchConfiguration().getClasses();
		MatchMappings mapGraphsToMapping = graphsWithMappings.getMappings();
		EGMatchConfiguration conf = graphsWithMappings.getMatchConfiguration();

		//		RepEGraphs mergedGraphs = new RepEGraphs();

		FusionStrategy fusionStrategy = new FusionStrategyIncrFilteringImpl(graphsWithMappings.getPlog().getXLog());
		//		FusionStrategy fusionStrategy = new FusionStrategyBruteForceBinaryVerificationImpl(
		//				graphsWithMappings.getPlog(), conf, 0);

		RepEGraphs mergedGraphs = fusionStrategy.fuseEGraphs(graphsWithMappings, mapGraphsToMapping);

		for (int threshold = 0; threshold < 100; threshold += 10) {
			QualityTFPN data = EGFusionTest.computeFusionQualityNoise(graphsWithMappings
					.getPlog().getXLog(), mergedGraphs.getHierarchy(), threshold);

			EGFusionTest.printSystemResult(conf, threshold, data);
		}
		
		mergedGraphs.setClasses(classes);
		mergedGraphs.setType(REGraphType.HIERARCHICAL);
		setLogs(mergedGraphs, graphsWithMappings);
		return mergedGraphs;
	}
	
	private void initContext(PluginContext context, int size) {
		this.context = context;
		if (this.context != null) {
			
			this.context.log("Started Fusing Execution Graphs...");
			this.context.getProgress().setValue(0);
			this.context.getProgress().setMaximum(size * size / 2);
		}
	}
}
