package org.processmining.xlu.matching.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.RepEGraphs;

@Plugin(name = "Matching Framework", returnLabels = { "RepEGraphs" }, returnTypes = { RepEGraphs.class }, parameterLabels = {
		"Event Log", "Parameters" }, categories = { PluginCategory.Discovery,
		PluginCategory.Analytics }, keywords = { "Clustering" }, help = "Compute mappings between traces", userAccessible = true)
public class EGMatchFusionAndRelabelPlugin {

	
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "L2Me: Log to Model Explorer (GUI)", requiredParameterLabels = { 0 })
	public RepEGraphs computeTraceMatchingViaGui(UIPluginContext context,
			XLog log) {
		EGMatchPlugin pluginMatcher = new EGMatchPlugin();
		EGraphsWithMappings mappings = pluginMatcher.computeTraceMatchingViaGui(context, log);
		
		EGFusionPlugin pluginFuser = new EGFusionPlugin();
		RepEGraphs fusedgraphs = pluginFuser.fuseIncGraphs(context, mappings);
		return fusedgraphs;
	}
	
	/**
	 * 
	 * @param context This input context could be null. 
	 * @param log
	 * @return
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "L2Me: Log to Model Explorer (Default)", requiredParameterLabels = { 0 })
	public RepEGraphs computeTraceMatchingDefault(PluginContext context,
			XLog log) {
		EGMatchPlugin pluginMatcher = new EGMatchPlugin();
		EGraphsWithMappings mappings = pluginMatcher.computeTraceMatchingGreedy(context, log);
		
		EGFusionPlugin pluginFuser = new EGFusionPlugin();
		RepEGraphs fusedgraphs = pluginFuser.fuseIncGraphs(context, mappings);
		return fusedgraphs;
	}
}
