package org.processmining.xlu.matching.plugins;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.partialorder.ptrace.model.PTrace;
import org.processmining.partialorder.ptrace.param.PTraceParameter;
import org.processmining.partialorder.ptrace.param.PTraceParameter.PTraceType;
import org.processmining.partialorder.ptrace.plugins.builder.PLogPlugin;
import org.processmining.xlu.matching.config.alg.EGMatchUserOrientedConfigAlg;
import org.processmining.xlu.matching.config.gui.EGMatchConfigUserPresetWizard;
import org.processmining.xlu.matching.config.gui.EGMatchConfigWizard;
import org.processmining.xlu.matching.models.Context;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.models.impl.ENodeProperties;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.models.impl.MatchMappingsImpl;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.plugins.alg.ENodeContextComparator;
import org.processmining.xlu.matching.plugins.alg.HeuristicStrategy;
import org.processmining.xlu.matching.plugins.alg.MatchingAlg;
import org.processmining.xlu.matching.plugins.alg.MatchingAllAlg;
import org.processmining.xlu.matching.plugins.alg.conf.impl.EGMatchConfigurationImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostFunctionImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostHeuristicFunctionComplexImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyConfiguration;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyWeightsImpl;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAlgAbstract;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAlgBackTrackingPruningImpl;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAlgGreedyImp;
import org.processmining.xlu.matching.plugins.alg.impl.MatchingAllAlgBackTrackingBFImpl;
import org.processmining.xlu.matching.preprocessing.PTraceToEGConverter;
import org.processmining.xlu.matching.preprocessing.PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl;

import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;
import nl.tue.astar.Trace;
import nl.tue.astar.util.PartiallyOrderedTrace;

/**
 * 
 * @author xlu
 * 
 */

@Plugin(name = "EGraph Mapping", returnLabels = {
		"EGraphs with mapping" }, returnTypes = {
				EGraphsWithMappings.class }, parameterLabels = { "Event Log",
						"Parameters" }, categories = { PluginCategory.Discovery,
								PluginCategory.Analytics }, keywords = {
										"Clustering" }, help = "Compute mappings between traces", userAccessible = true, level = PluginLevel.Local)
public class EGMatchPlugin {

	private PluginContext context;
	private PTraceParameter parameter;
	public double mappingEstimatedTime;
	public double computingEGEstimatedTime;

	/* ----------------------------------------------------------------
	 *
	 *   Individual Setting : Brute force alg.
	 *
	 * ---------------------------------------------------------------- */

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "EGraph Mappings - BF", requiredParameterLabels = { 0 })
	public EGraphsWithMappings computeTraceMatching(UIPluginContext context, XLog log) {

		initContext(context, log);
		this.parameter = new PTraceParameter(PTraceType.Data_Dependency);
		this.parameter.setComputeTransitive(true);
//		PLog pLog = PTraceConstructionFactory.callContructPLogWorkflow(context, log, parameter);
//		PLog pLog =  (new PLogPlugin()).computePTraces(context, log, parameter);

		
		EGMatchConfiguration conf = new EGMatchConfigurationImpl();
		// XEventClassifier classifier = new XEventAttributeClassifier("Org", "Type_Code"); // TODO
		XEventClassifier classifier = new XEventNameClassifier();
		// XEventClassifier classifier = new XEventAttributeClassifier("Org", "TrueLabel");
		XEventClasses classes = XEventClasses.deriveEventClasses(classifier, log);

		int k = 1;
		// for(int k = 1 ; k < 5; k++){
		
		//EGMatchConfigurationFactory conf = EGMatchConfigurationFactoryAbstract.experiment2MappingIndepWithConc(k);
		conf.setClasses(classes);
		//conf.setClassifier(classifier);
		conf.setConverter(
				new PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl(k, classes));
		// Init matching algorithm
		// MatchingAlg matchingAlg = conf.getMatchingAlgorithm();
		MatchingAllAlg matchingAlg = new MatchingAllAlgBackTrackingBFImpl();
		// MatchingAlg matchingAlg = new MatchingAlgDecomposedBackTrackingImpl();
		CostStrategy costfunction = new CostFunctionImpl();
		matchingAlg.setCostFunction(costfunction);
		conf.setMatchingAlgorithm(matchingAlg);
		conf.setCostStrategy(costfunction);

		return replayLog(context, log, conf);
		// return null;
	}

	/* ----------------------------------------------------------------
	 *
	 *   Individual Setting : greedy algorithm
	 *
	 * ---------------------------------------------------------------- */
//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "EGraph Mappings - Greedy", requiredParameterLabels = { 0 })
	public EGraphsWithMappings computeTraceMatchingGreedy(PluginContext context,
			XLog log) {

		initContext(context, log);
		this.parameter = new PTraceParameter(PTraceType.Sequential_Dependency);
		this.parameter.setComputeTransitive(true);
//		PLog pLog = PTraceConstructionFactory.callContructPLogWorkflow(context, log, parameter);

		//XEventClassifier classifier = new XEventAttributeClassifier("Org", "Verrichttekst_1");
		//XEventClassifier classifier = new XEventAttributeClassifier("Activity", "IPC6Label");
		//XEventClasses classes = XEventClasses.deriveEventClasses(classifier, log);
		//MUDATAPreprocess.updateMUDataLog(log);

		XEventClassifier classifier = new XEventNameClassifier();
		XEventClasses classes = XEventClasses.deriveEventClasses(classifier, log);

		int k = 1;
		EGMatchConfiguration conf = new EGMatchConfigurationImpl();
		PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl converter = new PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl(
				k, classes);
		conf.setClasses(classes);
		conf.setConverter(converter);
		MatchingAlgGreedyImp matchingAlg = new MatchingAlgGreedyImp();
		matchingAlg.setNodeComparator(new ENodeContextComparator());

		CostStrategy costfunction = new CostFunctionImpl();
		CostStrategyConfiguration csConf = new CostStrategyWeightsImpl();
		csConf.setWeightMatched(3);
		csConf.setWeightNoMatch(3);
		csConf.setWeightGlobalStructure(1);
		csConf.setWeightBaseCostForNoMatch(5);
		costfunction.setCostConfiguration(csConf);
		//Graph2: 		2, 2, 1, 1, 5
		//Orig code: 	1, 2, 3, 1, 5.
		//Top 3: 		2-1-1-1-5

		matchingAlg.setCostFunction(costfunction);
		HeuristicStrategy heuristicFunction = new CostHeuristicFunctionComplexImpl();
		matchingAlg.setHeuristicfunction(heuristicFunction);

		conf.setMatchingAlgorithm(matchingAlg);
		conf.setCostStrategy(costfunction);

		return replayLog(context, log, conf);
		//		return null;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "EGraph Mappings - GUI", requiredParameterLabels = {
			0 })
	public EGraphsWithMappings computeTraceMatchingViaGui(UIPluginContext context,
			XLog log) {
		EGraphsWithMappings res = null;

		// All gui sets
		EGMatchConfiguration conf = new EGMatchConfigurationImpl();
		EGMatchConfigWizard wizard = new EGMatchConfigWizard(context, conf, log);
		TaskListener.InteractionResult dialog = wizard.showWizard();
		if (dialog.equals(TaskListener.InteractionResult.FINISHED)) {

			// Run main plugin
			res = replayLog(context, log, wizard.getConfiguration());

		} else if (dialog.equals(TaskListener.InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
		}
		return res;
	}

	/* ----------------------------------------------------------------
	 *
	 *   "Auto" configuration: wanted to try but abandon.
	 *
	 * ---------------------------------------------------------------- */
//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "EGraph Mappings - User", requiredParameterLabels = { 0 })
	public EGraphsWithMappings computeTraceMatchingViaUser(UIPluginContext context,
			XLog log) {
		EGraphsWithMappings res = null;

		// All gui sets
		EGMatchConfiguration conf = new EGMatchConfigurationImpl();
		EGMatchConfigUserPresetWizard wizard = new EGMatchConfigUserPresetWizard(context,
				conf, log);
		TaskListener.InteractionResult finished = wizard.showWizard();
		if (finished.equals(TaskListener.InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
			return null;
		}

		EGMatchUserOrientedConfigAlg supervision = new EGMatchUserOrientedConfigAlg(
				context, log, conf);
		finished = supervision.showWizard();
		if (finished.equals(TaskListener.InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
			return null;
		}

		// Run main plugin
		EGMatchConfiguration matchingConfig = supervision.getConfiguration();

		res = replayLog(context, log, matchingConfig);
		return res;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", pack = "")
	@PluginVariant(variantLabel = "default", requiredParameterLabels = { 0 })
	public EGraphsWithMappings replayLog(PluginContext context, XLog log,
			EGMatchConfiguration conf) {
		return matchingTraces(log, conf);
	}

	public EGraphsWithMappings matchingTraces(XLog log, EGMatchConfiguration conf) {
		XEventClasses classes = XEventClasses.deriveEventClasses(conf.getClassifier(),
				log);
		conf.setClasses(classes);

		PLog pLog = PLogPlugin.computePTraces(log, conf.getPTraceParameter());

		PTraceToEGConverter converter = conf.getConverter();
		converter.setClasses(classes);
		MatchingAlg matchingAlg = conf.getMatchingAlgorithm();
		MatchMappings mapGraphsToMapping = new MatchMappingsImpl();
		if (conf.getLabels() != null) {
			matchingAlg.setMatchingLabelSet(conf.getLabels());
		}
		// Preprocessing
		List<EGraph> graphs = convertPTracesToComparableGraphs(pLog, conf.getClasses(),
				converter);
		if (matchingAlg instanceof MatchingAlgGreedyImp
				|| matchingAlg instanceof MatchingAlgBackTrackingPruningImpl) {
			HeuristicStrategy heuristic = ((MatchingAlgAbstract) matchingAlg)
					.getHeuristicfunction();
			if (heuristic == null) {
				heuristic = new CostHeuristicFunctionComplexImpl();
				heuristic.setCostConfiguration(
						conf.getCostStrategy().getCostConfiguration());
				((MatchingAlgAbstract) matchingAlg).setHeuristicfunction(heuristic);
			}
			heuristic.setCostConfiguration(conf.getCostStrategy().getCostConfiguration());
			if (heuristic instanceof CostHeuristicFunctionComplexImpl) {
				((CostHeuristicFunctionComplexImpl) heuristic)
						.setMapLabelToIndex(converter.getMapNodeLabelToIndex());

			}
		}

		// Compute mappings
		long s = System.currentTimeMillis();

		// Compute representatives
		Map<Trace, Integer> mapRep2Idx = new THashMap<>();
		Map<Integer, Integer> mapEachTrc2RepIdx = new THashMap<Integer, Integer>();
		for (int i = 0; i < graphs.size(); i++) {
			EGraph g = graphs.get(i);
			Trace t = getEfficientTrace(g, classes);
			Integer idx = mapRep2Idx.get(t);
			if (idx != null) {
				mapEachTrc2RepIdx.put(i, idx);
			} else {
				mapRep2Idx.put(t, i);
				mapEachTrc2RepIdx.put(i, i);
			}
		}

		// Compute mappings
		for (int i = 0; i < graphs.size(); i++) {
			
			if (i % 100 == 0) {
				System.out.format("Start with %d / %d at %d. %n", i, graphs.size(),
						System.currentTimeMillis());
			}
			EGraph g1 = graphs.get(i);
			for (int j = i + 1; j < graphs.size(); j++) {
				if(context != null){
					context.getProgress().inc();
				}
				
				// System.out.format("Start with j = %d / %d. %n", j, graphs.size());
				EGraph g2 = graphs.get(j);

				int idxI = mapEachTrc2RepIdx.get(i);
				int idxJ = mapEachTrc2RepIdx.get(j);
				if (idxJ == idxI) {
//					MatchMappingImpl newm = new MatchMappingImpl(g1, g2);
//					for (int k = 0; k < g1.getNodes().size(); k++) {
//						newm.addMatch(g1.getNodeById(k), g2.getNodeById(k));
//					}
//					mapGraphsToMapping.add(i, j, newm);
//					continue;
					idxJ = j;
				}
				List<MatchMapping> mappings = mapGraphsToMapping.getAllMappings(idxI,
						idxJ);
				if (mappings != null && mappings.size() > 0) {
					List<MatchMapping> clones = new ArrayList<MatchMapping>();
					for (MatchMapping m : mappings) {
						MatchMapping newm = null;
						if (m.getSourceGraph().getId() == idxI) {
							newm = new MatchMappingImpl(g1, g2);
							newm.clone(m);
						} else {
							newm = new MatchMappingImpl(g2, g1);
							newm.clone(m);
							newm.inverse();
						}

						clones.add(newm);
					}
					mapGraphsToMapping.add(i, j, clones);
				} else {

					if (matchingAlg instanceof MatchingAllAlg) {
						mappings = ((MatchingAllAlg) matchingAlg).getAllOptimalMapping(g1,
								g2);
						mapGraphsToMapping.add(i, j, mappings);
					} else {
						MatchMapping mapping = matchingAlg.getOptimalMapping(g1, g2);
						mapGraphsToMapping.add(i, j, mapping);
					}
				}
			}
		}
		long e = System.currentTimeMillis();
		mappingEstimatedTime = (e - s) / 1000.0 / 60.0;
		System.out.format("Computing mapping took %f min. %n", mappingEstimatedTime);

		EGraphsWithMappings graphsWithMappings = new EGraphsWithMappings(graphs,
				mapGraphsToMapping, conf);
		graphsWithMappings.setPlog(pLog);
		return graphsWithMappings;
	}

	private Trace getEfficientTrace(EGraph ptrc, XEventClasses classes) {
		int size = ptrc.getNodes().size();
		int[] activitySequence = new int[size];
		int[][] predecessors = new int[size][];

		List<List<Integer>> predecessorLists = new ArrayList<List<Integer>>(size);
		for (int i = 0; i < size; i++) {
			XEventClass c = classes.getByIdentity(ptrc.getNodeById(i).getLabel());
			activitySequence[i] = c == null ? 0 : c.getIndex();
			predecessorLists.add(new ArrayList<Integer>());
		}

		for (EEdge dep : ptrc.getEdges()) {
			List<Integer> preds = predecessorLists.get(dep.getTargetId());

			preds.add(dep.getSourceId());
			predecessorLists.set(dep.getTargetId(), preds);
		}
		for (int i = 0; i < size; i++) {
			List<Integer> list = predecessorLists.get(i);
			predecessors[i] = new int[list.size()];
			for (int j = 0; j < list.size(); j++) {
				predecessors[i][j] = list.get(j);
			}

		}
		Trace t = new PartiallyOrderedTrace(String.valueOf(ptrc.getId()),
				activitySequence, predecessors);

		return t;
	}

	private List<EGraph> convertPTracesToComparableGraphs(PLog pLog,
			XEventClasses classes, PTraceToEGConverter converter) {
		//		converter.setClasses(classes);
		long s = System.currentTimeMillis();
		List<EGraph> graphs = new ArrayList<EGraph>();

		for (PTrace ptrace : pLog) {
			EGraph g = converter.convert(ptrace);
			graphs.add(g);
		}

		if (converter instanceof PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl) {
			Map<Context, Integer> mapC2Freq = ((PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl) converter)
					.getMapContextToFreq();
			Map<String, Set<Context>> mapC2Num = computeLabelFrequent(mapC2Freq);
			double maxContextFreq = Double.MIN_VALUE;
			double maxLabelFreq = Double.MIN_VALUE;
			for (Integer v : mapC2Freq.values()) {
				if (v > maxContextFreq) {
					maxContextFreq = (double) v;
				}
			}
			for (Set<Context> c : mapC2Num.values()) {
				int size = c.size();
				if (size > maxLabelFreq) {
					maxLabelFreq = size;
				}
			}
			for (EGraph g : graphs) {
				for (ENode e : g.getNodes()) {

					Context context = (Context) e
							.getProperty(ENodeProperties.KEY_CONTEXT);

					if (context != null) {
						Integer i = mapC2Freq.get(context);
						e.setProperty(ENodeProperties.KEY_CONTEXTFREQ, i);
						int labelSize = mapC2Num.get(context.getLabel()).size();
						e.setProperty(ENodeProperties.KEY_LABELFREQ, labelSize);
						e.setProperty(ENodeProperties.KEY_CONTEXTFREQ_PROP,
								(double) i / maxContextFreq);
						e.setProperty(ENodeProperties.KEY_LABELFREQ_PROP,
								(double) labelSize / maxLabelFreq);
					}
				}
			}
		}
		long e = System.currentTimeMillis();
		computingEGEstimatedTime = (e - s) / 1000.0 / 60.0;
		System.out.format("Computing EGs took %f min. %n", computingEGEstimatedTime);
		return graphs;
	}

	private Map<String, Set<Context>> computeLabelFrequent(
			Map<Context, Integer> mapC2Freq) {
		Map<String, Set<Context>> mapC2num = new HashMap<String, Set<Context>>();
		for (Context c : mapC2Freq.keySet()) {
			String label = c.getLabel();
			Set<Context> set = mapC2num.get(label);
			if (set == null) {
				set = new THashSet<Context>();
			}
			set.add(c);
			mapC2num.put(label, set);
		}
		return mapC2num;
	}

	private void initContext(PluginContext context, XLog origLog) {
		this.context = context;
		if (this.context != null) {
			this.context.getProgress().setMaximum(origLog.size() * origLog.size() / 2);
			this.context.log("Started Matching...");
		}
	}

}
