package org.processmining.xlu.matching.plugins.alg.conf.impl;

import gnu.trove.set.hash.THashSet;

import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.processmining.partialorder.ptrace.param.PTraceParameter;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.plugins.alg.MatchingAlg;
import org.processmining.xlu.matching.plugins.alg.impl.CostFunctionImpl;
import org.processmining.xlu.matching.plugins.alg.impl.CostStrategyConfiguration;
import org.processmining.xlu.matching.preprocessing.PTraceToEGConverter;
import org.processmining.xlu.matching.preprocessing.PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl;


public class EGMatchConfigurationImpl implements EGMatchConfiguration {

	
	//	private boolean doConc;
	private XEventClasses classes;
	private XEventClassifier classifier;

	private MatchingAlg alg;
	private PTraceToEGConverter converter;
	private CostStrategy costfunction;
	private boolean isQualityOpt;

	private Set<String> labels;
	
	private PTraceParameter ptraceParm;
	
	private int dist = 1;
	private CostStrategyConfiguration costStrategyConfig;

	public EGMatchConfigurationImpl() {
	}

	public EGMatchConfigurationImpl(int dist, boolean b) {
		this.dist = dist;
	}

	public int getDist() {
		return dist;
	}

	public void setDist(int dist) {
		this.dist = dist;
	}

	public XEventClasses getClasses() {
		return this.classes;
	}

	public XEventClassifier getClassifier() {
		if(this.classifier == null){
			this.classifier = new XEventNameClassifier();
		}
		return this.classifier;
	}

	public void setClasses(XEventClasses classes) {
		this.classes = classes;
	}

	public void setClassifier(XEventClassifier classifier) {
		this.classifier = classifier;
	}

	public PTraceToEGConverter getConverter() {
		if (converter == null && this.getClasses() != null) {
			this.converter = new PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl(this.getDist(), getClasses());
		}
		return this.converter;
	}

	public MatchingAlg getMatchingAlgorithm() {
		return this.alg;
	}

	public CostStrategy getCostStrategy() {
		if (this.costfunction == null) {
			this.costfunction = new CostFunctionImpl();
		}
		return this.costfunction;
	}

	public void setConverter(PTraceToEGConverter converter) {
		this.converter = converter;
	}

	public void setMatchingAlgorithm(MatchingAlg alg) {
		this.alg = alg;
		this.alg.setCostFunction(getCostStrategy());
	}

	public void setCostStrategy(CostStrategy costfunction) {
		this.costfunction = costfunction;
	}

	public void setQualityOpt(boolean b) {
		this.isQualityOpt = b;

	}

	public boolean isQualityOpt() {
		return isQualityOpt;
	}

	public String toString() {
		return this.alg.toString() + "-" + this.ptraceParm.toString() + "-D" + this.dist + "M" + this.costfunction.getCostConfiguration().getWeightMatched() + "N"
				+ this.costfunction.getCostConfiguration().getWeightNoMatch() + "S"
				+ this.costfunction.getCostConfiguration().getWeightGlobalStructure() + "B"
				+ this.costfunction.getCostConfiguration().getWeightBaseCostForNoMatch();
	}

	public void setCostStrategyConfiguration(CostStrategyConfiguration costStrategyConfiguration) {
		this.costStrategyConfig = costStrategyConfiguration;
		getCostStrategy().setCostConfiguration(this.costStrategyConfig);
	}

	public PTraceParameter getPTraceParameter() {
		if (this.ptraceParm == null) {
			ptraceParm = new PTraceParameter();
		}
		return this.ptraceParm;
	}

	public void setPTraceParameter(PTraceParameter pTraceParameter) {
		this.ptraceParm = pTraceParameter;
	}

	public void setFocusOnMatchingLabels(Set<String> labels) {
		this.labels = labels;
		
	}

	/**
	 * Match only these labels. 			
	 */
	public Set<String> getLabels() {
		if(labels == null || labels.size() == 0){
			labels = new THashSet<String>();
			for(XEventClass c : classes.getClasses()){
				labels.add(c.toString());
			}
			
		}
		return this.labels;
	}

}
