package org.processmining.xlu.matching.plugins.alg.impl;

import static org.processmining.xlu.matching.plugins.alg.impl.CostStrategyAbstract.STRUCTURE_MAX_FLAT;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.plugins.alg.HeuristicStrategy;

/**
 * This class implements a cost function that is independent of the mapped nodes
 * given by the mapping.
 * 
 * @author xlu
 * 
 */
public class CostHeuristicFunctionComplexImpl extends CostStrategyAbstract implements HeuristicStrategy {

	
	private Map<String, Integer> mapLabelToIndex;


	public String toString() {
		return "HeuristicFunctionComplex";
	}
	
	public void setMapLabelToIndex(Map<String, Integer> mapLabelToIndex){
		this.mapLabelToIndex = mapLabelToIndex;
	}



	public int getCost(ENode source, ENode target, MatchMapping mapping) {
		// Component I
		int costCompIMatchs = 0;
		costCompIMatchs += getCosts(source.getPredIds(), target.getPredIds(), mapping);
		costCompIMatchs += getCosts(source.getSuccIds(), target.getSuccIds(), mapping);
		costCompIMatchs += getCosts(source.getConcIds(), target.getConcIds(), mapping);

		// Component III
		int costStructure = getStructureCostSofar(source, target, mapping);
		
		return csConfig.getWeightMatched() * costCompIMatchs + csConfig.getWeightGlobalStructure() * costStructure;
	}

	private int getStructureCostSofar(ENode source, ENode target, MatchMapping mappingSoFar) {
		int cost = 0;
		for (Entry<Integer, Integer> entry : mappingSoFar.getMatchs()) {
			int distSourcePair = mappingSoFar.getSourceGraph().getDist(entry.getKey(), source.getId());
			int distTargetPair = mappingSoFar.getTargetGraph().getDist(entry.getValue(), target.getId());
			int abs = Math.abs(distSourcePair - distTargetPair);
			cost += abs <= STRUCTURE_MAX_FLAT ? abs : STRUCTURE_MAX_FLAT;
//			cost += abs;
		}
		return cost;
	}

	private int getCosts(List<Integer> sourceContext, List<Integer> targetContext, MatchMapping mapping) {
		int cost = 0;
		List<Integer> unMappedSources = new ArrayList<Integer>();
		for (Integer sourceNeighbor : sourceContext) {
			if (mapping.inDomain(sourceNeighbor)) {
				// 1.1) sourceNeighbor is in the mapping so far
				Integer targetNeighbor = mapping.getTargetNodeIdMappedTo(sourceNeighbor);
				// sourceNeighbor is mapped to targetNeighbor
				if (targetNeighbor != null && targetContext.contains(targetNeighbor)) {
					// targetNeighbor in the neighbor of target

				} else {
					// targetNeighbor NOT in  the neighbor of target
					cost++;
				}
			} else {
				unMappedSources.add(sourceNeighbor);
			}
		}
		List<Integer> unMappedTargets = new ArrayList<Integer>();
		for (Integer targetNeighbor : targetContext) {
			if (mapping.inRange(targetNeighbor)) {
				// 1.1) targetNeighbor is in the mapping so far
				Integer sourceNeighbor = mapping.getSourceNodeIdMappedTo(targetNeighbor);
				// targetNeighbor is mapped to sourceNeighbor
				if (sourceNeighbor != null && sourceContext.contains(sourceNeighbor)) {
					// sourceNeighbor in the neighbor of source

				} else {
					// sourceNeighbor NOT in  the neighbor of source
					cost++;
				}
			} else {
				unMappedTargets.add(targetNeighbor);
			}
		}
		int[] unMappedSourceArray = getLabelMultiset(unMappedSources, mapping.getSourceGraph());
		int[] unMappedTargetArray = getLabelMultiset(unMappedTargets, mapping.getTargetGraph());
		for (int i = 0; i < unMappedSourceArray.length; i++) {
			cost += Math.abs(unMappedSourceArray[i] - unMappedTargetArray[i]);
		}
		return cost;
	}

	private int[] getLabelMultiset(List<Integer> unMappedSources, EGraph g) {
		int[] array = new int[this.mapLabelToIndex.size()];
		for(Integer i : unMappedSources){
			ENode e = g.getNodeById(i);
			int index = this.mapLabelToIndex.get(e.getLabel());
			array[index]++;
		}
		return array;
	}


	public int getCostNoMatch(ENode n, MatchMapping mapping) {
		int sumCost = this.csConfig.getWeightBaseCostForNoMatch();
		sumCost += n.getPredIds().size();
		sumCost += n.getSuccIds().size();
		sumCost += n.getConcIds().size();		
		return this.csConfig.getWeightNoMatch() * sumCost;
	}

}
