package org.processmining.xlu.matching.plugins.alg.impl;

import java.util.Map.Entry;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;

public abstract class CostStrategyAbstract implements CostStrategy {
	protected static final int STRUCTURE_MAX_FLAT = Integer.MAX_VALUE;
	
	protected CostStrategyConfiguration csConfig = new CostStrategyWeightsImpl();

	public CostStrategyAbstract(){
	}
	

	protected abstract int getCost(ENode sourceNode, ENode targetNode, MatchMapping mapping);
	protected abstract int getCostNoMatch(ENode nodeById, MatchMapping mapping);

	
	public int getTotalCost(EGraph source, EGraph target, MatchMapping mapping) {
		int cost = 0;
		int compI =  getCostOfMatchedPairs(source, target, mapping);
		int compII = getCostOfNoMatched(source, target, mapping);
		int compIII = getCostOfGlobalStructure(source, target, mapping);
		
//		writer.println(compI + "," + compII + ";" + compIII);
		cost += compI + compII + compIII;
		if (cost < 0) {
			System.out.println("");
		}
		return cost;
	}

	// Component I
	private int getCostOfMatchedPairs(EGraph source, EGraph target, MatchMapping mapping) {
		int cost = 0;
		for (Entry<Integer, Integer> match : mapping.getMatchs()) {
//			ENode sourceNode = source.getNodeById(match.getKey());
//			ENode targetNode = target.getNodeById(match.getValue());
			cost += getWeightedCostofMatchedPair(match.getKey(), match.getValue(), mapping);
		}
		return cost;
	}
	
	// Component II
	private int getCostOfNoMatched(EGraph source, EGraph target, MatchMapping mapping) {
		int cost = 0;
		for (Integer sourceNodeId : mapping.getSourcesToNull()) {
			cost += getWeightedCostNoMatch(sourceNodeId, source, mapping);
		}
		for (Integer targetNodeId : mapping.getTargetToNull()) {
			cost += getWeightedCostNoMatch(targetNodeId, target, mapping);
		}
		return cost;
	}
	
	// Component III
	private int getCostOfGlobalStructure(EGraph source, EGraph target, MatchMapping mapping) {
		int cost = 0;
//		List<Entry<Integer, Integer>> entries = new ArrayList<Entry<Integer, Integer>>(mapping.getMatchs());
//		while (entries.size() > 1) {
//			Entry<Integer, Integer> entry = entries.remove(0);
		if((!source.equals(mapping.getSourceGraph())) || !target.equals(mapping.getTargetGraph())){
			System.err.println("getCostOfGlobalStructure: Wrong mapping detected.");
		}
		for( Entry<Integer, Integer> entry : mapping.getMatchs()){
			cost = getWeightedCostOfGlobalStructure(entry.getKey(), entry.getValue(), mapping );
		}
		return cost;
	}

	

	public int getWeightedCostofMatchedPair(Integer sourceId, Integer targetId,
			MatchMapping mapping) {
		ENode sourceNode = mapping.getSourceGraph().getNodeById(sourceId);
		ENode targetNode = mapping.getTargetGraph().getNodeById(targetId);		
		return csConfig.getWeightMatched() * getCost(sourceNode, targetNode, mapping);
	}
	
	public int getWeightedCostNoMatch(Integer nodeId, EGraph NodeGraph, MatchMapping mapping) {
		return csConfig.getWeightNoMatch() * getCostNoMatch(NodeGraph.getNodeById(nodeId), mapping);
	}


	public int getWeightedCostOfGlobalStructure(Integer sourceId, Integer targetId, MatchMapping mapping) {
		int cost = 0;
		EGraph source = mapping.getSourceGraph();
		EGraph target = mapping.getTargetGraph();
		for (Entry<Integer, Integer> otherEntry : mapping.getMatchs()) {
			
			int distSourcePair = source.getDist(sourceId, otherEntry.getKey());
			int distTargetPair = target.getDist(targetId, otherEntry.getValue());
			int abs = Math.abs(distSourcePair - distTargetPair);
			cost += abs <= STRUCTURE_MAX_FLAT ? abs : STRUCTURE_MAX_FLAT ;
//			cost += abs;
		}
		cost = cost / 2;
		cost = csConfig.getWeightGlobalStructure() * cost;
		return cost;
	}
	
	
	public void setCostConfiguration(CostStrategyConfiguration config) {
		this.csConfig = config;
	}

	public CostStrategyConfiguration getCostConfiguration() {
		return this.csConfig;
	}


}
