package org.processmining.xlu.matching.plugins.alg.impl;

import java.util.Comparator;
import java.util.PriorityQueue;
import java.util.Queue;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.plugins.alg.ENodeContextComparator;
import org.processmining.xlu.matching.plugins.alg.MatchingAlg;

/**
 * Alg 4. Linear computation greedy...
 * 
 * 
 * @author xlu
 * 
 */
public class MatchingAlgGreedyImp extends MatchingAlgAbstract implements MatchingAlg {

	private Comparator<? super ENode> nodeComparator;
	
	//	@Override
	//	public List<MatchMapping> getAllOptimalMapping(EGraph g1, EGraph g2) {
	//		List<MatchMapping> mappings = new ArrayList<MatchMapping>();
	//		MatchMapping m = getOptimalMapping(g1, g2);
	//		mappings.add(m);
	//		return mappings;
	//	}

	@Override
	public MatchMapping getOptimalMapping(EGraph g1, EGraph g2) {
		MatchMapping mapping = new MatchMappingImpl(g1, g2);

		Queue<ENode> sources = getSortedNodesForMatching(g1);
		Queue<ENode> targets = getSortedNodesForMatching(g2);

		matchDummyStartEndAndUpdateMapping(sources, targets, mapping);

		while (!sources.isEmpty()) {
			// 1. Get the next source to be match. 
			ENode s = sources.poll();

			// 2. Select greedy the most matching node.
			int minCost = Integer.MAX_VALUE;
			ENode tOpt = null;
			// 2.1 Match source to a target node
			for (ENode t : targets) {
				if (t.getLabel().equals(s.getLabel())) {
					int cost = computeCost(s, t, mapping);
					if (cost < minCost) {
						tOpt = t;
						minCost = cost;
					}
				}
			}

			// 2.2 Match source to nothing
			int cost = computeNoMatchCost(s, mapping);
			if (cost < minCost) {
				tOpt = null;
				minCost = cost;
			}

			// 3. Update the choice
			targets.remove(tOpt);
			mapping.addMatch(s, tOpt);
		}

		for (ENode t : targets) {
			mapping.addMatch(null, t);
		}

		int cost = getCostStrategy().getTotalCost(g1, g2, mapping);
		mapping.setCost(cost);

		return mapping;
	}

	private Queue<ENode> getSortedNodesForMatching(EGraph g) {
		int size = g.getSize();
		Queue<ENode> queue = new PriorityQueue<ENode>(size, getENodeComparator());
		queue.addAll(g.getNodes());
		return queue;
	}

	private int computeNoMatchCost(ENode s, MatchMapping mapping) {
		return getHeuristicfunction().getCostNoMatch(s, mapping);
	}

	private int computeCost(ENode s, ENode t, MatchMapping mapping) {
		return getHeuristicfunction().getCost(s, t, mapping);
	}

	public Comparator<? super ENode> getENodeComparator() {
		if (nodeComparator == null) {
			nodeComparator = new ENodeContextComparator();
		}
		return nodeComparator;
	}

	public void setNodeComparator(Comparator<? super ENode> eNodeContextComparator) {
		nodeComparator = eNodeContextComparator;
	}

}
