package org.processmining.xlu.matching.plugins.alg.impl;

import java.util.ArrayList;
import java.util.List;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.plugins.alg.MatchingAllAlg;

/**
 * Algorithm 1: brute forcely compute all (optimal) mapping.
 * 
 * @author xlu
 * 
 */
public class MatchingAllAlgBackTrackingBFImpl extends MatchingAlgAbstract implements MatchingAllAlg {
	protected boolean computeOnlyAllOptMappings = true;
	
	public boolean isComputeOnlyAllOptMappings() {
		return computeOnlyAllOptMappings;
	}

	public void setComputeOnlyAllOptMappings(boolean computeOnlyAllOptMappings) {
		this.computeOnlyAllOptMappings = computeOnlyAllOptMappings;
	}

	public List<MatchMapping> getAllOptimalMapping(EGraph g1, EGraph g2) {
		MatchMapping mapping = new MatchMappingImpl(g1, g2);
		List<MatchMapping> mappings = new ArrayList<MatchMapping>();

		List<ENode> sources = new ArrayList<ENode>(g1.getNodes());
		List<ENode> targets = new ArrayList<ENode>(g2.getNodes());

		matchDummyStartEndAndUpdateMapping(sources, targets, mapping);

		recurse(sources, targets, mapping, mappings, Integer.MAX_VALUE);

		return mappings;
	}

	public MatchMapping getOptimalMapping(EGraph g1, EGraph g2) {
		return getAllOptimalMapping(g1, g2).get(0);
	}

	protected void recurse(List<ENode> sources, List<ENode> targets, MatchMapping mapping, List<MatchMapping> mappings,
			int minCost) {

		// 1. base case
		if (sources.isEmpty() || targets.isEmpty()) {
			computeCost(sources, targets, mapping, mappings, minCost);
			return;
		}

		// 2.1 recurse : select the source node to be mapped
		ENode source = sources.remove(0);
		List<Integer> indicesMatchbleInTarget = getNodesAllowedToMap(source, targets);

		// 2.2: select the target node to be mapped. 
		for (Integer targetIndex : indicesMatchbleInTarget) {
			ENode target = targets.remove((int) targetIndex); // select choice II. 

			// 2.3 add the choice to continue with recursion
			mapping.addMatch(source, target);
			
			if (isValid(mapping)) {
				recurse(sources, targets, mapping, mappings, minCost);
			}

			mapping.removeMatch(source, target);
			targets.add(targetIndex, target); // add element back to original index.
		}

		// 2.4: choose not to map the target node. 
		mapping.addMatch(source, null);
		if (isValid(mapping)) {
			recurse(sources, targets, mapping, mappings, minCost);
		}

		// undo choice. 
		mapping.removeMatch(source, null);
		sources.add(0, source);
	}

	private void computeCost(List<ENode> sources, List<ENode> targets, MatchMapping mapping,
			List<MatchMapping> mappings, int minCost) {
		// Complete choices
		for (ENode n : sources) {
			mapping.addMatch(n, null);
		}
		for (ENode n : targets) {
			mapping.addMatch(null, n);
		}

		// Compute cost and update optimal mapping
		int cost = this.getCostStrategy().getTotalCost(mapping.getSourceGraph(), mapping.getTargetGraph(), mapping);
		if (computeOnlyAllOptMappings) {
			addMappingOnlyIfOpt(mappings, mapping, cost);
		} else {
			cloneMappingAndAddTo(mappings, mapping, cost);
		}

		// remove choices
		for (ENode n : sources) {
			mapping.removeMatch(n, null);
		}
		for (ENode n : targets) {
			mapping.removeMatch(null, n);
		}

	}

	
	

}
