package org.processmining.xlu.matching.preprocessing;

import gnu.trove.map.hash.THashMap;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XEvent;
import org.processmining.partialorder.models.dependency.PDependency;
import org.processmining.partialorder.ptrace.model.PTrace;
import org.processmining.xlu.matching.models.Context;
import org.processmining.xlu.matching.models.ContextMultiSetPCSDistinct;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.impl.ENodeProperties;

public class PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl extends PTraceToEGConverterAbstract {

	Map<Context, Integer> mapContext2freq = new THashMap<Context, Integer>();
//	Map<String, Integer> mapLabel2NumContext = new THashMap<String, Integer>();
	
	public PTraceToEGConverterCAwareAdaptiveKwithDummyStartEndImpl(int distContext, XEventClasses classes) {
		super(distContext, classes);
	}
	
	public Map<Context, Integer> getMapContextToFreq() {
		return this.mapContext2freq;
	}

	public EGraph convert(PTrace ptrace) {
//		XEventClasses classes = this.getClasses();
		EGraphFactory factory = EGraphFactory.getInstance();
		EGraph eg = factory.createEmptyExecutionGraph(ptrace);

		int size = ptrace.getTrace().size();
		Map<Integer, List<Integer>> predMap = new HashMap<Integer, List<Integer>>(size);
		Map<Integer, List<Integer>> succMap = new HashMap<Integer, List<Integer>>(size);
		Map<Integer, List<Integer>> concMap = new HashMap<Integer, List<Integer>>(size);

		int[][] distMatrix = computeNodeContexts(eg, ptrace, predMap, succMap, concMap);
		factory.setGraphDist(eg, distMatrix);

		for (int i = 0; i < size; i++) {
			
			XEvent e = ptrace.getEvent(i);
			XEventClass c = getClass(e);
			int[] preds = convertPredToArray(ptrace, predMap.get(i));
			int[] succs = convertSuccToArray(ptrace, succMap.get(i)); // Dist = 1 (i.e. direct) preds. 
			int[] concs = convertToArray(ptrace, concMap.get(i));
			Context context = new ContextMultiSetPCSDistinct(c.toString(), preds, succs, concs);
			Integer freq = this.mapContext2freq.get(context);
			this.mapContext2freq.put(context, freq != null ? ++freq : 1 );
			
			ENode node = factory.createNode(eg, c, predMap.get(i), succMap.get(i), concMap.get(i), preds, succs, concs);
			node.setProperty(ENodeProperties.KEY_CONTEXT, context);
		}

		for (PDependency dep : ptrace.getDependencies()) {
			if (dep.isDirect()) {
				EEdge e = factory.createEdge(eg, dep.getSource(), dep.getTarget());
			}
		}
//		if( eg.getEdges().size() == ptrace.getDependencies().size()){
//			System.out.println("Trace " + eg.getId());
//		}

		addArtificalStartAndEnd(eg, ptrace, factory);

		return eg;
	}





}
