package org.processmining.xlu.matching.vis;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.math3.util.Pair;
import org.processmining.partialorder.models.graph.POEdge;
import org.processmining.partialorder.models.graph.PONode;
import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.partialorder.models.graph.edge.POEdgeImp;
import org.processmining.partialorder.plugins.vis.PartialOrderGraphFactory;
import org.processmining.partialorder.plugins.vis.PartialVisualType;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.partialorder.comparison.vis.PGraphInfoViewComparableGImp;
import org.processmining.xlu.pgraph.vis.PGraphParameter;
import org.processmining.xlu.pgraph.vis.VisSettingUtil;
import org.processmining.xlu.pgraph.vis.converter.PGraphConverter;
import org.processmining.xlu.pgraph.vis.panel.PGraphInfoView;
import org.processmining.xlu.pgraph.vis.panel.PGraphVisView;

import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;
import edu.uci.ics.jung.graph.DirectedSparseGraph;

public class EGraphToPGraphConverterBinaryImpl implements PGraphConverter<RepEGraph> {
	Map<PartialOrderGraph, Map<Object, Map<String, String>>> mapGraphToInfo = new HashMap<>();

	

	public PartialOrderGraph convert(RepEGraph object, PGraphParameter parameter) {
		PartialOrderGraph visgraph = new PartialOrderGraph("", object.getId());
		Map<Integer, PONode> map = new HashMap<>();
		Map<Object, Map<String, String>> mapElemToInfo = new HashMap<>();

		EGraphFactory factory = EGraphFactory.getInstance();
		List<Integer> graphIds = object.getRepresentedGraphIds();
		if (graphIds.size() > 2) {
			System.err.println("Binary converter encounted non binary RepEGraph.");
			return null;
		}

		for (ENode n : object.getNodes()) {
			Map<Integer, Integer> orgGraph2Node = object.getNodeRepresentedNodes().get(n.getId());
			PONode visnode = null;
			if (factory.isArtifical(n)) {
				visnode = new PODummyNode(visgraph, n.getLabel());
			} else {
				visnode = new POMatchNode(visgraph, n.getLabel(), n);
			}
			visgraph.addNode(visnode);
			Boolean correctness = object.getNodeToMappedCorrectly().get(n.getId());
			if (correctness == null || correctness) {
				visnode.setColor(getColor(orgGraph2Node, graphIds, VisSettingUtil.threeLightColor));
			} else {
				visnode.setColor(new Color(189, 0, 38));
			}
			Map<String, String> infoObject = getInfoForNode(orgGraph2Node);
			updateNodeInfo(n, object, infoObject);
			mapElemToInfo.put(visnode, infoObject);

			map.put(n.getId(), visnode); // book keeping

		}
		// add edges=================
		if (parameter.getVisType().equals(PartialVisualType.NO_PARALLEL_MINIMAL)) {
			List<Pair<Integer, Integer>> concurPairs = new ArrayList<Pair<Integer, Integer>>();
			DirectedSparseGraph<Integer, Integer> graph = getGraph(object);
			UnweightedShortestPath<Integer, Integer> dist = new UnweightedShortestPath<Integer, Integer>(graph);

			for (int i = 0; i < object.getNodes().size(); i++) {
				for (int j = i + 1; j < object.getNodes().size(); j++) {
					Number distI2J = dist.getDistance(i, j);
					Number distJ2I = dist.getDistance(j, i);

					if (distI2J != null && distJ2I != null) {
						Pair<Integer, Integer> pair = new Pair<Integer, Integer>(i, j);
						concurPairs.add(pair);
					} else if (distI2J == null && distJ2I != null && distJ2I.intValue() == 1) {
						EEdge e = object.getEdge(j, i);
						addEdge(visgraph, e, object, map, mapElemToInfo);
					} else if (distI2J != null && distJ2I == null && distI2J.intValue() == 1) {
						EEdge e = object.getEdge(i, j);
						addEdge(visgraph, e, object, map, mapElemToInfo);
					}
				}
			}
			for (Pair<Integer, Integer> pair : concurPairs) {
				PONode s = map.get(pair.getFirst());
				PONode t = map.get(pair.getSecond());
				updatePredAndSucc(visgraph, s, t);
				updatePredAndSucc(visgraph, t, s);
			}
			PartialOrderGraphFactory.removeRedundantEdges(visgraph);

		} else {
			for (EEdge e : object.getEdges()) {

				Map<Integer, Integer> orgGraph2Edge = object.getEdgeRepresentedEdges().get(e.getId());
				POEdge visedge = new POEdgeImp(map.get(e.getSourceId()), map.get(e.getTargetId()));
				visgraph.addEdge(visedge);
				visedge.setWidth(getWidth(orgGraph2Edge));
				visedge.setColor(getColor(orgGraph2Edge, graphIds, VisSettingUtil.threeDarkColor));
				Map<String, String> infoObject = getInfoForNode(orgGraph2Edge);

				mapElemToInfo.put(visedge, infoObject);
			}
		}
		mapGraphToInfo.put(visgraph, mapElemToInfo);

		return visgraph;
	}
	public PartialOrderGraph convert(RepEGraph object) {
		PGraphParameter p = new PGraphParameter();
		return convert(object, p);
	}
	
	
	// TODO TO REMOVE Feb 11. redundant ======================================================
	private void updatePredAndSucc(PartialOrderGraph visgraph, PONode s, PONode t) {
		for (POEdge inEdge : visgraph.getInEdges(s)) {
			POEdge visedge = new POEdgeImp(inEdge.getSource(), t);
			visgraph.addEdge(visedge);
		}
		for (POEdge outEdge : visgraph.getOutEdges(s)) {
			POEdge visedge = new POEdgeImp(t, outEdge.getTarget());
			visgraph.addEdge(visedge);
		}

	}

	private void addEdge(PartialOrderGraph visgraph, EEdge e, RepEGraph object, Map<Integer, PONode> map,
			Map<Object, Map<String, String>> mapElemToInfo) {
//		
//		Set<Integer> graphIds = object.getEdgeRepresentedEdges().get(e.getId()).keySet();
		Map<Integer, Integer> orgGraph2Edge = object.getEdgeRepresentedEdges().get(e.getId());
		
		POEdge visedge = new POEdgeImp(map.get(e.getSourceId()), map.get(e.getTargetId()));
		visgraph.addEdge(visedge);
		
		Map<String, String> infoObject = getInfoForNode(orgGraph2Edge);
		mapElemToInfo.put(visedge, infoObject);

	}

	private DirectedSparseGraph<Integer, Integer> getGraph(RepEGraph object) {
		DirectedSparseGraph<Integer, Integer> graph = new DirectedSparseGraph<Integer, Integer>();
		for (ENode n : object.getNodes()) {
			graph.addVertex(n.getId());
		}
		for (EEdge e : object.getEdges()) {
			graph.addEdge(e.getId(), e.getSourceId(), e.getTargetId());
		}

		return graph;
	}
	// END to remove ===========================================

	//TODO: added adhoc @Jan 24, 2015. 
	private void updateNodeInfo(ENode n, RepEGraph object, Map<String, String> infoObject) {
		int id = n.getId();
		Integer cost = object.getNodeToCost().get(id);
		Boolean t = object.getNodeToMappedCorrectly().get(id);
		infoObject.put("Cost", String.valueOf(cost));
		if (t != null) {
			infoObject.put("Correctness", String.valueOf(t));
		}
	}

	private Map<String, String> getInfoForNode(Map<Integer, Integer> orgGraph2Node) {
		Map<String, String> info = new HashMap<>();

		for (Entry<Integer, Integer> entry : orgGraph2Node.entrySet()) {
			Integer orgGraphId = entry.getKey();
			Integer orgNodeId = entry.getValue();
			String key = "GID " + orgGraphId;
			String value = "NID " + orgNodeId;
			info.put(key, value);
		}
		return info;
	}

	private float getWidth(Map<Integer, Integer> orgGraph2Edge) {
		return VisSettingUtil.widths[orgGraph2Edge.size() + 1];
	}

	private Color getColor(Map<Integer, Integer> orgGraph2Node, List<Integer> graphIds, Color[] colors) {
		if (orgGraph2Node.size() == 2) {
			return colors[2];
		} else {
			Entry<Integer, Integer> entry = orgGraph2Node.entrySet().iterator().next(); //just get the one
			return colors[graphIds.indexOf(entry.getKey())];
		}
	}


	public PGraphInfoView<RepEGraph> createPGraphInfoView(PartialOrderGraph g, RepEGraph object) {
		return new PGraphInfoViewComparableGImp(g, object, mapGraphToInfo.get(g));
	}

	public PGraphVisView<RepEGraph> createPGraphVisView(PartialOrderGraph g, PGraphInfoView<RepEGraph> infoView) {
		return new PGraphVisView<RepEGraph>(g, infoView);
	}

}
