package org.processmining.xlu.matching.vis;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.math3.util.Pair;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.partialorder.models.graph.POEdge;
import org.processmining.partialorder.models.graph.PONode;
import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.partialorder.models.graph.edge.POEdgeImp;
import org.processmining.partialorder.plugins.vis.PartialOrderGraphFactory;
import org.processmining.partialorder.plugins.vis.PartialVisualType;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.RepEGraphUtil;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.partialorder.comparison.vis.PGraphInfoViewComparableGImp;
import org.processmining.xlu.pgraph.vis.PGraphParameter;
import org.processmining.xlu.pgraph.vis.VisSettingUtil;
import org.processmining.xlu.pgraph.vis.converter.PGraphConverter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterEventFreqImpl;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterLabelAEventFreqImpl;
import org.processmining.xlu.pgraph.vis.panel.PGraphInfoView;
import org.processmining.xlu.pgraph.vis.panel.PGraphVisView;

import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;
import edu.uci.ics.jung.graph.DirectedSparseGraph;

public class EGraphToPGraphConverterImpl implements PGraphConverter<RepEGraph> {

	Map<PartialOrderGraph, Map<Object, Map<String, String>>> mapGraphToInfo = new HashMap<>();
	XLog log;
	private XEventClasses classes;

	int[] colormap = new int[5];
	int[] widthmap = new int[5];

	public EGraphToPGraphConverterImpl(XLog log, XEventClasses xClasses, RepEGraphs objects) {
		this.log = log;
		this.classes = xClasses;

		List<Integer> numberTracesPerNode = new ArrayList<Integer>();
		List<Integer> numberTracesPerEdge = new ArrayList<Integer>();
		for (RepEGraph graph : objects) {
			numberTracesPerEdge.addAll(RepEGraphUtil.getListOfNumTracesOfEdges(graph));
			numberTracesPerNode.addAll(RepEGraphUtil.getListOfNumTracesOfNodes(graph));
		}
		Collections.sort(numberTracesPerEdge);
		Collections.sort(numberTracesPerNode);
		for (int i = 0; i < 5; i++) {
			colormap[i] = numberTracesPerNode.get(numberTracesPerNode.size() / 5 * (i + 1) - 1);
			widthmap[i] = numberTracesPerEdge.get(numberTracesPerEdge.size() / 5 * (i + 1) - 1);
		}
	}

	public Color getNodeColor(int numTraces) {
		Color c = new Color(255, 255, 255);
		for (int i = 0; i < 5; i++) {
			if (numTraces <= colormap[i]) {
				c = VisSettingUtil.fiveGreenColors[i];
				break;
			}
		}
		return c;
	}

	public float getEdgeWidth(int numTraces) {
		float w = (float) 0.1;
		for (int i = 0; i < 5; i++) {
			if (numTraces <= widthmap[i]) {
				w = VisSettingUtil.widths[i];
				break;
			}
		}
		return w;
	}

	public PartialOrderGraph convert(RepEGraph object, PGraphParameter parameter) {
		//TODO
		Map<XEventClass, Color> mapEvtClass2Color = parameter.getEvtColorMap();
		Map<Integer, Color> mapTr2Color = parameter.getTraceColorMap();

		PartialOrderGraph visgraph = new PartialOrderGraph("", -1);
		Map<Integer, PONode> map = new HashMap<>();
		Map<Object, Map<String, String>> mapElemToInfo = new HashMap<>();

		//		List<ENode> nodes = new ArrayList<ENode>(object.getNodes());
		//		for(ENode n : nodes){
		//			if(!parameter.getNodeFilter().isSatisfied(n)){
		//				object.removeNode(n);
		//			}
		//		}
		Set<PONode> mapNode2Vis = new HashSet<PONode>();
		for (ENode n : object.getNodes()) {
			// TODO fix
			Map<Integer, Integer> orgGraph2Node = object.getNodeRepresentedNodes().get(n.getId());

			int size = orgGraph2Node.size();
			PONode visnode = null;
			if (EGraphFactory.getInstance().isArtifical(n)) {
				visnode = new PODummyNode(visgraph, n.getLabel());
			} else {
				visnode = new POMatchNode(visgraph, n.getLabel(), n);
			}
			visgraph.addNode(visnode);
			// get color 

			visnode.setColor(getNodeColor(size));
			Color color = null; //mapEvtClass2Color.get(classes.getClassOf(e));
			//			if (color != null) {
			//				visnode.setColor(color);
			//			}
			for (Integer tr : orgGraph2Node.keySet()) {
				color = mapTr2Color.get(tr);
				if (color != null) {
					visnode.setColor(color);
				}
			}

			if (!parameter.getNodeFilter().isSatisfied(n)) {
				visnode.setColor(VisSettingUtil.greyOutColor);
				if(visnode instanceof POMatchNode){
				((POMatchNode)visnode).setNodeLineColor(VisSettingUtil.greyOutColorLine);
				((POMatchNode)visnode).setNodeLabelColor(VisSettingUtil.greyOutColorText);
				}
				mapNode2Vis.add(visnode);
			}

			// book keeping
			map.put(n.getId(), visnode);

			Map<String, String> infoObject = getInfoForNode(orgGraph2Node);
			mapElemToInfo.put(visnode, infoObject);

		}

		if (parameter.getVisType().equals(PartialVisualType.NO_PARALLEL_MINIMAL)) {
			List<Pair<Integer, Integer>> concurPairs = new ArrayList<Pair<Integer, Integer>>();
			DirectedSparseGraph<Integer, Integer> graph = getGraph(object);
			UnweightedShortestPath<Integer, Integer> dist = new UnweightedShortestPath<Integer, Integer>(graph);

			for (int i = 0; i < object.getNodes().size(); i++) {
				for (int j = i + 1; j < object.getNodes().size(); j++) {
					Number distI2J = dist.getDistance(i, j);
					Number distJ2I = dist.getDistance(j, i);

					if (distI2J != null && distJ2I != null) {
						Pair<Integer, Integer> pair = new Pair<Integer, Integer>(i, j);
						concurPairs.add(pair);
					} else if (distI2J == null && distJ2I != null && distJ2I.intValue() == 1) {
						EEdge e = object.getEdge(j, i);
						addEdge(visgraph, e, object, map, mapTr2Color, mapElemToInfo);
					} else if (distI2J != null && distJ2I == null && distI2J.intValue() == 1) {
						EEdge e = object.getEdge(i, j);
						addEdge(visgraph, e, object, map, mapTr2Color, mapElemToInfo);
					}
				}
			}
			for (Pair<Integer, Integer> pair : concurPairs) {
				PONode s = map.get(pair.getFirst());
				PONode t = map.get(pair.getSecond());
				updatePredAndSucc(visgraph, s, t);
				updatePredAndSucc(visgraph, t, s);
			}
			PartialOrderGraphFactory.removeRedundantEdges(visgraph);

		} else {
			for (EEdge e : object.getEdges()) {

				//			if (parameter.getVisType().equals(PartialVisualType.NO_PARALLEL_MINIMAL)) {
				//				if (object.getEdge(e.getTargetId(), e.getSourceId()) != null) {
				//					continue;
				//				}
				//			}

				addEdge(visgraph, e, object, map, mapTr2Color, mapElemToInfo);
			}

			mapGraphToInfo.put(visgraph, mapElemToInfo);
		}

		boolean doRemove = false;
		if(parameter.getNodeFilter() instanceof PGraphFilterEventFreqImpl){
			doRemove =  ((PGraphFilterEventFreqImpl)parameter.getNodeFilter()).doRemoveNodes();
		}
		if(parameter.getNodeFilter()  instanceof PGraphFilterLabelAEventFreqImpl){
			doRemove =  ((PGraphFilterLabelAEventFreqImpl)parameter.getNodeFilter()).doRemoveNodes();
		}
		if (doRemove) {
			for (PONode n : mapNode2Vis) {
				Set<POEdge> toRemove = new HashSet<POEdge>();
				for (POEdge in : visgraph.getInEdges(n)) {
					toRemove.add(in);
					for (POEdge out : visgraph.getOutEdges(n)) {
						toRemove.add(out);
						if (!in.getSource().equals(out.getTarget())) {
							POEdge newEdge = new POEdgeImp(in.getSource(), out.getTarget());
							visgraph.addEdge(newEdge);
						}
					}
				}
				visgraph.removeEdges(toRemove);
				visgraph.removeNode(n);
			}
			PartialOrderGraphFactory.removeRedundantEdges(visgraph);
		}

		return visgraph;
	}

	private void updatePredAndSucc(PartialOrderGraph visgraph, PONode s, PONode t) {
		for (POEdge inEdge : visgraph.getInEdges(s)) {
			POEdge visedge = new POEdgeImp(inEdge.getSource(), t);
			visgraph.addEdge(visedge);
		}
		for (POEdge outEdge : visgraph.getOutEdges(s)) {
			POEdge visedge = new POEdgeImp(t, outEdge.getTarget());
			visgraph.addEdge(visedge);
		}

	}

	private void addEdge(PartialOrderGraph visgraph, EEdge e, RepEGraph object, Map<Integer, PONode> map,
			Map<Integer, Color> mapTr2Color, Map<Object, Map<String, String>> mapElemToInfo) {
		Map<Integer, Integer> mapGraph2Edge = object.getEdgeRepresentedEdges().get(e.getId());
		if (mapGraph2Edge == null) {
			mapGraph2Edge = new HashMap<Integer, Integer>();
		}
		Set<Integer> graphIds = mapGraph2Edge.keySet();
		float width = getEdgeWidth(graphIds.size());
		POEdge visedge = new POEdgeImp(map.get(e.getSourceId()), map.get(e.getTargetId()), width);
		visgraph.addEdge(visedge);
		for (Integer tr : graphIds) {
			Color color = mapTr2Color.get(tr);
			if (color != null) {
				visedge.setColor(color);
			}
		}
		Map<String, String> infoObject = getInfoForNode(graphIds);
		mapElemToInfo.put(visedge, infoObject);

	}

	private DirectedSparseGraph<Integer, Integer> getGraph(RepEGraph object) {
		DirectedSparseGraph<Integer, Integer> graph = new DirectedSparseGraph<Integer, Integer>();
		for (ENode n : object.getNodes()) {
			graph.addVertex(n.getId());
		}
		for (EEdge e : object.getEdges()) {
			graph.addEdge(e.getId(), e.getSourceId(), e.getTargetId());
		}

		return graph;
	}

	public PartialOrderGraph convert(RepEGraph object) {
		PGraphParameter p = new PGraphParameter();
		return convert(object, p);
	}

	private Map<String, String> getInfoForNode(Collection<Integer> graphIds) {
		Map<String, String> info = new HashMap<>();

		for (Integer id : graphIds) {
			String key = "GID " + id;
			info.put(key, id.toString());
		}
		return info;
	}

	private Map<String, String> getInfoForNode(Map<Integer, Integer> traceToEventOfNode) {
		Map<String, String> info = new HashMap<>();

		for (Entry<Integer, Integer> entry : traceToEventOfNode.entrySet()) {
			Integer orgGraphId = entry.getKey();
			Integer orgNodeId = entry.getValue();
			String key = "GID " + orgGraphId;
			String value = "NID " + orgNodeId;
			info.put(key, value);
		}
		return info;
	}

	public PGraphInfoView<RepEGraph> createPGraphInfoView(PartialOrderGraph g, RepEGraph object) {
		return new PGraphInfoViewComparableGImp(g, object, mapGraphToInfo.get(g));
	}

	public PGraphVisView<RepEGraph> createPGraphVisView(PartialOrderGraph g, PGraphInfoView<RepEGraph> infoView) {
		return new PGraphVisView<RepEGraph>(g, infoView);
	}
	
	

}
