package org.processmining.xlu.matching.vis;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.swing.JComponent;

import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.collection.REGraphType;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.models.collection.RepEGraphsUtil;
import org.processmining.xlu.pgraph.vis.PGraphController;
import org.processmining.xlu.pgraph.vis.converter.PGraphConverter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterEventFreqImpl;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterHierarchyImpl;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterSourceIdImpl;
import org.processmining.xlu.pgraph.vis.panel.PGraphListView;
import org.processmining.xlu.pgraph.vis.panel.PGraphMainView;

public class REGraphVisualizer {

	@Visualizer
	@PluginVariant(requiredParameterLabels = { 0 })
	@Plugin(name = "Visualize REGs", returnLabels = { "PTrace Clusters" },
			returnTypes = { JComponent.class }, parameterLabels = { "PO Traces" }, userAccessible = true)
	public JComponent visualize(PluginContext context, RepEGraphs graphs) {

		/* ----------------------------------------------------------------
		 * 1. Create converter
		 * ---------------------------------------------------------------- */
		PGraphConverter<RepEGraph> converter = null;
		REGraphType type = graphs.getType();
		if (type != null && type.equals(REGraphType.HIERARCHICAL)) {
			converter = new EGraphToPGraphConverterImpl(graphs.getLog(), graphs.getClasses(),
					graphs);
		} else {			
			converter = new EGraphToPGraphConverterBinaryImpl();
		}
		
		
		/* ----------------------------------------------------------------
		 * 2. Create controller 
		 * ---------------------------------------------------------------- */		 
		PGraphController<RepEGraph> controller = new PGraphController<RepEGraph>(converter, context, graphs.getLog());

		
		/* ----------------------------------------------------------------
		 * 3. Create filters
		 * ---------------------------------------------------------------- */	 
		
		List<PGraphFilter<RepEGraph>> filters = new ArrayList<PGraphFilter<RepEGraph>>();
		//filters.add(new PGraphFilterTraceId(controller, graphs.getInstanceGraphIds()));
		
		
//		filters.add(new PGraphFilterMaxNumber<RepEGraph>(controller));
		
		if (graphs.getHierarchy() != null) {
			filters.add(new PGraphFilterHierarchyImpl(controller, graphs.getHierarchy(), graphs.getLog()));
		} else {
			filters.add(new PGraphFilterSourceIdImpl(controller, RepEGraphsUtil.getIds(graphs)));
		}
		PGraphFilter<ENode> filter = new PGraphFilterEventFreqImpl(controller, graphs);
//		PGraphFilter<ENode> filter = new PGraphFilterLabelAEventFreqImpl(controller, graphs);
		
		controller.setFilters(filters);
		controller.addNodeFiltering(filter); //TODO REMOVE

		// 4. Create Coloring 
//		PGraphColoring<XEventClass> evtColoring = new PGraphColoringAbs<XEventClass>("EventClass", graphs.getClasses()
//				.getClasses(), controller);
//		controller.addColoring(evtColoring);
//		PGraphColoring<Integer> traceColoring = new PGraphColoringAbs<Integer>("Trace", graphs.getInstanceGraphIds(),
//				controller);
//		controller.addTrColoring(traceColoring);

		// TODO: here ad-hoc sorting... add better sorting. 
		List<RepEGraph> graphList = new ArrayList<RepEGraph>(graphs.getAllGraphs());
		Collections.sort(graphList, new Comparator<RepEGraph>() {
			public int compare(RepEGraph o1, RepEGraph o2) {
				return o1.getCost("Config 1") - o2.getCost("Config 1");
			}
		});

		// 5. Create list view and main view. 
		PGraphListView<RepEGraph> listView = new PGraphListView<RepEGraph>(graphList, controller);
		controller.setListView(listView);
		PGraphMainView main = new PGraphMainView(context, listView,
				controller.getControlView());

		return main;
	}

	

}
