package org.processmining.xlu.matching.zexperiment;

import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.partialorder.ptrace.param.PTraceParameter.PTraceType;
import org.processmining.xlu.experiment.alignment.XExperimentNoiseExtension;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.zexperiment.model.QualityTFPN;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterEventFreqImpl;

public class EGFusionTest {

	
	public static QualityTFPN computeFusionQualityNoise(XLog log, Tree<RepEGraph> hierarchy,
			PGraphFilter<ENode> eventFilter) {
		QualityTFPN data = new QualityTFPN();

		for (int i = 0; i < hierarchy.getChildCount(); i++) {
			RepEGraph reg = ((Tree<RepEGraph>) hierarchy.getChildAt(i)).getObject();
			
			for (ENode n : reg.getNodes()) {
				Map<Integer, Integer> mapG2E = reg.getNodeRepresentedNodes().get(n.getId());
				if(EGraphFactory.getInstance().isArtifical(n)){
					continue;
				}	
				if (eventFilter.isSatisfied(n)) {
					// retain node
					for (Entry<Integer, Integer> entry : mapG2E.entrySet()) {
						XTrace trace = log.get(entry.getKey());
//						if (entry.getValue() < trace.size()) { // else start or end 
							XEvent e = trace.get(entry.getValue());
							if (XExperimentNoiseExtension.isNoise(e)) {
								// Noise is retained as an event
								data.falseNegativeNoise++;
							} else {
								// An Event is truly retained as an event
								data.trueNegativeNoise++;
							}
//						}
					}
				} else { // remove node
					for (Entry<Integer, Integer> entry : mapG2E.entrySet()) {
						XTrace trace = log.get(entry.getKey());
//						if (entry.getValue() < trace.size()) { // else start or end
							XEvent e = trace.get(entry.getValue());
							if (XExperimentNoiseExtension.isNoise(e)) {
								data.truePositiveNoise++; // Noise is correctly removed as an Noise
							} else {
								data.falsePositiveNoise++; // Event is wrongly removed as a noise
							}
//						}
					}
				}
			}
		}
		if(data.trueNegativeNoise == 0){
			System.err.println("Event retained is null");
		}
		return data;
	}

	public static QualityTFPN computeFusionQualityNoise(XLog log, Tree<RepEGraph> hierarchy, int threshold) {
		QualityTFPN data = new QualityTFPN();

		for (int i = 0; i < hierarchy.getChildCount(); i++) {
			RepEGraph reg = ((Tree<RepEGraph>) hierarchy.getChildAt(i)).getObject();
			int maxFreq = PGraphFilterEventFreqImpl.computeMaxFreq(reg);
			if (maxFreq == 0) {
				System.out.println("MaxFreq == 0");
			}
			for (ENode n : reg.getNodes()) {
				Map<Integer, Integer> mapG2E = reg.getNodeRepresentedNodes().get(n.getId());
				int freq = mapG2E.size();
				if (freq * 100 / maxFreq >= threshold) {
					// retain node
					for (Entry<Integer, Integer> entry : mapG2E.entrySet()) {
						XTrace trace = log.get(entry.getKey());
						if (entry.getValue() < trace.size()) { // else start or end 
							XEvent e = trace.get(entry.getValue());
							if (XExperimentNoiseExtension.isNoise(e)) {
								// Noise is retained as an event
								data.falseNegativeNoise++;
							} else {
								// An Event is truly retained as an event
								data.trueNegativeNoise++;
							}
						}
					}
				} else { // remove node
					for (Entry<Integer, Integer> entry : mapG2E.entrySet()) {
						XTrace trace = log.get(entry.getKey());
						if (entry.getValue() < trace.size()) { // else start or end
							XEvent e = trace.get(entry.getValue());
							if (XExperimentNoiseExtension.isNoise(e)) {
								data.truePositiveNoise++; // Noise is correctly removed as an Noise
							} else {
								data.falsePositiveNoise++; // Event is wrongly removed as a noise
							}
						}
					}
				}
			}
		}
		if(data.trueNegativeNoise == 0){
			System.err.println("Event retained is null");
		}
		return data;
	}

	public static String[] printColumnNames() {
		String[] record = new String[] { 
				"NumNois", "NumRepeat",
				
				"PTraceType", "KNeighbors", //TODO concurrent true or false
//				"CostFunction", "CostFunctionConfig",

				"WMatched", "WNoMatch", "WGlobal", "WBaseCostNoMatch",

				"MatchingAlg", 
				"Filter", 
				"Threshold",

				"TP", "FP", "FN", "TN", "Fitness", "Precision"
		//, "Generalisation"
		};
		return record;
	}

	public static String[] printResult(PTraceType type, EGMatchConfiguration conf, int threshold,
			QualityTFPN data, ModelQuality quality, String filter, int numRepeat, int numNoise) {
		String[] record = new String[17];
		int i =0;
		record[i++] = String.valueOf(numNoise);
		record[i++] = String.valueOf(numRepeat);
		record[i++] = type.toString();
		record[i++] = String.valueOf(conf.getDist());
//		record[i++] = conf.getCostStrategy().toString();
//		record[i++] = conf.getCostStrategy().getCostConfiguration().getName();

		record[i++] = String.valueOf(conf.getCostStrategy().getCostConfiguration().getWeightMatched());
		record[i++] = String.valueOf(conf.getCostStrategy().getCostConfiguration().getWeightNoMatch());
		record[i++] = String.valueOf(conf.getCostStrategy().getCostConfiguration().getWeightGlobalStructure());
		record[i++] = String.valueOf(conf.getCostStrategy().getCostConfiguration().getWeightBaseCostForNoMatch());

		record[i++] = conf.getMatchingAlgorithm().toString();
		record[i++] = String.valueOf(filter);
		record[i++] = String.valueOf(threshold);

		record[i++] = String.valueOf(data.truePositiveNoise);
		record[i++] = String.valueOf(data.falsePositiveNoise);
		record[i++] = String.valueOf(data.falseNegativeNoise);
		record[i++] = String.valueOf(data.trueNegativeNoise);

		record[i++] = String.valueOf(quality.fitness);
		record[i++] = String.valueOf(quality.precision);
		//		record[17] = String.valueOf(quality.generalisation);

		return record;
	}

	public static void printSystemResult(EGMatchConfiguration conf, int threshold, QualityTFPN data) {
		System.out.println("New Result ==================");
		System.out.println("Threshold : " + threshold);
		//		System.out.println("PTraceType " + type.toString());
		//		System.out.println("KNeighbors " + conf.getDist());
		//		System.out.println("CostFunction " + conf.getCostStrategy().toString());
		//		System.out.println("CostFunctionConfig " + conf.getCostStrategy().getCostConfiguration().getName());
		//		System.out.println("MatchingAlg " + conf.getMatchingAlgorithm().toString());
		System.out.println("TP " + data.truePositiveNoise);
		System.out.println("FP " + data.falsePositiveNoise);
		System.out.println("FN " + data.falseNegativeNoise);
		System.out.println("TN " + data.trueNegativeNoise);
		//		System.out.println("FN " + );
	}

	
}
