package org.processmining.xlu.matching.zexperiment;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.in.XesXmlParser;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IMPetriNet;
import org.processmining.plugins.astar.petrinet.PetrinetReplayerWithILP;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.multietc.plugins.MultiETCPlugin;
import org.processmining.plugins.multietc.res.MultiETCResult;
import org.processmining.plugins.multietc.sett.MultiETCSettings;
import org.processmining.plugins.petrinet.replayer.PNLogReplayer;
import org.processmining.plugins.petrinet.replayer.algorithms.IPNReplayAlgorithm;
import org.processmining.plugins.petrinet.replayer.algorithms.costbasedcomplete.CostBasedCompleteParam;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;
import org.processmining.pnetreplayer.utils.TransEvClassMappingUtils;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.ExpModelSimplicityUtil;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.matching.zexperiment.precision.EventToEnabledSets;
import org.processmining.xlu.matching.zexperiment.precision.MapTrcEnabled;
import org.processmining.xlu.relabel.experiment.ExpDuplicatedTaskUtil;
import org.processmining.xlu.relabel.experiment.ExpImprLabelSetI;
import org.processmining.xlu.relabel.experiment.ExpRelabelingParam;
import org.processmining.xlu.relabel.plugin.hyber.RelabelHyberApproachIM;

import nl.tue.astar.AStarException;

/**
 * 
 * This plugin test the EGMatchPlugin using randomly generated noise
 * 
 * @author xlu
 * 
 */
@Plugin(name = "Trace Discovery Expriment", returnLabels = { "Results" }, returnTypes = { ModelQuality.class }, parameterLabels = {
		"Event Log", "Parameters" }, categories = { PluginCategory.Discovery,
		PluginCategory.Analytics }, keywords = { "Experiment" }, help = "Execute an experiment", userAccessible = true, level = PluginLevel.Local)
public class EGLogDiscoveryTestPluginImpl {

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Trace Discovery Expriment, no dup", requiredParameterLabels = { 0 })
	public ModelQuality test(PluginContext context, XLog origLog) throws IOException {
		return runDefaultTest(context, origLog);
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Trace Discovery Expriment, dup", requiredParameterLabels = { 0 })
	public ModelQuality testDup(PluginContext context, XLog origLog) throws IOException {
		XEventClasses classes = XEventClasses.deriveEventClasses(
				new XEventNameClassifier(), origLog);
		ExpImprLabelSetI<String> dupMapping = (new ExpRelabelingParam()).getDuplicationMap(classes);
		ExpDuplicatedTaskUtil.print(dupMapping);
		XLog dupLog = ExpDuplicatedTaskUtil.addDuplicatedLabels(origLog, classes,
				dupMapping);

		return runDefaultTest(context, dupLog);
	}

	public static ModelQuality runTest(PluginContext context, XLog origLog) {
		XEventClassifier classifier = new XEventNameClassifier();

		// Load filtered logs
		List<XLog> logs = LoadLogs();

		for (XLog log : logs) {
			//Fit to discover algorithm to get model
			Object[] nets = IMPetriNet.minePetriNet(context, log,
					new MiningParametersIM());

			//Get original log and the discovered model => alignment and trace fitness
			// log + model + alignment => precision and generalization
			ModelQuality quality = runReplayToGetFitnessPrecisionAndGeneralization(
					context, nets, origLog, classifier);

			System.out.println("Fitness: " + quality.fitness + "; Precision: "
					+ quality.precision);
		}
		return null;
	}

	public static ModelQuality runDefaultTest(PluginContext context, XLog origLog) {
//		return runDefaultTest(context, origLog, new ExpDiscoveryHyberILP());
//		return runDefaultTest(context, origLog, new ExpDiscoveryInductiveMiner());
//		return runDefaultTest(context, origLog, new ExpDiscoveryFodina());
		return runDefaultTest(context, origLog, new RelabelHyberApproachIM());
	}

	public static ModelQuality runDefaultTest(PluginContext context, XLog origLog,
			ExpDiscoveryAlgorithm alg) {
		XEventClassifier classifier = new XEventNameClassifier();
//		XEventClassifier classifier = new XEventAttributeClassifier("NewLabel", "NewLabel");
		//Fit to discover algorithm to get model
		MarkedPetrinet net = alg.minePetrinet(context, origLog, classifier);

		//Get original log and the discovered model => alignment and trace fitness
		// log + model + alignment => precision and generalization
		ModelQuality quality = runReplayToGetFitnessPrecisionAndGeneralization(context,
				net, origLog, classifier);
		quality.setMarkedNet(net);

		System.out.println("Fitness: " + quality.fitness + "; Precision: "
				+ quality.precision);
		context.getProvidedObjectManager().createProvidedObject("Log",
				(XLog) origLog.clone(), XLog.class, context);

		context.getProvidedObjectManager().createProvidedObject("Net", net.getPetrinet(),
				Petrinet.class, context);
		return quality;
	}

	
	/**
	 * 
	 * @param context
	 * @param nets
	 * @param perfectLog
	 * @param classifier
	 * @return
	 */
	public static ModelQuality runReplayToGetFitnessPrecisionAndGeneralization(
			PluginContext context, Object[] nets, XLog perfectLog, XEventClassifier classifier) {
		Petrinet net = (Petrinet) nets[0];
		Marking initM = (Marking) nets[1];
		Marking finalM = (Marking) nets[2];
		return runReplayToGetFitnessPrecisionAndGeneralization(context, perfectLog,
				classifier, net, initM, finalM);
	}

	/**
	 * 
	 * @param context
	 * @param net
	 * @param perfectLog
	 * @param classifier
	 * @return
	 */
	public static ModelQuality runReplayToGetFitnessPrecisionAndGeneralization(
			PluginContext context, MarkedPetrinet net, XLog perfectLog,
			XEventClassifier classifier) {
		return runReplayToGetFitnessPrecisionAndGeneralization(context, perfectLog,
				classifier, net.getPetrinet(), net.getInitialMarking(),
				net.getFinalMarking());
	}

	/**
	 * 
	 * @param context
	 * @param perfectLog
	 * @param classifier
	 * @param net
	 * @param initM
	 * @param finalM
	 * @return
	 */
	private static ModelQuality runReplayToGetFitnessPrecisionAndGeneralization(
			PluginContext context, XLog perfectLog, XEventClassifier classifier, Petrinet net,
			Marking initM, Marking finalM) {

		XEventClasses newClasses = XEventClasses.deriveEventClasses(
				classifier, perfectLog);
		TransEvClassMapping etMapping = TransEvClassMappingUtils.getInstance()
				.getMapping(net, new HashSet<>(newClasses.getClasses()),
						newClasses.getClassifier());

		return runReplayToGetFitnessPrecisionAndGeneralization(context, perfectLog, net,
				initM, finalM, newClasses, etMapping);
	}

	public static ModelQuality runReplayToGetFitnessPrecisionAndGeneralization(
			PluginContext context, XLog perfectLog, Petrinet net, Marking initM,
			Marking finalM, XEventClasses newClasses, TransEvClassMapping etMapping) {
		ModelQuality quality = new ModelQuality();
		// use cost-based prefix algorithm
		CostBasedCompleteParam replayParam = ExprReplayUtils.getCostBasedCompleteParam(
				initM, finalM, newClasses, etMapping, 1, 1);
		IPNReplayAlgorithm replayAlgorithm = new PetrinetReplayerWithILP();

		if (replayAlgorithm.isAllReqSatisfied(context, net, perfectLog, etMapping,
				replayParam)) {
			PNLogReplayer replayer = new PNLogReplayer();
			PNRepResult replayResult = null;
			try {
				replayResult = replayer.replayLog(context, net, perfectLog, etMapping,
						replayAlgorithm, replayParam);

				/* ---------------------------------------------------------------
				 * Compute my recall and precision
				 * -------------------------------------------------------------- */
				computeMyRecallAndPrecision(replayResult, quality);

				/* ---------------------------------------------------------------
				 * Compute fitness & Precision
				 * -------------------------------------------------------------- */
				quality.fitness = getAlignmentValue(replayResult);

				MultiETCSettings precisionSettings = new MultiETCSettings();
				precisionSettings.put(MultiETCSettings.ALGORITHM,
						MultiETCSettings.Algorithm.ALIGN_1);
				precisionSettings.put(MultiETCSettings.REPRESENTATION,
						MultiETCSettings.Representation.ORDERED);
//				precisionSettings.put(MultiETCSettings.WINDOW,
//						MultiETCSettings.Window.BACKWARDS);

				MultiETCPlugin precisionPlugin = new MultiETCPlugin();
				Object[] precisions = precisionPlugin.checkMultiETCAlign1(context,
						perfectLog, net, precisionSettings, replayResult);
				quality.precision = (double) ((MultiETCResult) precisions[0])
						.getAttribute(MultiETCResult.BALANCED_PRECISION);
				
				// Only for experiment
//				quality.precisionDistinctlabels = (double) ((MultiETCResult) precisions[0])
//						.getAttribute(MultiETCResult.BALANCED_PRECISION_DISTINCT_LABEL);
//				quality.precisionDistinctlabelsTau = (double) ((MultiETCResult) precisions[0])
//						.getAttribute(MultiETCResult.BALANCED_PRECISION_DISTINCT_LABELANDTAU);
				
				/* ----------------------------------------------------------------
				 * Compute Wil's recall and precision w.r.t. log
				 * ---------------------------------------------------------------- */
				 MapTrcEnabled enabledSet = EventToEnabledSets.getEventToEnabledSets(replayResult, net, initM);
				 quality.setEnabledSet(enabledSet);

			} catch (AStarException e) {
				e.printStackTrace();
			} catch (ConnectionCannotBeObtained e) {
				e.printStackTrace();
			}

		}
		ExpModelSimplicityUtil.computeAllSimplicities(net, quality);
		return quality;
	}

	private static void computeMyRecallAndPrecision(PNRepResult replayResult,
			ModelQuality quality) {
		int numSM = 0, numMM = 0, numLM = 0, numMMinv = 0;
		for (SyncReplayResult alignment : replayResult) {
			int numTraces = alignment.getTraceIndex().size();
			if (alignment.isReliable()) {
				for (StepTypes type : alignment.getStepTypes()) {
					if (type.equals(StepTypes.L)) {
						numLM += numTraces;
					} else if (type.equals(StepTypes.MREAL)) {
						numMM += numTraces;
					} else if (type.equals(StepTypes.MINVI)) {
						numMMinv += numTraces;
					} else if (type.equals(StepTypes.LMGOOD)) {
						numSM += numTraces;
					}
				}
			}
		}
		quality.set("numSM", numSM);
		quality.set("numMM", numMM);
		quality.set("numLM", numLM);
		quality.set("numMMinv", numMMinv);
	}

	/**
	 * Return the trace fitness if more than 50% of traces have a reliable alignment, otherwise return -1. 
	 * @param pnRepResult
	 * @return
	 */
	public static double getAlignmentValue(PNRepResult pnRepResult) {
		int unreliable = 0;
		int total = 0;
		for (SyncReplayResult srp : pnRepResult) {
			if (!srp.isReliable()) {
				unreliable += srp.getTraceIndex().size();
			}
			total += srp.getTraceIndex().size();
		}
		if (unreliable > total / 2) {
			System.err.println("Fitness unreliable");
			return -1.0;
		} else {
			return (Double) pnRepResult.getInfo().get(PNRepResult.TRACEFITNESS);
		}
	}

	private static List<XLog> LoadLogs() {
		List<XLog> logs = new ArrayList<XLog>();
		for (int dist = 1; dist < 3; dist++) {
			for (int wMatch = 1; wMatch < 3; wMatch++) {
				for (int wNoMatch = 1; wNoMatch < 3; wNoMatch++) {
					for (int wGlobal = 1; wGlobal < 3; wGlobal++) {
						for (int wBase = 1; wBase < 6; wBase += 2) {
							for (int threshold = 0; threshold < 100; threshold += 10) {
								XLog log = loadLog(dist, wMatch, wNoMatch, wGlobal,
										wBase, threshold);
								if (log != null) {
									logs.add(log);
								}

							}
						}
					}
				}
			}
		}
		return logs;
	}

	private static XLog loadLog(int dist, int wMatch, int wNoMatch, int wGlobal,
			int wBase, int threshold) {
		XesXmlParser parser = new XesXmlParser();
		String file = "Log" + "D" + dist + "M" + wMatch + "N" + wNoMatch + "S" + wGlobal
				+ "B" + wBase + "F" + threshold + ".xes";

		try {
			List<XLog> res = parser.parse(new File("ExperimentResult/logs/" + file));
			return res.get(0);

		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

}
