package org.processmining.xlu.matching.zexperiment.precision;

import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;

import org.deckfour.xes.model.XLog;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;
import org.processmining.partialorder.models.palignment.Move;
import org.processmining.partialorder.models.palignment.MoveFactory;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

public class EventToEnabledSets {

	// TODO: need a wrapper for alignment...
	public static MapTrcEnabled getEventToEnabledSets(PNRepResult replayResult,
			Petrinet net, Marking initM) {
		PetrinetSemantics f = PetrinetSemanticsFactory
				.regularPetrinetSemantics(Petrinet.class);

		MapTrcEnabled tracemap = new MapTrcEnabled();
		for (SyncReplayResult alignment : replayResult) {
			if (alignment.isReliable()) {
				SortedSet<Integer> traceIndeces = alignment.getTraceIndex();
				f.initialize(net.getTransitions(), initM);
				int evtIdx = 0;
//				for (StepTypes type : alignment.getStepTypes()) {
				List<Transition> firingSequence = new ArrayList<Transition>();
				Map<Move, EnabledSet> map = new THashMap<Move, EnabledSet>();
				EnabledSet set = getEnabledSet(f, firingSequence, net, initM);
//				Move m = MoveFactory.createSyncMove(-1, -1, null, null);
//				map.put(, set);

				for (int i = 0; i < alignment.getStepTypes().size(); i++) {
					StepTypes type = alignment.getStepTypes().get(i);
					Object obj = alignment.getNodeInstance().get(i);
					Move newMove = null;
					EnabledSet nextSet = null;

					if (type.equals(StepTypes.LMGOOD)) {
						firingSequence.add((Transition) obj);
						nextSet = getEnabledSet(f, firingSequence, net, initM);

						newMove = MoveFactory.createSyncMove(i, evtIdx, null,
								(Transition) obj);
						evtIdx++;
					} else if (type.equals(StepTypes.MINVI)) {
						firingSequence.add((Transition) obj);
						nextSet = getEnabledSet(f, firingSequence, net, initM);

						newMove = MoveFactory.createInvisibleModelMove(i,
								(Transition) obj);

					} else if (type.equals(StepTypes.MREAL)) {
						firingSequence.add((Transition) obj);
						nextSet = getEnabledSet(f, firingSequence, net, initM);

						newMove = MoveFactory.createVisibleModelMove(i, (Transition) obj);

					} else if (type.equals(StepTypes.L)) {
						newMove = MoveFactory.createLogMove(i, evtIdx, null);
						evtIdx++;

					}

					map.put(newMove, set);
					if(newMove.isSilentModelMove()){
						nextSet.addAll(set.getAllEnabled());
					}
					set = nextSet;
//					m = newMove;
				}
				
				for (Integer i : traceIndeces) {
					tracemap.put(i, map);
				}
			}
		}
		return tracemap;
	}

	private static EnabledSet getEnabledSet(PetrinetSemantics f,
			List<Transition> firingSequence, Petrinet net, Marking initM) {
		Marking current = getMarkingOfFiringSequence(firingSequence, initM);
		f.setCurrentState(current);
		EnabledSet set = new EnabledSet(f.getExecutableTransitions());

		propagateTAU(f, firingSequence, net, initM, set, new THashSet<Transition>());
		return set;
	}

	private static void propagateTAU(PetrinetSemantics f,
			List<Transition> firingSequence, Petrinet net, Marking initM, EnabledSet set,
			Set<Transition> invsTested) {
		Set<Transition> invs = new THashSet<>(set.getInvisibleTransitions());
//		Set<Transition> invsTested = new THashSet<>();
//		int size = set.size();
		invs.removeAll(invsTested);

		while (!invs.isEmpty() || !invsTested.containsAll(invs)) {

			Transition t = invs.iterator().next();
			invsTested.add(t);
			invs.remove(t);

			firingSequence.add(t);

			EnabledSet newSet = getEnabledSetRecurse(f, firingSequence, net, initM,
					invsTested);

			set.addAll(newSet.getAllEnabled());

			for (Transition newInv : newSet.getInvisibleTransitions()) {
				if (!invsTested.contains(newInv)) {
					invs.add(newInv);
				}
			}

			firingSequence.remove(firingSequence.size() - 1);

		}
	}

	private static EnabledSet getEnabledSetRecurse(PetrinetSemantics f,
			List<Transition> firingSequence, Petrinet net, Marking initM,
			Set<Transition> invsTested) {
		Marking current = getMarkingOfFiringSequence(firingSequence, initM);
		f.setCurrentState(current);
		EnabledSet set = new EnabledSet(f.getExecutableTransitions());

		propagateTAU(f, firingSequence, net, initM, set, invsTested);
		return set;
	}

	private static Marking getMarkingOfFiringSequence(List<Transition> firingSequence,
			Marking initM) {
		Marking current = new Marking();
		current.addAll(initM);
		for (Transition t : firingSequence) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> p : t
					.getGraph().getOutEdges(t)) {
				Place post = (Place) p.getTarget();
				current.add(post);
			}
		}
		for (Transition t : firingSequence) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> p : t
					.getGraph().getInEdges(t)) {
				Place pre = (Place) p.getSource();
				current.remove(pre);
			}
		}
		return current;
	}

	public static RPDiagnostics compareEanebledSets(XLog log, MapTrcEnabled system,
			MapTrcEnabled model) {

		return null;
	}

}
