package org.processmining.xlu.matching.zexperiment.verification;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.experiment.alignment.XExperimentNoiseExtension;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.zexperiment.model.FusionBinaryResult;

public class EGMatchTestVerificationMethodMUDataImpl implements EGMatchTestVerificationMethod {

	

	public void computeCorrectness(RepEGraph representiveGraph, EGraph g1, EGraph g2, MatchMapping matchMapping,
			XLog log, FusionBinaryResult mappingQuality) {
		Map<Integer, Boolean> mapNode2Corrent = new HashMap<Integer, Boolean>();

		for (Entry<Integer, Map<Integer, Integer>> entry : representiveGraph.getNodeRepresentedNodes().entrySet()) {
			int newNodeId = entry.getKey();
			Map<Integer, Integer> mappedOrgGraphToNode = entry.getValue();
			Integer sourceNodeId = mappedOrgGraphToNode.get(g1.getId());
			Integer targetNodeId = mappedOrgGraphToNode.get(g2.getId());

			if (sourceNodeId == null && targetNodeId == null) {
				System.err.println("FusionBinaryVerification: Artifical node mapped to non artificial node");
			} else if (sourceNodeId != null && targetNodeId != null) {
				if (isArtificalMatch(g1, g2, sourceNodeId, targetNodeId)) {
					continue;
				}
				// Get original events and labels. 
				XEvent source = log.get(g1.getId()).get(sourceNodeId);
				String sourceLabel = getTrueLabel(source);
				XEvent target = log.get(g2.getId()).get(targetNodeId);
				String targetLabel = getTrueLabel(target);

				boolean equals = sourceLabel.equals(targetLabel);
				if (equals) {
					mappingQuality.truePositive++;

//					if (isNoise(source)) {
//						 Such that TP+FN = Sum Relevant =  All event + |Noise|
//						mappingQuality.truePositive++;
//					}
				} else {
					mappingQuality.falsePositive++;
				}

				mapNode2Corrent.put(newNodeId, equals);

			} else {
				XEvent e = null;
				Boolean foundLabel = null;
				if (targetNodeId == null) {
					e = log.get(g1.getId()).get(sourceNodeId);
					foundLabel = findLabel(e, log.get(g2.getId()), matchMapping.getTargetToNull());

				} else { // sourceNodeId == null
					e = log.get(g2.getId()).get(targetNodeId);
					foundLabel = findLabel(e, log.get(g1.getId()), matchMapping.getSourcesToNull());
				}

				if (isNoise(e) || !foundLabel) {
					mappingQuality.truePositive++;
					mapNode2Corrent.put(newNodeId, true);
				} else if (foundLabel) {
					mappingQuality.falsePositive++;
					mapNode2Corrent.put(newNodeId, false);
				}
			}
		}

		representiveGraph.setNodeToCorrect(mapNode2Corrent);
	}
	
	

	private Boolean findLabel(XEvent source, XTrace trace, Set<Integer> set) {
		String eTrueLabel = getTrueLabel(source);
		for (Integer i : set) { // MU-Data
			XEvent e = trace.get(i);
//		for (XEvent e : trace) { // Generated data

			if (eTrueLabel.equals(getTrueLabel(e)) && !isNoise(e)) {
				return true;
			}
		}
		return false;
	}

	private boolean isNoise(XEvent e) {
		return XExperimentNoiseExtension.isNoise(e);
	}

	private String getTrueLabel(XEvent e) {
		return XExperimentNoiseExtension.getTrueLabel(e);
	}

	private boolean isArtificalMatch(EGraph g1, EGraph g2, Integer sourceNodeId, Integer targetNodeId) {
		boolean isArtificalSource = EGraphFactory.getInstance().isArtifical(g1.getNodeById(sourceNodeId));
		boolean isArtificalTarget = EGraphFactory.getInstance().isArtifical(g2.getNodeById(targetNodeId));
		if (isArtificalSource != isArtificalTarget) {
			System.err.println("FusionBinaryVerification: Artifical node mapped to non artificial node");
		}

		return isArtificalSource && isArtificalTarget;
	}
}
