package org.processmining.xlu.pgraph.vis;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.partialorder.plugins.vis.PartialVisualType;
import org.processmining.plugins.inductiveVisualMiner.InductiveVisualMiner;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.pgraph.vis.color.PGraphColoring;
import org.processmining.xlu.pgraph.vis.converter.PGraphConverter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilter;
import org.processmining.xlu.pgraph.vis.filter.PGraphFilterHierarchyImpl;
import org.processmining.xlu.pgraph.vis.panel.PGraphInfoView;
import org.processmining.xlu.pgraph.vis.panel.PGraphListView;
import org.processmining.xlu.pgraph.vis.panel.PGraphVisView;
import org.processmining.xlu.relabel.visualization.RelabelInteractivePluginVisualizer;

public class PGraphController<T> {

	private PGraphListView<T> listView;
	private JPanel controlView;
	private PluginContext context;

	private List<PGraphFilter<T>> filters = new ArrayList<PGraphFilter<T>>();
	private PGraphFilter<ENode> eventfreqFilter; // TODO REMOVE!!!!!
	private PGraphColoring<XEventClass> evtColoring;
	private PGraphColoring<Integer> trColoring;
	private PGraphVisSetting setting;

	private PGraphConverter<T> converter;
	private JFrame frameRelabel;

	private XLog log;
	
	private boolean isShowInfoView = true;

	public PGraphController(PGraphConverter<T> converter, PluginContext context,
			XLog xLog) {
		this.converter = converter;
		this.context = context;
		this.log = xLog;
		setting = new PGraphVisSetting(this);
	}

	public PartialOrderGraph convert(T object) {
		PGraphParameter p = new PGraphParameter();
		if (evtColoring != null) {
			p.setEvtMap(evtColoring.getColorMap());
		}
//		p.setTraceMap(trColoring.getColorMap());
		p.setVisType(setting.getPGraphVisType() != null ? setting.getPGraphVisType()
				: PartialVisualType.AS_IS);
		p.setNodeFilter(this.eventfreqFilter);

		return converter.convert(object, p);
	}

	public boolean isUsingREGs() {
		return setting.isUsingREGs();
	}

	public boolean isUsingDiscovery() {
		return setting.isUsingDiscovery();
	}

	public boolean isUsingRelabel() {
		return setting.getVisType().equals("Relabel");
	}
	
	public boolean isUsingInductiveVisualMiner(){
		return setting.getVisType().equals("IvM");
	}

	public PGraphInfoView<T> getPGraphInfoView(PartialOrderGraph graph, T object) {
		return converter.createPGraphInfoView(graph, object);
	}

	public PGraphVisView<T> getPGraphVisView(PartialOrderGraph graph,
			PGraphInfoView<T> infoView) {
		return converter.createPGraphVisView(graph, infoView);
	}

	public void initializeFilters() {
		for (PGraphFilter<T> filter : filters) {
			filter.initialize();
		}
	}

	public boolean isSatisfied(T object) {
		boolean b = true;
		for (PGraphFilter<T> filter : filters) {
			b &= filter.isSatisfied(object);
			if (!b) {
				return b;
			}
		}
		return b;
	}

	public JComponent getControlView() {
		if (controlView == null) {
			controlView = new JPanel();
			controlView.setLayout(new BoxLayout(controlView, BoxLayout.Y_AXIS));
			controlView.setPreferredSize(new Dimension(300, 1000));
			controlView.setMaximumSize(new Dimension(300, 1000));

			controlView.setBackground(Color.DARK_GRAY);
			controlView.setForeground(Color.WHITE);

//			GridLayout experimentLayout = new GridLayout(0,1);
//			controlView.setLayout(experimentLayout);

			// add vis settings
//			Box boxContainer = Box.createHorizontalBox();
//			boxContainer.add(setting.getPanel());
//			boxContainer.setAlignmentX(Component.LEFT_ALIGNMENT);
//			boxContainer.add(setting.getPanel());
			controlView.add(setting.getPanel());

			// add filters
			for (PGraphFilter<T> filter : filters) {
//				 boxContainer = Box.createHorizontalBox();
//					boxContainer.add(filter.getVisualPanel());
//					boxContainer.setAlignmentX(Component.LEFT_ALIGNMENT);
//					boxContainer.add();
				JComponent panel = filter.getVisualPanel();
				panel.setAlignmentX(Component.LEFT_ALIGNMENT);
				if (filter instanceof PGraphFilterHierarchyImpl) {
					controlView.add(panel, 0);
				} else {
					controlView.add(panel);
				}
			}

			// add color settings
			if (evtColoring != null) {
				controlView.add(evtColoring.getPanel());
			}
//			controlView.add(trColoring.getPanel());

			Box boxContainer = Box.createHorizontalBox();
			boxContainer.add(eventfreqFilter.getVisualPanel());
			boxContainer.setAlignmentX(Component.LEFT_ALIGNMENT);
			boxContainer.add(Box.createHorizontalGlue());
			controlView.add(boxContainer);
		}
		return controlView;
	}

	public void setListView(PGraphListView<T> listView) {
		this.listView = listView;
	}

	public void update() {
		listView.update();
	}

	public void add(PGraphFilter<T> filter) {
		filters.add(filter);
	}

	public void setFilters(List<PGraphFilter<T>> filters) {
		this.filters = filters;
	}

	public void addColoring(PGraphColoring<XEventClass> evtColoring) {
		this.evtColoring = evtColoring;
	}

	public void addTrColoring(PGraphColoring<Integer> traceColoring) {
		this.trColoring = traceColoring;
	}

	public void addNodeFiltering(PGraphFilter<ENode> nodeFilter) {
		this.eventfreqFilter = nodeFilter;
	}

	public PluginContext getContext() {
		return this.context;

	}

	public PGraphFilter<ENode> getENodeFilter() {
		return eventfreqFilter;
	}

	public void updateRelabel(XLog newLog) {
		if (frameRelabel == null) {
			frameRelabel = new JFrame();
			frameRelabel.setPreferredSize(new Dimension(800, 400));
			frameRelabel.setMinimumSize(new Dimension(800, 400));
			frameRelabel.setVisible(true);

		} else {

			frameRelabel.setVisible(true);
		}
		JComponent panel = RelabelInteractivePluginVisualizer
				.visualizeByCreateSeperateFrame(getContext(), newLog);
//		panel.setPreferredSize(new Dimension(800, 400));
//		panel.setMinimumSize(new Dimension(800, 400));
		frameRelabel.getContentPane().removeAll();
		frameRelabel.getContentPane().add(panel);
		frameRelabel.repaint();
		frameRelabel.revalidate();
	}

	public void relabelNewView(T object) {
		if (object instanceof RepEGraph) {
			XLog newlog = getREGLog((RepEGraph) object, log);
			RelabelInteractivePluginVisualizer vis = new RelabelInteractivePluginVisualizer();
			updateRelabel(newlog);
		}
	}

	public JComponent relabelInView(T object) {
		if (object instanceof RepEGraph) {
			XLog newlog = getREGLog((RepEGraph) object, log);
			RelabelInteractivePluginVisualizer vis = new RelabelInteractivePluginVisualizer();
			return vis.visualizeByCreateSeperateFrame(context, newlog);
		}
		return null;
	}
	
	public JComponent showIvM(T object) {
		if (object instanceof RepEGraph) {
			XLog newlog = getREGLog((RepEGraph) object, log);
			InductiveVisualMiner vis = new InductiveVisualMiner();
			JComponent ivMComponent = vis.visualise(context, newlog, ProMCanceller.NEVER_CANCEL);
			ivMComponent.setPreferredSize(new Dimension(800, 500));
			ivMComponent.setMaximumSize(new Dimension(800, 500));
			return ivMComponent;
		}
		return null;
	}


	public MarkedPetrinet discover(T object) {
		// TODO Auto-generated method stub
		MarkedPetrinet net = null;
		ExpDiscoveryAlgorithm alg = setting.getDiscoveryAlg();
		if (object instanceof RepEGraph) {
			XLog newlog = getREGLog((RepEGraph) object, log);
			net = alg.minePetrinet(context, newlog);
		}
		return net;
	}

	private XLog getREGLog(RepEGraph repgraph, XLog log) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		XLog orgLog = log;
		XLog newLog = factory.createLog(orgLog.getAttributes());

		for (int traceIndex : repgraph.getRepresentedGraphIds()) {
			XTrace orgTrace = orgLog.get(traceIndex);

			XTrace newTrace = factory.createTrace(orgTrace.getAttributes());

			for (int eventIndex = 0; eventIndex < orgTrace.size(); eventIndex++) {
				ENode n = repgraph.getNodeRepresenting(traceIndex, eventIndex);

				if (this.getENodeFilter().isSatisfied(n)) {

					XEvent newEvent = (XEvent) orgTrace.get(eventIndex).clone();
					newTrace.add(newEvent);
				}
			}
			if (newTrace.size() > 0) {
				newLog.add(newTrace);
			}
		}
		if (newLog.size() == 0) {
			System.err.println("Log size 0");
		}
		return newLog;
	}

	public void setShowInfoView(boolean b) {
		isShowInfoView = b;
		
	}

	public boolean isShowInfoView() {
		return isShowInfoView;
	}


	


}
