package org.processmining.xlu.pgraph.vis;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.Border;

import org.processmining.partialorder.plugins.vis.PartialVisualType;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryHyberILP;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryInductiveMiner;
import org.processmining.xlu.relabel.visualization.RelabelMainPanel;

import com.fluxicon.slickerbox.factory.SlickerDecorator;

public class PGraphVisSetting {

	private Object visType = new ExpDiscoveryInductiveMiner();

	
	private JComponent panel;

	private PGraphController<?> controller;
	
	private static Border paddingBorder = BorderFactory.createEmptyBorder(10, 10, 10, 10);

	
	public PGraphVisSetting(PGraphController<?> controller){
		this.controller = controller;
	}
	
	public JComponent getPanel(){
		if(panel == null){
			panel = new JPanel();
			panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
			panel.setBackground(Color.DARK_GRAY);
			panel.setForeground(Color.WHITE);
			
			final JCheckBox box = new JCheckBox("Show info view");
			box.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {

					if (((JCheckBox) e.getSource()).isSelected()) {
						controller.setShowInfoView(true);
					} else {
						controller.setShowInfoView(false);
					}
					controller.update();
				}
			});
			box.setSelected(controller.isShowInfoView());
			SlickerDecorator.instance().decorate(box);
			box.setBackground(Color.DARK_GRAY);
			box.setForeground(Color.WHITE);
			box.setAlignmentX(Component.LEFT_ALIGNMENT);
			panel.add(box);
			
			final JLabel labelVisType = RelabelMainPanel.createLabel(panel, "Set visualization type: ");
			Box boxContainer = Box.createHorizontalBox();
			boxContainer.add(labelVisType);
			boxContainer.add(Box.createHorizontalGlue());
			boxContainer.setAlignmentX(Component.LEFT_ALIGNMENT);
			panel.add(boxContainer);
			
			final JComboBox<Object> visCombo = new JComboBox<Object>(
					getAllPossibleVisTypes());
			
			visCombo.setSelectedItem(getVisType());
			SlickerDecorator.instance().decorate(visCombo);
			visCombo.setBackground(Color.DARK_GRAY);
			visCombo.setForeground(Color.WHITE);
			visCombo.setAlignmentX(Component.LEFT_ALIGNMENT);
			visCombo.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					setVisType(visCombo.getSelectedItem());
					controller.update();
				}
			});
			panel.add(visCombo);
		}
		return panel;
	}

	private Object[] getAllPossibleVisTypes() {
		PartialVisualType[] comparableGraphVisualType = PartialVisualType.toComparableGraphVisualType();
		int numVisTypes = comparableGraphVisualType.length + 5;
		Object[] visTypes = new Object[numVisTypes];
		for(int i = 0; i < comparableGraphVisualType.length; i++){
			visTypes[i] = comparableGraphVisualType[i];
		}
		int i = comparableGraphVisualType.length;
		visTypes[i++] = visType;
		visTypes[i++] = "IvM";
		visTypes[i++] = new ExpDiscoveryHyberILP();
		visTypes[i++] = "Relabel";
		visTypes[i++] = "Relabel (new window)";
//		visTypes
		
		return visTypes;
	}

	public PartialVisualType getPGraphVisType() {
		if(isUsingREGs()){
			return (PartialVisualType)visType;
		}
		return null;
	}
	
	public ExpDiscoveryAlgorithm getDiscoveryAlg(){
		if(visType instanceof ExpDiscoveryAlgorithm){
			return (ExpDiscoveryAlgorithm)visType;
		}
		return null;
	}

	public void setVisType(Object visType) {
		this.visType = visType;
	}
	
	public Object getVisType(){
		return this.visType;
	}

	public boolean isUsingREGs() {
		return visType instanceof PartialVisualType;
	}

	public boolean isUsingDiscovery() {
		return visType instanceof ExpDiscoveryAlgorithm;
	}
}
