package org.processmining.xlu.pgraph.vis.color;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.processmining.partialorder.plugins.vis.projection.PProjectionConfigPanel;
import org.processmining.xlu.pgraph.vis.PGraphController;

public class PGraphColoringAbs<T> implements PGraphColoring<T> {

	String title;

	Map<T, Color> map = new HashMap<>();

	JPanel panel;

	private PGraphController<?> controller;

	private Collection<T> values;

	public PGraphColoringAbs(String title, Collection<T> values, PGraphController<?> controller) {
		this.title = title;
		this.controller = controller;
		this.values = values;

	}

	public Color getColor(T object) {
		Color c = map.get(object);
		if (c == null) {
			return null;
		}
		return c;
	}

	public JPanel getPanel() {
		if (panel == null) {
			panel = new JPanel();
			panel.setLayout(new BorderLayout());
			
			final JList<Object> list = new JList<Object>(values.toArray());
			list.getSelectionModel().addListSelectionListener(new ListSelectionListener() {

				@SuppressWarnings("unchecked")
				public void valueChanged(ListSelectionEvent e) {
					if (e.getValueIsAdjusting() == false) {
						if (list.getSelectedIndex() == -1) {
							//No selection, disable fire button.
							//fireButton.setEnabled(false);
						} else {
							Color newColor = JColorChooser.showDialog(panel, "Choose Background Color", null);
							map.put((T)list.getSelectedValue(), newColor);
							controller.update();
						}
					}
				}
			}
			);
			JComponent comp = PProjectionConfigPanel.configureSingleSelectionList(list, "Select " + title,
					"Visualize ");
			list.setCellRenderer(new SelectedListCellRenderer());
			panel.add(comp, BorderLayout.CENTER);
		}

		return panel;
	}

	public class SelectedListCellRenderer extends DefaultListCellRenderer {
		private static final long serialVersionUID = -2089347684127611653L;

		@SuppressWarnings("rawtypes")
		@Override
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
				boolean cellHasFocus) {
			Component c = super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

			c.setForeground(map.get(value));
			return c;
		}
	}

	public Map<T, Color> getColorMap() {
		
		return this.map;
	}

}
