package org.processmining.xlu.pgraph.vis.filter;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.pgraph.vis.PGraphController;
import org.processmining.xlu.relabel.visualization.RelabelMainPanel;

import com.fluxicon.slickerbox.factory.SlickerDecorator;

public class PGraphFilterEventFreqImpl extends PGraphFilterAbstract<ENode> implements ChangeListener, ActionListener {

	RepEGraphs graphs;
	int threshold = 0;
	boolean doRemoveNodes = false;

	JPanel panel;

	Map<Integer, Integer> mapGraphId2Max;
	private JLabel labelFilterValue;

	public PGraphFilterEventFreqImpl(PGraphController<?> controller, RepEGraphs graphs) {
		super(controller);
		this.graphs = graphs;
		initialize();
	}

	public boolean isSatisfied(ENode object) {
		Integer max = this.mapGraphId2Max.get(object.getGraphId());
		RepEGraph graph = graphs.getRepEGraph(object.getGraphId());
		if (max == null || graph == null) {
			System.err.println("PGraphFilterEventFreqImpl: No graph found");
			return false;
		}

		int ratio = graph.getNodeRepresentedNodes().get(object.getId()).size() * 100 / max;
		if (ratio >= this.getThreshold()) {
			return true;
		}
		return false;
	}

	public void initialize() {
		if (mapGraphId2Max == null) {

			this.mapGraphId2Max = new HashMap<Integer, Integer>();
			for (RepEGraph g : graphs) {
				int max = computeMaxFreq(g);

				this.mapGraphId2Max.put(g.getId(), max);
			}
		}
	}

	public static int computeMaxFreq(RepEGraph g) {
		int max = 0;
		Map<Integer, Map<Integer, Integer>> maps = g.getNodeRepresentedNodes();
		for (Entry<Integer, Map<Integer, Integer>> entry : maps.entrySet()) {
			Integer id = entry.getKey();
			if(EGraphFactory.getInstance().isArtifical(g.getNodeById(id))){
				continue;
			}
			Map<Integer, Integer> map = entry.getValue();
			max = map.size() > max ? map.size() : max;
		}
		return max;
	}

	public JComponent getVisualPanel() {
		if (panel == null) {
			panel = new JPanel();
			panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
			panel.setBackground(Color.DARK_GRAY);
			panel.setForeground(Color.WHITE);
			
			
			int initValue = threshold;
			final JLabel labelFilter = RelabelMainPanel.createLabel(panel, "Set Filter : ");
			Box boxContainer = Box.createHorizontalBox();
			boxContainer.add(labelFilter);
			labelFilterValue = RelabelMainPanel.createLabel(panel, String.valueOf(initValue));
			boxContainer.add(labelFilterValue);
			boxContainer.add(Box.createHorizontalGlue());
			boxContainer.setAlignmentX(Component.LEFT_ALIGNMENT);
			panel.add(boxContainer);
			
			JSlider framesPerSecond = new JSlider(JSlider.HORIZONTAL, 0, 100, initValue);
			framesPerSecond.setBackground(Color.DARK_GRAY);
			framesPerSecond.setForeground(Color.WHITE);
			framesPerSecond.addChangeListener(this);
			framesPerSecond.setAlignmentX(Component.LEFT_ALIGNMENT);
			panel.add(framesPerSecond);
			
			
//			SlickerButton button = new SlickerButton("Export logs");
//			button.addActionListener(this);
//			button.setAlignmentX(Component.LEFT_ALIGNMENT);
//			panel.add(button);
			
			final JCheckBox box = new JCheckBox("Hide Infreq. Nodes");
			box.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					doRemoveNodes = box.isSelected();
					controller.update();
				}
			});
			SlickerDecorator.instance().decorate(box);
			box.setBackground(Color.DARK_GRAY);
			box.setForeground(Color.WHITE);
			box.setAlignmentX(Component.LEFT_ALIGNMENT);
			
			panel.add(box);
			
		}
		return panel;
	}

	public void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider) e.getSource();
		int threshold = (int) source.getValue();
		labelFilterValue.setText(String.valueOf(threshold));
		if (!source.getValueIsAdjusting()) {
			
			
			setThreshold(threshold);
			
			controller.update();
		}
	}

	public void actionPerformed(ActionEvent e) {
		XLog newLog = getFilteredLog();
		controller.getContext().getProvidedObjectManager()
				.createProvidedObject("Filtered Log with " + getThreshold(), newLog, XLog.class, controller.getContext());

	}

	public int getThreshold() {
		return threshold;
	}

	public void setThreshold(int threshold) {
		System.out.println("Threshold set to " + threshold);
		this.threshold = threshold;
	}

	public XLog getFilteredLog() {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog orgLog = this.graphs.getLog();
		XLog newLog = factory.createLog(orgLog.getAttributes());

		//		System.out.println("Rep has " + this.graphs.getGraphs().size() + " graphs");

		for (int traceIndex = 0; traceIndex < orgLog.size(); traceIndex++) {
			XTrace orgTrace = orgLog.get(traceIndex);

			XTrace newTrace = factory.createTrace(orgTrace.getAttributes());

			// Find the representative graph for this trace.
			RepEGraph repgraph = null;
			for (int i = 0; i < this.graphs.getHierarchy().getChildCount(); i++) { 
				// FIXME: buggy. Assuming the regs are sorted based on the hierarchy..

				RepEGraph g = ((Tree<RepEGraph>) this.graphs.getHierarchy().getChildAt(i)).getObject();
				if (g.getRepresentedGraphIds().contains(traceIndex)) {
					repgraph = g;
					break;
				}
			}
			if (repgraph == null) {
				System.err.println("Trace " + traceIndex + " name "
						+ XConceptExtension.instance().extractName(orgTrace) + " not found in RepGraphs");
			}
			for (int eventIndex = 0; eventIndex < orgTrace.size(); eventIndex++) {
				ENode n = repgraph.getNodeRepresenting(traceIndex, eventIndex);
				if (this.isSatisfied(n)) {
					// clone event and add to new trace
					XEvent newEvent = (XEvent) orgTrace.get(eventIndex).clone();
					newTrace.add(newEvent);
				}
			}
			if (newTrace.size() > 0) {
				newLog.add(newTrace);
			} else {
//				System.out.println("Empty trace");
			}
		}
		if(newLog.size() == 0){
			System.err.println("Log size 0");
		}
		return newLog;
	}

	public boolean doRemoveNodes() {
		return doRemoveNodes;
	}

	public void setRemoveNodes(boolean doRemoveNodes) {
		this.doRemoveNodes = doRemoveNodes;
	}

}
