package org.processmining.xlu.pgraph.vis.filter;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeSelectionModel;

import org.apache.commons.collections15.map.HashedMap;
import org.deckfour.xes.classification.XEventAttributeClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.pgraph.vis.PGraphController;
import org.processmining.xlu.relabel.visualization.RelabelMainPanel;

import com.fluxicon.slickerbox.components.SlickerButton;
import com.fluxicon.slickerbox.factory.SlickerDecorator;

public class PGraphFilterHierarchyImpl extends PGraphFilterAbstract<RepEGraph> {

	private List<RepEGraph> selectedGraphs = new ArrayList<>();
	private Tree<RepEGraph> hierarchy;
	private XLog log;
	JTree visTree;

	private boolean isShowSubcluster = false;

	public PGraphFilterHierarchyImpl(PGraphController<?> controller,
			Tree<RepEGraph> hierarchy, XLog xLog) {
		super(controller);
		this.hierarchy = hierarchy;
		this.log = xLog;
	}

	public boolean isSatisfied(RepEGraph object) {
		return selectedGraphs.contains(object);
	}

	public void initialize() {
		// TODO Auto-generated method stub

	}

	public JComponent getVisualPanel() {
		//		Tree<RepEGraph> hierarchy = graphs.getHierarchy();
		//		if(hierarchy != null){
		JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
		panel.setBackground(Color.DARK_GRAY);
		panel.setForeground(Color.WHITE);

		JLabel labelCost = RelabelMainPanel.createLabel(panel,
				"Select the cluster to view: ");
		labelCost.setAlignmentX(Component.LEFT_ALIGNMENT);
		panel.add(labelCost);

		if (visTree == null) {
			visTree = new JTree(hierarchy);

			//			visTree.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
			visTree.setBackground(Color.GRAY);
			visTree.setForeground(Color.WHITE);

			visTree.getSelectionModel()
					.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

			//Listen for when the selection changes.
			visTree.addTreeSelectionListener(new TreeSelectionListener() {

				@SuppressWarnings("unchecked")
				public void valueChanged(TreeSelectionEvent e) {
					//Returns the last path element of the selection.
					//This method is useful only when the selection model allows a single selection.
					Tree<RepEGraph> node = (Tree<RepEGraph>) visTree
							.getLastSelectedPathComponent();

					updateSelectedGraphs(node, isShowSubcluster);
					controller.update();
				}
			});

			visTree.setCellRenderer(new MyTreeCellRenderer());
			
//			ClusterExpanderBlockStructure expander = new ClusterExpanderBlockStructure(log);
//			expander.expandTree(visTree,hierarchy);
		}

		visTree.setAlignmentX(Component.LEFT_ALIGNMENT);
		//		visTree.setBorder(BorderFactory.createEmptyBorder());
		JScrollPane treeView = new JScrollPane(visTree);
		//		treeView.setMaximumSize(new Dimension(1000, 300));
		treeView.setBorder(BorderFactory.createEmptyBorder());
		treeView.setAlignmentX(Component.LEFT_ALIGNMENT);
		panel.add(treeView);

		SlickerButton button = new SlickerButton("Export Selected Cluster");
		button.setAlignmentX(Component.LEFT_ALIGNMENT);
		//		SlickerDecorator.instance().decorate(button);
		button.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				XLog newLog = getFilteredLog();
				controller.getContext().getProvidedObjectManager().createProvidedObject(
						"Cluster Log", newLog, XLog.class, controller.getContext());
			}
		});

//		SlickerButton bMineModel = new SlickerButton("Show mining model with relabel");
//		bMineModel.setAlignmentX(Component.LEFT_ALIGNMENT);
//		bMineModel.addActionListener(new ActionListener() {
//			public void actionPerformed(ActionEvent e) {
//				XLog newLog = getFilteredLog();
//
//				controller.updateRelabel(newLog);
//			}
//		});

		//		JButton dfgbutton = new JButton("Show cluster as DFG");
		//		button.addActionListener(new ActionListener() {
		//			public void actionPerformed(ActionEvent e) {
		////				XLog newLog = getFilteredLog();
		////				XLog2Dfg.log2Dfg();
		//			}
		//		});
		//		

		final JCheckBox box = new JCheckBox("Show child-clusters");
		box.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				Tree<RepEGraph> node = (Tree<RepEGraph>) visTree
						.getLastSelectedPathComponent();

				if (((JCheckBox) e.getSource()).isSelected()) {
					isShowSubcluster = true;
				} else {
					isShowSubcluster = false;
				}
				updateSelectedGraphs(node, isShowSubcluster);
				controller.update();
			}
		});
		box.setSelected(isShowSubcluster);
		SlickerDecorator.instance().decorate(box);
		box.setBackground(Color.DARK_GRAY);
		box.setForeground(Color.WHITE);
		box.setAlignmentX(Component.LEFT_ALIGNMENT);

		panel.add(button);
//		panel.add(bMineModel);
		panel.add(box);
		//		panel.add(dfgbutton, BorderLayout.PAGE_END);

		return panel;
	}



	private void updateSelectedGraphs(Tree<RepEGraph> node,
			boolean isShowSubcluster) {
		if (node == null) {
			//Nothing is selected.     
			return;
		}
		selectedGraphs.clear();
		
		RepEGraph g = node.getObject();
		selectedGraphs.add(g);				
		if (isShowSubcluster) {
			for (int i = 0; i < node.getChildCount(); i++) {
				selectedGraphs
						.add(((Tree<RepEGraph>) node.getChildAt(i)).getObject());
			}
		}

	}

	public XLog getFilteredLog() {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		//		XLog orgLog = this.graphs.getLog();
		XLog orgLog = this.log;
		XLog newLog = factory.createLog(orgLog.getAttributes());

		//		System.out.println("Rep has " + this.graphs.getGraphs().size() + " graphs");

		//		for (int traceIndex = 0; traceIndex < orgLog.size(); traceIndex++) {
		RepEGraph repgraph = selectedGraphs.get(0);
		// Labeling...
		Map<Integer, String> relabels = new HashedMap<>();

		for (int traceIndex : repgraph.getRepresentedGraphIds()) {
			XTrace orgTrace = orgLog.get(traceIndex);

			XTrace newTrace = factory.createTrace(orgTrace.getAttributes());

			// Find the representative graph for this trace.

			//			for (int i = 0; i < this.graphs.getHierarchy().getChildCount(); i++) { // FIXME: buggy. Assuming the regs are sorted based on the hierarchy..

			//				RepEGraph g = ((Tree<RepEGraph>) this.graphs.getHierarchy().getChildAt(i)).getObject();
			//				if (g.getRepresentedGraphIds().contains(traceIndex)) {
			//					repgraph = g;
			//					break;
			//				}
			//			}
			if (repgraph == null) {
				System.err.println("Trace " + traceIndex + " name "
						+ XConceptExtension.instance().extractName(orgTrace)
						+ " not found in RepGraphs");
			}
			for (int eventIndex = 0; eventIndex < orgTrace.size(); eventIndex++) {
				ENode n = repgraph.getNodeRepresenting(traceIndex, eventIndex);

				if (this.controller.getENodeFilter().isSatisfied(n)) {

					//Relabel only nodes to be exported...
					if (!relabels.containsKey(n.getId())) {
						String label = n.getLabel();
						String newLabel = label;
						int i = 1;
						while (relabels.values().contains(newLabel)) {
							newLabel = label + "_R" + i;
							i++;
						}
						relabels.put(n.getId(), newLabel);
					}

					// clone event and add to new trace
					XEvent newEvent = (XEvent) orgTrace.get(eventIndex).clone();
					newEvent.getAttributes().put("NewLabel",
							factory.createAttributeLiteral("NewLabel",
									relabels.get(n.getId()), null));
					newTrace.add(newEvent);
				}
			}
			if (newTrace.size() > 0) {
				newLog.add(newTrace);
			} else {
				//				System.out.println("Empty trace");
			}
		}
		if (newLog.size() == 0) {
			System.err.println("Log size 0");
		}
		newLog.getClassifiers()
				.add(new XEventAttributeClassifier("Relabeled", "NewLabel"));
		return newLog;
	}

	public class MyTreeCellRenderer extends DefaultTreeCellRenderer {

		/**
		 * 
		 */
		private static final long serialVersionUID = 3541685795888538905L;

		@Override
		public Component getTreeCellRendererComponent(JTree tree, Object value,
				boolean sel, boolean exp, boolean leaf, int row, boolean hasFocus) {
			super.getTreeCellRendererComponent(tree, value, sel, exp, leaf, row,
					hasFocus);

			// Assuming you have a tree of Strings
			RepEGraph node = (RepEGraph) ((Tree<RepEGraph>) value).getObject();

			// If the node is a leaf and ends with "xxx"
			//	        if (leaf && node.endsWith("xxx")) {
			// Paint the node in blue
			setForeground(new Color(13, 57, 115));
			setBackground(Color.GRAY);
			if (node != null) {
				int number = node.getRepresentedGraphIds().size();
				setText("Cluster " + node.getId() + " (" + number + " trcs)");
			} else {
				setText("Root");
			}
			//	        }

			return this;
		}
	}

}
