package org.processmining.xlu.pgraph.vis.filter;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.pgraph.vis.PGraphController;

public class PGraphFilterLabelAEventFreqImpl extends PGraphFilterAbstract<ENode> implements ChangeListener, ActionListener {

	RepEGraphs graphs;
	int threshold = 50;
	boolean doRemoveNodes = false;



	JPanel panel;

	Map<Integer, Map<String, Integer>> mapGraphId2Max;

	public PGraphFilterLabelAEventFreqImpl(PGraphController<?> controller, RepEGraphs graphs) {
		super(controller);
		this.graphs = graphs;
		initialize();
	}

	public boolean isSatisfied(ENode object) {
		Map<String, Integer> max = this.mapGraphId2Max.get(object.getGraphId());
		RepEGraph graph = graphs.getRepEGraph(object.getGraphId());
		if (max == null || graph == null) {
			System.err.println("PGraphFilterEventFreqImpl: No graph found");
			return false;
		}

		int ratio = graph.getNodeRepresentedNodes().get(object.getId()).size() * 100 / max.get(object.getLabel());
		if (ratio >= this.getThreshold()) {
			return true;
		}
		return false;
	}

	public void initialize() {
		if (mapGraphId2Max == null) {

			this.mapGraphId2Max = new HashMap<Integer, Map<String, Integer>>();
			for (RepEGraph g : graphs) {
				 Map<String, Integer> map = getLabelToFreqMap(g);
				
				this.mapGraphId2Max.put(g.getId(), map);
			}
		}
	}
	
	public static Map<String, Integer> getLabelToFreqMap(RepEGraph graph) {
		Map<String, Integer> mapL2Max = new HashMap<String, Integer>();
		Map<Integer, Map<Integer, Integer>> maps = graph.getNodeRepresentedNodes();
		for (Entry<Integer, Map<Integer, Integer>> entry : maps.entrySet()) {
			String label = graph.getNodeById(entry.getKey()).getLabel();
			Integer max = mapL2Max.get(label);
			int newSize = entry.getValue().size();
			int newMax = max != null && max > newSize ? max : newSize;
			mapL2Max.put(label, newMax);
		}
		return mapL2Max;		
	}


	/* duplicated code!!!!!!*/
	public JComponent getVisualPanel() {
		if (panel == null) {
			panel = new JPanel();
			JSlider framesPerSecond = new JSlider(JSlider.HORIZONTAL, 0, 100, 50);
			framesPerSecond.addChangeListener(this);
			panel.add(framesPerSecond);
//			JButton button = new JButton("Export logs"); 
//			button.addActionListener(this);
			final JCheckBox box = new JCheckBox("Remove Nodes");
			box.addActionListener(new ActionListener() {
				
				public void actionPerformed(ActionEvent e) {
					doRemoveNodes = box.isSelected();
					controller.update();					
				}
			});
			panel.add(box);
//			panel.add(button);
		}
		return panel;
	}

	public void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider) e.getSource();
		if (!source.getValueIsAdjusting()) {
			setThreshold((int) source.getValue());
			controller.update();
		}
	}

	public void actionPerformed(ActionEvent e) {
		XLog newLog = getFilteredLog();
		controller.getContext().getProvidedObjectManager().createProvidedObject("Filtered Log with "+this.getThreshold(), newLog, XLog.class, controller.getContext());
		
	}
	
	public int getThreshold() {
		return threshold;
	}

	public void setThreshold(int threshold) {
		System.out.println("Threshold is set to " + threshold);
		this.threshold = threshold;
	}


	public XLog getFilteredLog() {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog orgLog = this.graphs.getLog();
		XLog newLog = factory.createLog(orgLog.getAttributes());
		
//		System.out.println("Rep has " + this.graphs.getGraphs().size() + " graphs");
	
		for(int traceIndex = 0; traceIndex < orgLog.size(); traceIndex++){
			XTrace orgTrace = orgLog.get(traceIndex);
			
			XTrace newTrace = factory.createTrace(orgTrace.getAttributes());

			// Find the representative graph for this trace.
			RepEGraph repgraph = null;
			for (int i = 0; i < this.graphs.getHierarchy().getChildCount(); i++) { // FIXME: buggy. Assuming the regs are sorted based on the hierarchy..
				
				RepEGraph g = ((Tree<RepEGraph>)this.graphs.getHierarchy().getChildAt(i)).getObject();
				if(g .getRepresentedGraphIds().contains(traceIndex)){
					repgraph = g;
					break;
				}
			}
			if(repgraph == null){
				System.err.println("Trace " + traceIndex + " name " + XConceptExtension.instance().extractName(orgTrace) + " not found in RepGraphs");
			}
			for(int eventIndex = 0; eventIndex <orgTrace.size(); eventIndex++){
				ENode n = repgraph.getNodeRepresenting(traceIndex, eventIndex);
				if(this.isSatisfied(n)){
					// clone event and add to new trace
					XEvent newEvent = (XEvent)orgTrace.get(eventIndex).clone();
					newTrace.add(newEvent);
				}
			}
			newLog.add(newTrace);
		}
		return newLog;
	}
	
	
	public boolean doRemoveNodes() {
		return doRemoveNodes;
	}

	public void setRemoveNodes(boolean doRemoveNodes) {
		this.doRemoveNodes = doRemoveNodes;
	}

}
