package org.processmining.xlu.pgraph.vis.filter.expander;

import javax.swing.JTree;
import javax.swing.tree.TreePath;

import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.relabel.plugin.detection.ILDetectorIMGlobal;
import org.processmining.xlu.relabel.plugin.detection.ILDetectorIMLocal;

public class ClusterExpanderBlockStructure implements ClusterExpander<RepEGraph> {
	
	private XLog log;

	public ClusterExpanderBlockStructure(XLog log){
		this.log = log;
	}

	public void expandTree(JTree tree, Tree<RepEGraph> hierarchy) {
//		DefaultMutableTreeNode node = (DefaultMutableTreeNode) hierarchy.getRoot();
		for (int i = 0; i < hierarchy.getChildCount(); i++) {
			Tree<RepEGraph> childAt = (Tree<RepEGraph>) hierarchy.getChildAt(i);
			RepEGraph egraph = childAt.getObject();
			if (doExpand(egraph)) {
				tree.expandPath(new TreePath(childAt.getPath()));
			}
			expandTree(tree, childAt);
		}
	}

	private boolean doExpand(RepEGraph egraph) {
		XLog newLog = getSublog(egraph, log);
		MiningParametersIM param = new MiningParametersIM();
		ProcessTree tree = IMProcessTree.mineProcessTree(newLog, param);
		for (Node n : tree.getNodes()) {
			if (ILDetectorIMLocal.isTauLoop(n) && ILDetectorIMGlobal.recursivelyGetLeaves(n).size() > 2) {
				return true;
			}
		}
		return false;
	}

	private XLog getSublog(RepEGraph repgraph, XLog orgLog) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog newLog = factory.createLog(orgLog.getAttributes());
		for (int traceIndex : repgraph.getRepresentedGraphIds()) {
			XTrace orgTrace = orgLog.get(traceIndex);
			newLog.add(orgTrace);
		}
		return newLog;
	}

	public String toString(){
		return "expand until no loop";
	}
}
