package org.processmining.xlu.pgraph.vis.filter.expander;

import java.util.Date;
import java.util.concurrent.TimeUnit;

import javax.swing.JTree;
import javax.swing.tree.TreePath;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;

public class ClusterExpanderPerformance implements ClusterExpander<RepEGraph> {
	
	private XLog log;

	public ClusterExpanderPerformance(XLog log){
		this.log = log;
	}

	public void expandTree(JTree tree, Tree<RepEGraph> hierarchy) {
//		DefaultMutableTreeNode node = (DefaultMutableTreeNode) hierarchy.getRoot();
		for (int i = 0; i < hierarchy.getChildCount(); i++) {
			Tree<RepEGraph> childAt = (Tree<RepEGraph>) hierarchy.getChildAt(i);
			RepEGraph child = childAt.getObject();
			if (doExpand(child, hierarchy)) {
				tree.expandPath(new TreePath(childAt.getPath()));
			}
			expandTree(tree, childAt);
		}
	}

	private boolean doExpand(RepEGraph current, Tree<RepEGraph> hierarchy) {
//		RepEGraph parent = ((Tree<RepEGraph>) hierarchy.getRoot()).getObject();
//		XLog newLog = getSublog(parent, log);
		
		XLog newLog = getSublog(current, log);
		long d = 0;
		for(XTrace t : newLog){
			if(t.size() > 1){
				Date start = XTimeExtension.instance().extractTimestamp(t.get(0));
				Date end = XTimeExtension.instance().extractTimestamp(t.get(t.size()-1));
				d += getDateDiff(start,end,TimeUnit.DAYS);
			}
			
		}
		d = d / newLog.size();	
	
		return d > 30.0;
	}
	
	public static long getDateDiff(Date date1, Date date2, TimeUnit timeUnit) {
	    long diffInMillies = date2.getTime() - date1.getTime();
	    return timeUnit.convert(diffInMillies,TimeUnit.MILLISECONDS);
	}

	private XLog getSublog(RepEGraph repgraph, XLog orgLog) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog newLog = factory.createLog(orgLog.getAttributes());
		for (int traceIndex : repgraph.getRepresentedGraphIds()) {
			XTrace orgTrace = orgLog.get(traceIndex);
			newLog.add(orgTrace);
		}
		return newLog;
	}

	public String toString(){
		return "expand until no loop";
	}
}
