package org.processmining.xlu.pgraph.vis.panel;

import java.awt.BorderLayout;
import java.awt.Color;
import java.util.Collection;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;

import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.pgraph.vis.PGraphController;
import org.processmining.xlu.relabel.visualization.RelabelMainPanel;

public class PGraphListView<T> extends JComponent {

	/**
	 * 
	 */
	private static final long serialVersionUID = -365686697259804291L;
	private Collection<T> objects;
//	private PGraphConverter<T> converter;
	private PGraphController<T> controller;

	public PGraphListView(Collection<T> objects, PGraphController<T> controller) {
		super();
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		setBackground(Color.DARK_GRAY);
		setForeground(Color.WHITE);

//		this.converter = converter;
		this.objects = objects;
		this.controller = controller;

		addGraphs();
	}

	private void addGraphs() {
		int count = 0;
		for (T object : objects) {
			// filter
			if (!controller.isSatisfied(object)) {
				continue;
			}

			System.out.println("Adding " + count++);
			JComponent graphViewPanel = new JPanel();
//			graphViewPanel.setLayout(new BoxLayout(graphViewPanel, BoxLayout.X_AXIS));
			graphViewPanel.setLayout(new BorderLayout());
			graphViewPanel.setBackground(Color.DARK_GRAY);
			graphViewPanel.setForeground(Color.WHITE);
			TitledBorder border = BorderFactory.createTitledBorder("Cluster ");
			border.setTitleColor(Color.WHITE);
//			border.setTitleFont(border.getTitleFont().deriveFont((float) 12.0));
			border.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
			graphViewPanel.setBorder(border);

			// add graph and info panel
			PartialOrderGraph g = null;
			PGraphInfoView<T> graphInfoViewPanel = null;
			JComponent graphVisViewPanel = null;

			if (controller.isUsingREGs()) {
				g = controller.convert(object);

				graphVisViewPanel = controller.getPGraphVisView(g, graphInfoViewPanel);

			} else if (controller.isUsingDiscovery()) {
				// is using a discovery
				MarkedPetrinet net = controller.discover(object);
				graphVisViewPanel = RelabelMainPanel
						.getPanelOfModel(controller.getContext(), net);
			} else if (controller.isUsingInductiveVisualMiner()) {

				graphVisViewPanel = controller.showIvM(object);

			} else if (controller.isUsingRelabel()) {

				// is relabeling
				graphVisViewPanel = controller.relabelInView(object);
			} else {
				// relabel and show a new frame
				g = controller.convert(object);
				graphVisViewPanel = controller.getPGraphVisView(g, graphInfoViewPanel);
				controller.relabelNewView(object);
			}
			if (controller.isShowInfoView()) {
				graphInfoViewPanel = controller.getPGraphInfoView(g, object);
				graphViewPanel.add(graphInfoViewPanel, BorderLayout.WEST);
			}

			graphViewPanel.add(graphVisViewPanel, BorderLayout.CENTER);

			add(graphViewPanel);
		}

	}

	public void update() {
		this.removeAll();
		controller.initializeFilters();
		addGraphs();

		this.revalidate();
		this.repaint();
//		pTracesScrollPane.setViewportView(this);
	}
}
