package org.processmining.xlu.pgraph.vis.panel;

import java.awt.BorderLayout;
import java.awt.Color;

import javax.swing.SwingConstants;

import org.jgraph.event.GraphSelectionEvent;
import org.jgraph.event.GraphSelectionListener;
import org.processmining.framework.util.ui.scalableview.ScalableViewPanel;
import org.processmining.framework.util.ui.scalableview.interaction.ZoomInteractionPanel;
import org.processmining.framework.util.ui.widgets.WidgetColors;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.models.jgraph.elements.ProMGraphCell;
import org.processmining.models.jgraph.elements.ProMGraphEdge;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.partialorder.models.graph.POEdge;
import org.processmining.partialorder.models.graph.PONode;
import org.processmining.partialorder.models.graph.PartialOrderGraph;
import org.processmining.partialorder.plugins.vis.GraphExportablePanel;
import org.processmining.partialorder.util.VisUtil;

public class PGraphVisView<T> extends GraphExportablePanel implements GraphSelectionListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = 7497413316434260532L;
	
	PGraphInfoView<T> infoView;

	private PartialOrderGraph graph;
	
	public PGraphVisView(PartialOrderGraph g, PGraphInfoView<T> graphInfoViewPanel) {
		super(10, 5, 0);
		initLayout();
		
		this.infoView = graphInfoViewPanel;	
		this.graph = g;
		
		comp = VisUtil.getProMSimpleJGraph(graph, VisUtil.LightBackgroundColor);
		
		if(comp != null){
			comp.addGraphSelectionListener(this);
			ProMJGraphPanel scalable = getJGraphScrollPane(comp);
			comp.addMouseListener(this);
			this.add(scalable, BorderLayout.CENTER);
		}
	}
	
	
	private static ProMJGraphPanel getJGraphScrollPane(ProMJGraph comp) {
		ProMJGraphPanel panel = new ProMJGraphPanel(comp);
		//		panel.setSize(300, 300);
		panel.addViewInteractionPanel(new ZoomInteractionPanel(panel, ScalableViewPanel.MAX_ZOOM),
				SwingConstants.WEST);
		panel.scaleToFit();
		return panel;
	}


	private void initLayout() {
		setBackground(WidgetColors.COLOR_ENCLOSURE_BG);
		setLayout(new BorderLayout());
//		setMinimumSize(new Dimension(200, 100));
//		setMaximumSize(new Dimension(1000, 1000));
//		setPreferredSize(new Dimension(1000, 200));
//		add(Box.createHorizontalStrut(5), BorderLayout.WEST);
//		add(Box.createHorizontalStrut(5), BorderLayout.EAST);
	}

	/* Selected a node or an edge, update infoView
	 */
	public void valueChanged(GraphSelectionEvent evt) {
		for (Object cell : evt.getCells()) {
			if (!evt.isAddedCell(cell)) {
				continue;
			}
			if (cell instanceof ProMGraphCell) {
				if (((ProMGraphCell) cell).getNode() instanceof PONode) {
					// Partially Ordered Trace Information visualization
					infoView.updateInfo((PONode) ((ProMGraphCell) cell).getNode());
				} 

			} else if (cell instanceof ProMGraphEdge && ((ProMGraphEdge) cell).getEdge() instanceof POEdge) {
				infoView.updateInfo((POEdge) ((ProMGraphEdge) cell).getEdge());
			} else {
				infoView.updateGraphGeneralInfo();
			}
			
		}
	}
	
	
	public void updateNodeColor(PONode node, Color c) {
		comp.getModel().beginUpdate();
		// 
		for (PONode n : graph.getNodes()) {

			if(n.equals(node))
			if (c != null) {
				n.setColor(c);
			}
		}

		comp.getModel().endUpdate();
		comp.refresh();
		comp.revalidate();
		comp.repaint();
	}

}
