package org.processmining.xlu.relabel.experiment;

import gnu.trove.set.hash.THashSet;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.processtree.Block;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;
import org.processmining.processtree.impl.AbstractTask.Automatic;
import org.processmining.processtree.impl.AbstractTask.Manual;
import org.processmining.xlu.ageneral.UtilMathRandom;

public class ExpDuplicatedTaskUtil {

	public static final String ORG_LABEL = "OrgLabel";

	public static XLog addDuplicatedLabels(XLog origLog, XEventClasses classes,
			ExpImprLabelSetI<String> mapSourceToTarget) {

		XLog log = (XLog) origLog.clone();
		for (XTrace t : log) {
			for (XEvent e : t) {
				String eclass = classes.getClassOf(e).toString();
				XAttribute a = XFactoryRegistry.instance().currentDefault()
						.createAttributeLiteral(ORG_LABEL, eclass.toString(), null);
				e.getAttributes().put(ORG_LABEL, a);
				
				if (!mapSourceToTarget.getNodeAs().contains(eclass)) {
					continue;
				}

				String targetClass = mapSourceToTarget.getNodeB();
				XConceptExtension.instance().assignName(e, targetClass);
			}
		}
		return log;
	}

	/**
	 * 
	 * @param tree
	 * @param numLeaves
	 *            >= 2
	 * @return
	 */
	public static ExpImprLabelPair<String> getRandomDupTaskSet(ProcessTree tree,
			int numLeaves, int numAllowinLoop) {
		ExpImprLabelPair<String> pair = new ExpImprLabelPair<String>();

		// Get all nodes that have the input type
		List<Node> selectedNodes = new ArrayList<Node>();
		for (Node n : tree.getNodes()) {
			if (n.isLeaf() && n instanceof org.processmining.processtree.Task.Manual) {
				selectedNodes.add(n);
			}
		}
		// Skip if no such nodes 
		if (selectedNodes.size() == 0) {
			return null;
		}

		List<Integer> randomNodeIndices = UtilMathRandom.getRandomUniqueInt(
				selectedNodes.size(), selectedNodes.size());

		Set<String> leaves = new THashSet<>();
//		for (int i = 0; i < randomNodeIndices.size() - 1; i++) {
		int i = 0;
		while(leaves.size() < numLeaves && i < randomNodeIndices.size()){
			Node node = selectedNodes.get(randomNodeIndices.get(i));
			boolean isInLoop = isInLoop(node);
			
			if (isInLoop) {
				if(numAllowinLoop > 0){
					numAllowinLoop--;
					leaves.add(node.getName());
				} // else skip 
			} else {
				if(numAllowinLoop + leaves.size() < numLeaves){
					leaves.add(node.getName());
				}
			}
			i++;
		}
		if(leaves.size() == 0){
			return null;
		}
		String nodeB = leaves.iterator().next();
		leaves.remove(nodeB);
		pair.setNodeA(leaves);
		pair.setNodeB(nodeB);
//		pair.setMinimalSharedParent("Multi");

		return pair;
	}

	private static boolean isInLoop(Node node) {
		Set<Block> ances = new THashSet<Block>();
		getAncestors(node, ances);
		
		for(Block parent : ances){
			if (parent instanceof Block.XorLoop){
				return true;
			}
		}		
		return false;
	}

	private static void getAncestors(Node node, Set<Block> ances) {
		for(Block b : node.getParents()){
			ances.add(b);
			getAncestors(b, ances);			
		}		
	}

	public static ExpImprLabelPair<String> getRandomDupTaskPair(ProcessTree tree,
			Type type) {
		ExpImprLabelPair<String> pair = new ExpImprLabelPair<String>();

		// Get all nodes that have the input type
		List<Node> selectedNodes = new ArrayList<Node>();
		for (Node n : tree.getNodes()) {
			if (type.equals(tree.getType(n))) {
				selectedNodes.add(n);
			}
		}
		// Skip if no such nodes 
		if (selectedNodes.size() == 0) {
			return null;
		}

		List<Integer> randomNodeIndices = UtilMathRandom.getRandomUniqueInt(
				selectedNodes.size(), selectedNodes.size());

		for (int index = 0; index < randomNodeIndices.size(); index++) {
			Block selectedNode = (Block) selectedNodes.get(randomNodeIndices.get(index));

			List<Integer> randomChildrenIndex = UtilMathRandom.getRandomUniqueInt(
					selectedNode.getChildren().size(), 2);
			List<String> pathToA = new ArrayList<String>();
			getRandomPath(selectedNode.getChildren().get(randomChildrenIndex.get(0)),
					pathToA);
			List<String> pathToB = new ArrayList<String>();
			getRandomPath(selectedNode.getChildren().get(randomChildrenIndex.get(1)),
					pathToB);
			List<String> pathToRoot = new ArrayList<String>();
			getPathToRoot(selectedNode, pathToRoot);

			// If no path to a visible transition found, try next node with type.  
			if (pathToA.size() == 0 || pathToB.size() == 0) {
				continue;
			} else {
				pair.setNodeA(pathToA.get(pathToA.size() - 1));
				pair.setPathToA(pathToA.get(pathToA.size() - 1), pathToA);
				pair.setNodeB(pathToB.get(pathToB.size() - 1));
				pair.setPathToB(pathToB);
				pair.setMinimalSharedParent(type.toString());
				pair.setPathToRoot(pathToRoot);
				break;
			}
		}
		print(pair);
		if (pair.getMinimalSharedParent() == null) {
			return null;
		}

		return pair;
	}

	public static void print(ExpImprLabelSetI<String> pair) {
		PrintStream writer = System.out;
		String[] sentences = getStrings(pair);

		for (String s : sentences) {
			writer.println(s);
		}
	}

	public static String[] getStrings(ExpImprLabelSetI<String> pair) {
		if (pair == null) {
			return new String[] { "", "", "", "", "", "" };
		}
		String[] sentences = new String[6];
		sentences[0] = StringUtils.join(pair.getNodeAs(), ",");
		sentences[1] = pair.getNodeB();
		sentences[2] = pair.getMinimalSharedParent();
		sentences[3] = StringUtils.join(pair.getPathToA(), ">");
		sentences[4] = StringUtils.join(pair.getPathToB(), ">");
		sentences[5] = StringUtils.join(pair.getPathToRoot(), ">");
		return sentences;
	}

	private static void getPathToRoot(Block selectedNode, List<String> pathToParent) {
//		Assert.assertTrue(selectedNode.getParents().size() <= 1);
		for (Block parent : selectedNode.getParents()) {
			getPathToRoot(parent, pathToParent);
			pathToParent.add(parent.toStringShort());
		}
	}

	private static boolean getRandomPath(Node node, List<String> pathToA) {
		if (node instanceof Manual) {
			pathToA.add(((Manual) node).getName());
			return true;
		} else if (node instanceof Automatic) {
			return false;
		}

		if (node instanceof Block) {

			// Get a random child and recurse.
			Block b = (Block) node;
			List<Integer> indices = UtilMathRandom.getRandomUniqueInt(b.getChildren()
					.size());
			int i = 0;
			while (i < indices.size()) {
				Node randonChild = b.getChildren().get(indices.get(i));

				boolean isMaunal = getRandomPath(randonChild, pathToA);
				if (isMaunal) {
					pathToA.add(0, node.toStringShort());
					return true;
				}
				i++;
			}
		}

		return false;
	}

}
