package org.processmining.xlu.relabel.experiment;

import gnu.trove.set.hash.THashSet;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;

public class ExpMatchMappingWriter {

	public static void __testMappings(EGraphsWithMappings graphsWithMappings,
			Map<MatchMapping, Integer> mapToNewCost, ExpRelabelingParam param,
			ExpImprLabelSetI<String> mapOrg2Dup, String gloOrLoc) {
		
		String file =  param.getFileDir()
		+ "mappingDump/" + param.getModelName() + "_" + param.getLogNum() + "_" + mapOrg2Dup.getMinimalSharedParent() + "_" + gloOrLoc + "_" + param.getRunday();
	
//		new File(fileDirectory).mkdirs();
		String fileName = file + "_raw.csv";
		String fileOverviewName = file + "_overview.csv";
		PrintStream writer = null, overview = null;
		try {
			writer = new PrintStream(new File(fileName));
			overview = new PrintStream(new File(fileOverviewName));
		} catch (IOException e) {
			e.printStackTrace();
		}

		writer.println("trcI;trcJ;evtI;evtJ;Labeli;Labelj;TransI;TransJ;Correct;cMatch;cNoMatch;cGlob");
		overview.println("ti;tj;numWrong;oldCost;newCost");
		CostStrategy costfunction = graphsWithMappings.getMatchConfiguration()
				.getCostStrategy();

		MatchMappings mappings = graphsWithMappings.getMappings();
		int size = graphsWithMappings.getPlog().size();
		XLog log = graphsWithMappings.getPlog().getXLog();
		List<Object[]> records = new ArrayList<Object[]>();

		for (int ti = 0; ti < size; ti++) {
			for (int tj = ti + 1; tj < size; tj++) {

				int numWrong = 0;

				MatchMapping mapping = mappings.getAnOptimalMapping(ti, tj);

				// testing single mapping
				XTrace xiTrace = log.get(ti);
				XTrace xjTrace = log.get(tj);
				for (Entry<Integer, Integer> match : mapping.getMatchs()) {
					if (match.getKey() < xiTrace.size()
							&& match.getValue() < xjTrace.size()) {
						XEvent ei = xiTrace.get(match.getKey());
						XEvent ej = xjTrace.get(match.getValue());

						String eilabel = __getOriginalLabel(ei);
						String ejlabel = __getOriginalLabel(ej);

						Integer costMatch = costfunction.getWeightedCostofMatchedPair(
								match.getKey(), match.getValue(), mapping);
						Integer costNoMatch = 0;
						Integer costGlob = costfunction.getWeightedCostOfGlobalStructure(
								match.getKey(), match.getValue(), mapping);

						boolean isCorrect = eilabel.equals(ejlabel);
						numWrong += isCorrect ? 0 : 1;
						Object[] record = new Object[] { ti, tj, match.getKey(),
								match.getValue(),
								XConceptExtension.instance().extractName(ei),
								XConceptExtension.instance().extractName(ej), eilabel,
								ejlabel, isCorrect, costMatch, costNoMatch, costGlob };
						records.add(record);
					}
				}

				Set<String> tiNoMatchOrgLabels = __getOrgLabels(xiTrace,
						mapping.getSourcesToNull());
				Set<String> tjNoMatchOrgLabels = __getOrgLabels(xjTrace,
						mapping.getTargetToNull());

				for (Integer ei : mapping.getSourcesToNull()) {
					String eilabel = __getOriginalLabel(xiTrace.get(ei));
					boolean isCorrect = !tjNoMatchOrgLabels.contains(eilabel);
					numWrong += isCorrect ? 0 : 1;

					Integer costMatch = 0;
					Integer costNoMatch = costfunction.getWeightedCostNoMatch(ei,
							mapping.getSourceGraph(), mapping);
					Integer costGlob = 0;

					Object[] record = new Object[] { String.valueOf(ti),
							String.valueOf(tj), String.valueOf(ei), "null",
							XConceptExtension.instance().extractName(xiTrace.get(ei)),
							"null", eilabel, "null", String.valueOf(isCorrect),
							costMatch, costNoMatch, costGlob };
					records.add(record);
				}

				for (Integer ej : mapping.getTargetToNull()) {
					String ejlabel = __getOriginalLabel(xjTrace.get(ej));
					boolean isCorrect = !tiNoMatchOrgLabels.contains(ejlabel);
					numWrong += isCorrect ? 0 : 1;

					Integer costMatch = 0;
					Integer costNoMatch = costfunction.getWeightedCostNoMatch(ej,
							mapping.getTargetGraph(), mapping);
					Integer costGlob = 0;

					Object[] record = new Object[] { ti, String.valueOf(tj), "null",
							String.valueOf(ej), "null",
							XConceptExtension.instance().extractName(xjTrace.get(ej)),
							"null", ejlabel, String.valueOf(isCorrect), costMatch,
							costNoMatch, costGlob };
					records.add(record);
				}

				overview.println(ti + ";" + tj + ";" + numWrong + ";" + mapping.getCost()
						+ ";" + mapToNewCost.get(mapping));
			}
		}

		for (Object[] record : records) {
			writer.println(org.apache.commons.lang3.StringUtils.join(record, ";"));
		}
		writer.close();
		overview.close();
	}

	private static Set<String> __getOrgLabels(XTrace xTrace, Set<Integer> sourcesToNull) {
		Set<String> set = new THashSet<String>();
		for (Integer i : sourcesToNull) {
			set.add(__getOriginalLabel(xTrace.get(i)));
		}
		return set;
	}

	private static String __getOriginalLabel(XEvent ei) {
		XAttributeLiteral a = (XAttributeLiteral) ei.getAttributes().get(
				ExpDuplicatedTaskUtil.ORG_LABEL);
		return a.getValue();
	}



}
