package org.processmining.xlu.relabel.experiment;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.out.XesXmlGZIPSerializer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.providedobjects.ProvidedObjectID;
import org.processmining.framework.util.collection.HashMultiSet;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.partialorder.models.palignment.Move;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryInductiveMiner;
import org.processmining.xlu.experiment.discovery.ExpLocRefiningDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.zexperiment.EGLogDiscoveryTestPluginImpl;
import org.processmining.xlu.matching.zexperiment.ModelQuality;
import org.processmining.xlu.matching.zexperiment.precision.EnabledSet;
import org.processmining.xlu.matching.zexperiment.precision.MapTrcEnabled;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelParam;
import org.processmining.xlu.relabel.param.RelabelParamPT;
import org.processmining.xlu.relabel.plugin.RelabelGeneralPlugin;
import org.processmining.xlu.relabel.plugin.detection.ExpILDetectorOracleGlobal;
import org.processmining.xlu.relabel.plugin.detection.ExpILDetectorOracleLocal;
import org.processmining.xlu.relabel.plugin.detection.ILDetector;
import org.processmining.xlu.relabel.plugin.detection.RelabelScope;
import org.processmining.xlu.relabel.plugin.hyber.RelabelHyberApproachIM;
import org.processmining.xlu.relabel.plugin.matching.ModelCreator;

import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;
import junit.framework.Assert;

public class ExpRelabelingAlg {

	public static void runExperimentDuplicatedLabelRelabeling(PluginContext context,
			XLog origLog, ExpRelabelingParam param) throws IOException {

		ExpDiscoveryAlgorithm[] algs = param.getDiscoveryAlgs();

		PrintStream outWriter = param.getOutWriter();

		// Log 1
		XLog log1 = (XLog) origLog.clone();
		XEventClassifier classifier = new XEventNameClassifier();
		XEventClasses classes = XEventClasses.deriveEventClasses(classifier, log1);
		// Log 2
//		Set<ExpDupTaskPair<String>> pairs = param.getAllDuplicationMap(classes);
		List<ExpImprLabelPair<String>> pairs = param.getDupPairs();
//		ExpDupTaskPair<String> mapSourceToTarget = param.getDupPair();
//		ExpDupTaskPair<String> mapSourceToTarget = param.getDuplicationMap(classes);

//		printDuplicatedMapping(outWriter, mapSourceToTarget);

		for (ExpImprLabelSetI<String> mapOrg2Dup : pairs) {
			param.setDetected(true);

			XLog log2Dup = ExpDuplicatedTaskUtil.addDuplicatedLabels(log1, classes,
					mapOrg2Dup);
//			context.getProvidedObjectManager().createProvidedObject("Log2", log2Dup,
//					XLog.class, context);
			int numEvtClassesLabelLog = XEventClasses.deriveEventClasses(classifier,
					log2Dup).size();

			serializeLog(param, mapOrg2Dup, log2Dup, "LogD");

			ExpDuplicatedTaskUtil.print(mapOrg2Dup);

			// Log 3
			RelabelParamPT relabelParam = RelabelParamPT.createGlobalDefaultParam();
//			relabelParam.getRelabelerParam().setMatchingConfig(
//					param.getTraceMatchingConfiguration());

			RelabelScope<?, ?> scope = null;
			boolean isUsingDetectorOracle = param.isUsingOracle();
			if (isUsingDetectorOracle) {
				ExpILDetectorOracleGlobal detector = new ExpILDetectorOracleGlobal();
//			XLog log3R = RelabelGeneralPlugin.relabelLogGeneral(log2Dup, relabelParam);
				detector.addImpreciseLabel(mapOrg2Dup.getNodeB());
				scope = detector.detectImpreciseLabels(log2Dup,
						relabelParam.getDetectorParam());
			} else {
				ILDetector<?, ?> detector = relabelParam.getDetector();
				scope = detector.detectImpreciseLabels(log2Dup,
						relabelParam.getDetectorParam());
			}

			if (scope.getLabelScope() == null || scope.getLabelScope().isEmpty()) {
				param.setDetected(false);
			}

			ModelCreator algCreator = relabelParam.getRelabeler();
			RelabelModel model = algCreator.createRelabelModel(log2Dup,
					scope.getLabelScope(), relabelParam);

//			ExpMatchMappingWriter.__testMappings(model.getMappings(), model.getMappingsInfo().getMapToCost(), param, mapOrg2Dup, "glo");

			OverviewMap resOverview = new OverviewMap();
			resOverview.put("M0_gen", param.getSystemModelQuality());

			ModelQuality imAlgLB = null;
			ModelQuality imAlgUB = null;
			for (ExpDiscoveryAlgorithm alg : algs) {
//				relabelParam.getRelabelParam().reset();
				/* -----------------------------------------------
				 * M1: Run discovery on the original log and compute model quality.
				 * ------------------------------------------------*/
				MarkedPetrinet net1 = alg.minePetrinet(context, log1, classifier);

//				addNetToContext(context, net1, "Net1");
				writeModelImage(context, param, mapOrg2Dup, alg, net1, "M1");

				ModelQuality qualityL1M1 = computeQuality(context, classifier, log1, net1);

				/* --------------------------------------------------------------
				 * M2: Add duplicated tasks and run exp.
				 * ------------------------------------------------------------- */
				MarkedPetrinet net2 = alg.minePetrinet(context, log2Dup, classifier);
//				addNetToContext(context, net2, "Net2");
				writeModelImage(context, param, mapOrg2Dup, alg, net2, "M2");

				ModelQuality qualityL2M2 = computeQuality(context, classifier, log2Dup,
						net2);
				if (alg instanceof ExpDiscoveryInductiveMiner) {
					imAlgLB = qualityL2M2;
					imAlgUB = qualityL1M1;
				}

				/* --------------------------------------------------------------
				 * Adapt the model of quality2 and use origLog L1 to assess the model 
				 * ------------------------------------------------------------- */
//				getModifiedModelAndUpdateMappinng(net2, mapOrg2Dup);

//				ModelQuality qualityL1M2 = computeQuality(context, classifier, origLog,
//						net2);

				/* --------------------------------------------------------------
				 * TODO Use origLog to assess the model of quality3
				 * - need a more complex mapping creator ... Hmm much complex..
				 * ------------------------------------------------------------- */

				/* --------------------------------------------------------------
				 * Use L2 to assess Model 1
				 * ------------------------------------------------------------- */
//				modifiedModel1ToL2(net1, mapOrg2Dup);
//				ModelQuality qualityL2M1 = computeQuality(context, classifier, log2Dup,
//						net1);
				ModelQuality qualityL2M1 = qualityL1M1;

				/* --------------------------------------------------------------
				 * M3. Run our relabeling.
				 * ------------------------------------------------------------- */
				XLog log3R = algCreator.applyRelabeling(model, relabelParam);
//				createAndExportDFGs(context, param, mapOrg2Dup, log3R, "glo");				

				MarkedPetrinet net3 = alg.minePetrinet(context, log3R, classifier);
//				addNetToContext(context, net3, "Net3");

//				ModelQuality qualityL3M3 = computeQuality(context, classifier, log3R,
//						net3);

//				context.getProvidedObjectManager().createProvidedObject("Dup.Mapping",
//						graphsWithMappings, EGraphsWithMappings.class, context);

				/* --------------------------------------------------------------
				 * TODO Use L2 to assess Model 3
				 * ------------------------------------------------------------- */
				modifiedModel3ToL2(net3);

				ModelQuality qualityL2M3 = computeQuality(context, classifier, log2Dup,
						net3);

				boolean found = false;
				if (alg instanceof ExpDiscoveryInductiveMiner) {
					loop: for (int edgeCostThreshold = 101; edgeCostThreshold >= 0; edgeCostThreshold -= 5) {
						for (int unfthreshold = 0; unfthreshold < 60; unfthreshold += 10) {

//				loop: for (int edgeCostThreshold = 8; edgeCostThreshold >= 0; edgeCostThreshold--) {
//					for (int unfthreshold = 0; unfthreshold < 60; unfthreshold += 10) {
//					edgeCostThreshold--; // testing adaptive edge cost. 
							relabelParam.getRelabelParam().setMaxCostOfEdge(
									edgeCostThreshold);
							relabelParam.getRelabelParam().setUnfoldingThreshold(
									unfthreshold);

							log3R = algCreator.applyRelabeling(model, relabelParam);
//					debugTest(log2Dup, log3R);

							int numEvtClasses = XEventClasses.deriveEventClasses(
									relabelParam.getClassifier(), log3R).size();
							if (numEvtClasses < numEvtClassesLabelLog
									+ scope.getLabelScope().size() * 10) {
								net3 = alg.minePetrinet(context, log3R, classifier);
								modifiedModel3ToL2(net3);

								if (net3.getPetrinet().getTransitions().size() < 40) {
//						addNetToContext(context, net3, "Net3");
									qualityL2M3 = computeQuality(context, classifier,
											log2Dup, net3);
								}
//					}

								if (qualityL2M3.precision > qualityL2M2.precision) {

									found = true;
									break loop;
								}
							}
						}
					}
				}
				if (!found) {
					relabelParam.getRelabelParam().reset();
				}
				serializeLog(param, mapOrg2Dup, log3R, "LogR_" + alg.toString());
				writeModelImage(context, param, mapOrg2Dup, alg, net3, "M3");

				if (qualityL2M3.precision <= qualityL2M2.precision) {

//					qualityL2M2.setImproved(false);
//					qualityL2M3.precision = qualityL2M2.precision;
//					qualityL2M3.recallWrtMGen = qualityL2M2.recallWrtMGen;
//					qualityL2M3.precWrtMGen = qualityL2M2.precWrtMGen;
					qualityL2M3 = qualityL2M2;
					qualityL2M3.setImproved(false);
				}
//				addNetToContext(context, net3, "Net3forL2");

				computePrecisionRecall(origLog, "M1", mapOrg2Dup, alg, qualityL2M1, param);
				computePrecisionRecall(origLog, "M2", mapOrg2Dup, alg, qualityL2M2, param);
				computePrecisionRecall(origLog, "M3", mapOrg2Dup, alg, qualityL2M3, param);

				computeBaseLine(qualityL2M1, qualityL2M2, qualityL2M3);

				resOverview.put("M1_org", alg, qualityL2M1);
				resOverview.put("M2_impr", alg, qualityL2M2);
				resOverview.put("M3_ref", alg, qualityL2M3);

//				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M1_org;L1_org",
//						qualityL1M1, alg.toString(), outWriter, param);
//				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M2_impr;L1_org",
//						qualityL1M2, alg.toString(), outWriter, param);
				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M1_org" + ";"
						+ "L2_impr", qualityL2M1, alg.toString(), outWriter, param,
						relabelParam);

////				printPrecisionRecall(log2Dup, "M1", mapOrg2Dup, alg, qualityL2M1, param);
				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M2_impr" + ";"
						+ "L2_impr", qualityL2M2, alg.toString(), outWriter, param,
						relabelParam);
////				printPrecisionRecall(log2Dup, "M2", mapOrg2Dup, alg, qualityL2M2, param);
//
				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M3_ref" + ";"
						+ "L2_impr", qualityL2M3, alg.toString(), outWriter, param,
						relabelParam);
////				printPrecisionRecall(log2Dup, "M3", mapOrg2Dup, alg, qualityL2M3, param);
//
//				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M3_ref;L3_ref",
//						qualityL3M3, alg.toString(), outWriter, param);
////				printPrecisionRecallOverview(qualityL2M1, qualityL2M2, qualityL2M3);
			}

			for (ExpLocRefiningDiscoveryAlgorithm alg : param.getHyberRelabelAlgs()) {
				RelabelParamPT relabelLocalParam = RelabelParamPT
						.createLocalDefaultParam();
//				relabelParam.getRelabelerParam().setMatchingConfig(
//						param.getTraceMatchingConfiguration());

				if (isUsingDetectorOracle) {
					ExpILDetectorOracleLocal detector = new ExpILDetectorOracleLocal();
					detector.addImpreciseLabel(mapOrg2Dup.getNodeB());
					relabelLocalParam.setDetector(detector);
					alg.setRelabelParam(relabelLocalParam);
				}

				MarkedPetrinet net4 = alg.minePetrinet(context, log2Dup, classifier);

				XLog loclog = alg.getLocalRefinedLog();
				RelabelModel rmodel = alg.getRelabelModel();
//				ExpMatchMappingWriter.__testMappings(rmodel.getMappings(), rmodel.getMappingsInfo().getMapToCost(), param, mapOrg2Dup, "loc");

//				createAndExportDFGs(context, param, mapOrg2Dup, loclog, "loc");

//				context.getProvidedObjectManager().createProvidedObject("Net4",
//						, Petrinet.class, context);

//				addNetToContext(context, net4, "Net4");

				writeModelImage(context, param, mapOrg2Dup, alg, net4, "M4");

				ModelQuality qualityL2M4 = computeQuality(context, classifier, log2Dup,
						net4);
				if (alg instanceof RelabelHyberApproachIM
						&& qualityL2M4.precision <= imAlgLB.precision) {

					qualityL2M4 = imAlgLB;
					qualityL2M4.setImproved(false);
				}

				computePrecisionRecall(origLog, "M4", mapOrg2Dup, alg, qualityL2M4, param);

				computeBaseLine(imAlgUB, imAlgLB, qualityL2M4);

				resOverview.put("M4_loc", alg, qualityL2M4);

				ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M4_loc" + ";"
						+ "L2_impr", qualityL2M4, alg.toString(), outWriter, param,
						relabelLocalParam);
////				printPrecisionRecall(log2Dup, "M4", mapOrg2Dup, alg, qualityL2M4, param);

			}
//			printQuality(log2Dup, mapOrg2Dup, resOverview, outWriter, param);
			printPrecisionRecall(log2Dup, mapOrg2Dup, resOverview, param);
		}
	}

	private static void debugTest(XLog log2Dup, XLog log3R) {
		Assert.assertEquals(log3R.size(), log2Dup.size());
		for (int i = 0; i < log3R.size(); i++) {
			Assert.assertEquals(log3R.get(i).size(), log2Dup.get(i).size());

			for (int j = 9; j < log3R.get(i).size(); j++) {
				String newlabel = XConceptExtension.instance().extractName(
						log3R.get(i).get(j));
				int indexOf = newlabel.indexOf("_X_");
				if (indexOf >= 0) {
					newlabel = newlabel.substring(0, indexOf);
				}
				String oldlabel = XConceptExtension.instance().extractName(
						log2Dup.get(i).get(j));
				Assert.assertEquals(newlabel, oldlabel);
			}
		}
	}

	private static void computeBaseLine(ModelQuality qualityL2M1,
			ModelQuality qualityL2M2, ModelQuality qualityL2M3) {

		qualityL2M3.properties.put(ModelQuality.LOG_PREC_LB,
				String.valueOf(qualityL2M2.precision));
		qualityL2M3.properties.put(ModelQuality.LOG_PREC,
				String.valueOf(qualityL2M3.precision - qualityL2M2.precision));
		qualityL2M3.properties.put(ModelQuality.LOG_PREC_TO_UB,
				String.valueOf(qualityL2M1.precision - qualityL2M3.precision));

		qualityL2M3.properties.put(ModelQuality.SYS_PREC_LB,
				String.valueOf(qualityL2M2.precWrtMGen));
		qualityL2M3.properties.put(ModelQuality.SYS_PREC_TO_LB,
				String.valueOf(qualityL2M3.precWrtMGen - qualityL2M2.precWrtMGen));
		qualityL2M3.properties.put(ModelQuality.SYS_PREC_TO_UB,
				String.valueOf(qualityL2M1.precWrtMGen - qualityL2M3.precWrtMGen));
	}

	private static void createAndExportDFGs(PluginContext context,
			ExpRelabelingParam param, ExpImprLabelSetI<String> mapOrg2Dup, XLog log,
			String type) {
//
//		IMLog2IMLogInfo info = new IMLog2IMLogInfoDefault();
//
//		XEventAttributeClassifier org = new XEventAttributeClassifier("0OrgLabel",
//				ExpDuplicatedTaskUtil.ORG_LABEL);
//		exportDFG(param, log, info, org, type, mapOrg2Dup);
//
//		XEventAttributeClassifier impr = new XEventAttributeClassifier("1Imprecise",
//				ModelCreatorUsingMatching.IMPRECISE_LABEL);
//		exportDFG(param, log, info, impr, type, mapOrg2Dup);
//
//		XEventAttributeClassifier ref = new XEventAttributeClassifier("2Refined",
//				"concept:name");
//		exportDFG(param, log, info, ref, type, mapOrg2Dup);
	}

//	private static void exportDFG(ExpRelabelingParam param, XLog log,
//			IMLog2IMLogInfo info, XEventAttributeClassifier activityClassifier,
//			String type, ExpImprLabelSetI<String> mapOrg2Dup) {
//		Dfg dfg = info.createLogInfo(new IMLog(log, activityClassifier)).getDfg();
//		Dot dot = GraphvizDirectlyFollowsGraphClone.dfg2Dot(dfg, true);
//		DotPanel panel = new DotPanel(dot);
//		dot = panel.getDot();
//		Dot2Image.dot2image(
//				dot,
//				new File(param.getFileDir() + "images/_dfg_" + param.getModelName() + "_"
//						+ param.getLogNum() + "_" + mapOrg2Dup.getMinimalSharedParent()
//						+ "_" + type + "_" + activityClassifier.name() + "_"
//						+ param.getRunday() + ".png"),
//				org.processmining.plugins.graphviz.dot.Dot2Image.Type.png);
//	}

	private static void printQuality(XLog log2Dup, ExpImprLabelSetI<String> mapOrg2Dup,
			OverviewMap resOverview, PrintStream outWriter, ExpRelabelingParam param) {
//		for(alg : resOverview.getListOfAlgs()){
//			ExpWriterRelabelingUtil.printQuality(mapOrg2Dup, "M4_loc;L2_impr",
//					qualityL2M4, alg, outWriter, param);
//		}

	}

	private static void computePrecisionRecall(XLog log, String string,
			ExpImprLabelSetI<String> mapOrg2Dup, ExpDiscoveryAlgorithm alg,
			ModelQuality quality, ExpRelabelingParam param) {

		MapTrcEnabled systemMap = param.getSystemModelQuality().getMapTrcEnable();

		int sumAllSystem = 0;
		int sumAllModel = 0;
		int sumSystemAndModel = 0;
		Map<String, MultiSet<String>> mapEvtClass2Set = getEmptyMap(log);

		for (int trc = 0; trc < log.size(); trc++) {
			int size = log.get(trc).size();
			// get list of events. 
			List<EnabledSet> systemList = systemMap.getEnabledTraceList(trc, size);
			List<EnabledSet> modelList = quality.getMapTrcEnable().getEnabledTraceList(
					trc, size);

			for (int i = 0; i < size; i++) {

//				Set<String> systemSet = getLabels(systemList.get(i), mapOrg2Dup);
//				sumAllSystem += systemSet.size();

//				Set<String> modelSet = getLabels(modelList.get(i), mapOrg2Dup);
//				sumAllModel += modelSet.size();

//				systemSet.retainAll(modelSet);
//				sumSystemAndModel += systemSet.size();

//				String label = XConceptExtension.instance().extractName(
//						log.get(trc).get(i));

				Set<String> systemSet = getLabels(systemList.get(i), mapOrg2Dup);
//				sumAllSystem += systemSet.size();
				sumAllSystem += systemSet.size();

				Set<String> modelSet = getLabels(modelList.get(i), mapOrg2Dup);
//				sumAllModel += modelSet.size();

				systemSet.retainAll(modelSet);
				sumSystemAndModel += systemSet.size();
				sumAllModel += modelSet.size();
//				int numOverflow = modelSet.size() - systemSet.size();

//				record.add(String.valueOf(numMissing));
//				record.add(String.valueOf(numOverflow));

			}
		}
		quality.precWrtMGen = (double) sumSystemAndModel / sumAllModel;
		quality.recallWrtMGen = (double) sumSystemAndModel / sumAllSystem;
//		quality.mapEvtClass2Set = mapEvtClass2Set;

	}

	private static Map<String, MultiSet<String>> getEmptyMap(XLog log) {
		XEventClassifier c = new XEventNameClassifier();
		XEventClasses cs = XEventClasses.deriveEventClasses(c, log);

		Map<String, MultiSet<String>> map = new THashMap<String, MultiSet<String>>();
		for (XEventClass cl : cs.getClasses()) {
			map.put(cl.toString(), new HashMultiSet<String>());
		}
		return map;
	}

	private static Set<String> getLabels(EnabledSet enabledSet,
			ExpImprLabelSetI<String> mapOrg2Dup) {
		Set<String> labels = new THashSet<String>();
//		if (enabledSet != null) {
		for (Transition t : enabledSet.getVisibleTransitions()) {

			if (mapOrg2Dup.getNodeAs().contains(t.getLabel())) {
				labels.add(mapOrg2Dup.getNodeB());
			} else {
				labels.add(t.getLabel());
			}
		}
//		}
		return labels;
	}

	private static void printPrecisionRecall(XLog log2Dup,
			ExpImprLabelSetI<String> mapOrg2Dup, OverviewMap resOverview,
			ExpRelabelingParam param) {
//		PrintStream writer = param.getRPWriter();
		String minimalSharedParent = mapOrg2Dup == null ? "" : mapOrg2Dup
				.getMinimalSharedParent();
		PrintStream writer = null;
		try {
			writer = new PrintStream(new File(param.getFileDir() + "precsv/_ov_"
					+ param.getModelName() + "_" + param.getLogNum() + "_"
					+ minimalSharedParent + "_" + param.getRunday() + ".csv"));
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		String columnNames = "Trc;evt;label;orgLabel";
//		for(String alg :  resOverview.getListOfAlgs()){
		for (int i = 0; i < resOverview.getListOfAlgs().size(); i++) {
			String algName = resOverview.getListOfAlgs().get(i);
			columnNames += ";" + algName;
			if (i > 0) {
				columnNames += ";" + "miss_" + algName;
				columnNames += ";" + "add_" + algName;
			}
		}
		writer.println(columnNames);
//		MapTrcEnabled map = qualityL2M1.getMapTrcEnable();
//		
		for (int trc = 0; trc < log2Dup.size(); trc++) {

			List<List<EnabledSet>> listOfTrcPosSet = resOverview.getListOfListEnabledSet(
					trc, log2Dup.get(trc).size());

//			Map<Move, EnabledSet> mmap = map.get(trc);

			for (int evt = 0; evt < log2Dup.get(trc).size(); evt++) {
				List<String> record = new ArrayList<String>();

				record.add(String.valueOf(trc));
				record.add(String.valueOf(evt));
				record.add(XConceptExtension.instance().extractName(
						log2Dup.get(trc).get(evt)));
				record.add(((XAttributeLiteral) log2Dup.get(trc).get(evt).getAttributes()
						.get(ExpDuplicatedTaskUtil.ORG_LABEL)).getValue());

//				for (List<EnabledSet> trcPosSet : listOfTrcPosSet) {
				for (int i = 0; i < listOfTrcPosSet.size(); i++) {

					List<EnabledSet> trcPosSet = listOfTrcPosSet.get(i);
					if (isLogMove(trcPosSet.get(evt))) {
						continue;
					}
					record.add(trcPosSet.get(evt).getVisibleSetToString());
					if (i > 0) {
						// add missing and overflowing columns. 
						Set<String> systemSet = getLabels(
								listOfTrcPosSet.get(0).get(evt), mapOrg2Dup);
//						sumAllSystem += systemSet.size();
						int sizeSystem = systemSet.size();

						Set<String> modelSet = getLabels(listOfTrcPosSet.get(i).get(evt),
								mapOrg2Dup);
//						sumAllModel += modelSet.size();

						systemSet.retainAll(modelSet);
						int numMissing = sizeSystem - systemSet.size();
						int numOverflow = modelSet.size() - systemSet.size();

						record.add(String.valueOf(numMissing));
						record.add(String.valueOf(numOverflow));
					}
				}
				writer.println(StringUtils.join(record, ";"));
			}
		}
		writer.close();
	}

	private static boolean isLogMove(EnabledSet enabledSet) {
		return enabledSet == null;
	}

	private static void printPrecisionRecall(XLog log2Dup, String string,
			ExpImprLabelSetI<String> mapOrg2Dup, ExpDiscoveryAlgorithm alg,
			ModelQuality qualityL2M1, ExpRelabelingParam param) {
//		PrintStream writer = param.getRPWriter();
		String minimalSharedParent = mapOrg2Dup == null ? "" : mapOrg2Dup
				.getMinimalSharedParent();
		PrintStream writer = null;
		try {
			writer = new PrintStream(new File(param.getFileDir() + "precsv/"
					+ param.getModelName() + "_" + param.getLogNum() + "_" + string + "_"
					+ alg.toString() + "_" + minimalSharedParent + "_"
					+ param.getRunday() + ".csv"));
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		writer.println("Trc;move;evt;label;enabledset");
		MapTrcEnabled map = qualityL2M1.getMapTrcEnable();
		for (int trc = 0; trc < log2Dup.size(); trc++) {
			Map<Move, EnabledSet> mmap = map.get(trc);
			for (Entry<Move, EnabledSet> entry : mmap.entrySet()) {
				if (!entry.getKey().getTransition().isInvisible()) {
					String label = entry.getKey().getEventIndex() >= 0 ? XConceptExtension
							.instance().extractName(
									log2Dup.get(trc).get(entry.getKey().getEventIndex()))
							: (entry.getKey().getTransition() != null ? entry.getKey()
									.getTransition().getLabel() : "");
//					label = ;
					writer.println(trc + ";" + entry.getKey().getMoveIndex() + ";"
							+ entry.getKey().getEventIndex() + ";" + label + ";"
//							+ entry.getValue().toString());
							+ entry.getValue().getVisibleSetToString());
				}
			}
		}

		//==================================

		PrintStream writeroverview = null;
		try {
			writeroverview = new PrintStream(new File(param.getFileDir() + "precsv/_ov_"
					+ param.getModelName() + "_" + param.getLogNum() + "_" + string + "_"
					+ alg.toString() + "_" + minimalSharedParent + "_"
					+ param.getRunday() + ".csv"));
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		writeroverview.println("Label;PossibleSet;count");
		Map<String, MultiSet<String>> mapoverview = qualityL2M1.getMapEvtClass2Set();
		for (Entry<String, MultiSet<String>> entry : mapoverview.entrySet()) {
			for (String pos : entry.getValue().baseSet()) {
				writeroverview.println(entry.getKey() + ";" + pos + ";"
						+ entry.getValue().occurrences(pos));
			}
		}
	}

	private static void writeModelImage(PluginContext context, ExpRelabelingParam param,
			ExpImprLabelSetI<String> mapOrg2Dup, ExpDiscoveryAlgorithm alg,
			MarkedPetrinet net1, String modelType) {
		String minimalSharedParent = mapOrg2Dup == null ? "" : mapOrg2Dup
				.getMinimalSharedParent();
//		minimalSharedParent = minimalSharedParent == null ? "" : minimalSharedParent;
		ExpRelabelingPlugin.export2Image(net1, new File(param.getFileDir()
				+ "images/" + param.getModelName() + "_" + param.getLogNum() + "_"
				+ modelType + "_" + alg.toString() + "_" + minimalSharedParent + "_"
				+ param.getRunday() + ".png"));
	}

	private static void serializeLog(ExpRelabelingParam param,
			ExpImprLabelSetI<String> mapOrg2Dup, XLog log2Dup, String logType)
			throws FileNotFoundException, IOException {
		XesXmlGZIPSerializer serializer = new XesXmlGZIPSerializer();
		OutputStream output = new FileOutputStream(getLogFile(param, mapOrg2Dup, logType));
		serializer.serialize(log2Dup, output);
	}

	private static File getLogFile(ExpRelabelingParam param,
			ExpImprLabelSetI<String> mapOrg2Dup, String logType) {
		File f = new File(param.getFileDir() + "logs/" + param.getModelName() + "_"
				+ param.getLogNum() + "_" + logType + "_"
				+ mapOrg2Dup.getMinimalSharedParent() + "_" + param.getRunday()
				+ ".xes.gz");
		return f;
	}

	public static ModelQuality computeQuality(PluginContext context,
			XEventClassifier classifier, XLog log2Dup, MarkedPetrinet net3) {

		return EGLogDiscoveryTestPluginImpl
				.runReplayToGetFitnessPrecisionAndGeneralization(context, net3, log2Dup,
						classifier);
	}

	private static ProvidedObjectID addNetToContext(PluginContext context,
			MarkedPetrinet net1, String name) {
		return context.getProvidedObjectManager().createProvidedObject(name,
				net1.getPetrinet(), Petrinet.class, context);
//		return null;
	}

	public static void runRealLifeLog(PluginContext context, XLog origLog,
			ExpRelabelingParam param) {
		ExpDiscoveryAlgorithm[] algs = param.getDiscoveryAlgs();

		PrintStream outWriter = param.getOutWriter();
		ExpWriterRelabelingUtil.printColumn(outWriter);

		// Log 2
		XLog log2Dup = (XLog) origLog.clone();
		XEventClassifier classifier = new XEventNameClassifier();
		// Log 3

		EGMatchConfiguration conf = param.getTraceMatchingConfiguration();
//		EGraphsWithMappings graphsWithMappings = (new EGMatchPlugin()).replayLog(context,
//				(XLog) log2Dup.clone(), conf);
		//TODO
		RelabelParam<?, ?> relabelParam = RelabelParamPT.createGlobalDefaultParam();
		XLog log3relabeled = (new RelabelGeneralPlugin()).relabelLogGeneral(
				(XLog) log2Dup.clone(), relabelParam);

		for (ExpDiscoveryAlgorithm alg : algs) {

			/* --------------------------------------------------------------
			 * M2: Add duplicated tasks and run exp.
			 * ------------------------------------------------------------- */

			MarkedPetrinet net2 = alg.minePetrinet(context, log2Dup, classifier);
			writeModelImage(context, param, null, alg, net2, "M2");
			ModelQuality qualityL2M2 = computeQuality(context, classifier, log2Dup, net2);

			context.getProvidedObjectManager().createProvidedObject("Net2",
					net2.getPetrinet(), Petrinet.class, context);

			/* --------------------------------------------------------------
			 * M3. Run our relabeling.
			 * ------------------------------------------------------------- */

			MarkedPetrinet net3 = alg.minePetrinet(context, log3relabeled, classifier);
			ModelQuality qualityL3M3 = computeQuality(context, classifier, log3relabeled,
					net3);

			context.getProvidedObjectManager().createProvidedObject("Net3",
					net3.getPetrinet(), Petrinet.class, context);
//			context.getProvidedObjectManager().createProvidedObject("Dup.Mapping",
//					graphsWithMappings, EGraphsWithMappings.class, context);

			/* --------------------------------------------------------------
			 * TODO Use L2 to assess Model 3
			 * ------------------------------------------------------------- */
			modifiedModel3ToL2(net3);
			writeModelImage(context, param, null, alg, net3, "M3");
			ModelQuality qualityL2M3 = computeQuality(context, classifier, log2Dup, net3);
			context.getProvidedObjectManager().createProvidedObject("Net3forL2",
					net3.getPetrinet(), Petrinet.class, context);

			String m = "RealLife";
			ExpWriterRelabelingUtil.printQuality(null, m + ";M2;L2;", qualityL2M2,
					alg.toString(), outWriter, param, relabelParam);
			ExpWriterRelabelingUtil.printQuality(null, m + ";M3;L2;", qualityL2M3,
					alg.toString(), outWriter, param, relabelParam);

			ExpWriterRelabelingUtil.printQuality(null, m + ";M3;L3;", qualityL3M3,
					alg.toString(), outWriter, param, relabelParam);
		}

//		for (ExpDiscoveryAlgorithm alg : param.getHyberRelabelAlgs()) {
//			MarkedPetrinet net4 = alg.minePetrinet(context, log2Dup, classifier);
//			context.getProvidedObjectManager().createProvidedObject("Net4",
//					net4.getPetrinet(), Petrinet.class, context);
//
//			writeModelImage(context, param, null, alg, net4, "M4");
//
//			ModelQuality qualityL2M4 = computeQuality(context, classifier, log2Dup, net4);
//
//			ExpWriterRelabelingUtil.printQuality(null, "M4;L2", qualityL2M4,
//					alg.toString(), outWriter, param, );
//
//		}
	}

	public static String printDuplicatedMapping(
			Map<XEventClass, XEventClass> mapSourceToTarget) {
		String s = "";
		for (Entry<XEventClass, XEventClass> entry : mapSourceToTarget.entrySet()) {
			s += entry.getKey().toString() + " to " + entry.getValue().toString();
		}
		return s;
	}

	public static void modifiedModel3ToL2(MarkedPetrinet mnet) {
		Petrinet pnet = mnet.getPetrinet();
		Collection<Transition> transitions = new THashSet<Transition>(
				pnet.getTransitions());
		for (Transition t : transitions) {
			if (t.isInvisible()) {
				continue;
			}

			String label = t.getLabel();
			if (label.indexOf("_X_") > 0) {
				label = label.substring(0, label.indexOf("_X_"));
				Transition newTrans = pnet.addTransition(label);
				replaceOldTransWithNewTrans(pnet, t, newTrans);
			}
		}

	}

	private static void modifiedModel1ToL2(MarkedPetrinet mnet,
			ExpImprLabelSetI<String> mapSourceToTarget) {
		Petrinet pnet = mnet.getPetrinet();

		Set<Transition> ts = getTransition(pnet.getTransitions(),
				mapSourceToTarget.getNodeAs());

		for (Transition t : ts) {
			Transition newTrans = pnet.addTransition(mapSourceToTarget.getNodeB());

			replaceOldTransWithNewTrans(pnet, t, newTrans);
		}
	}

	private static void replaceOldTransWithNewTrans(Petrinet pnet, Transition t,
			Transition newTrans) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> in : pnet
				.getInEdges(t)) {
			Place source = (Place) in.getSource();
			pnet.addArc(source, newTrans);

			pnet.removeEdge(in);
		}
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> out : pnet
				.getOutEdges(t)) {
			Place target = (Place) out.getTarget();
			pnet.addArc(newTrans, target);

			pnet.removeEdge(out);
		}

		pnet.removeTransition(t);
	}

	private static Set<Transition> getTransition(Collection<Transition> transitions,
			Set<String> set) {
		Set<Transition> ts = new THashSet<Transition>();
		for (Transition t : transitions) {
			if (set.contains(t.getLabel())) {
				ts.add(t);
			}
		}
		return ts;
	}

}
