package org.processmining.xlu.relabel.experiment;

import gnu.trove.set.hash.THashSet;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryInductiveMiner;
import org.processmining.xlu.experiment.discovery.ExpLocRefiningDiscoveryAlgorithm;
import org.processmining.xlu.matching.config.EGMatchConfigurationFactory;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.matching.zexperiment.ModelQuality;
import org.processmining.xlu.relabel.plugin.hyber.RelabelHyberApproachIM;

/**
 * @author xlu
 *
 */
public class ExpRelabelingParam {
	protected PrintStream writer;
	protected PrintStream prwriter;
	protected List<ExpDiscoveryAlgorithm> existingAlgs;
	protected List<ExpLocRefiningDiscoveryAlgorithm> hyberRelabelAlgs;


	protected boolean isAddingDuplicatedTasks;
	protected List<ExpImprLabelPair<String>> dupPairs;
	protected String runday = "";

	public ExpRelabelingParam() {
		existingAlgs = new ArrayList<ExpDiscoveryAlgorithm>();
		existingAlgs.add(new ExpDiscoveryInductiveMiner());
//		existingAlgs.add(new ExpDiscoveryHyberILP());
//		existingAlgs.add(new ExpDiscoveryFodina());
//		existingAlgs.add(new ExpDiscoveryETM());
		
		hyberRelabelAlgs = new ArrayList<ExpLocRefiningDiscoveryAlgorithm>();
		RelabelHyberApproachIM expDiscoveryIMwithRelabeling = new RelabelHyberApproachIM();
//		expDiscoveryIMwithRelabeling.setTraceMatchingConfig(this.getTraceMatchingConfiguration());
//		hyberRelabelAlgs.add(expDiscoveryIMwithRelabeling);
		
		isAddingDuplicatedTasks = true;
		dupPairs = new ArrayList<ExpImprLabelPair<String>>();
//		createFileName();
	}

	public ExpRelabelingParam(String runday){
		this();
		this.runday = runday;
	}


	public String getRunday() {
		return runday;
	}
	public void setRunday(String runday) {
		this.runday = runday;
	}
	
	public boolean isAddingDuplicatedTasks() {
		return isAddingDuplicatedTasks;
	}

//	public ExpDiscoveryAlgorithm getDiscoveryAlg(){
//		if(algs == null || algs.length == 0){
//			return new ExpDiscoveryInductiveMiner();
//		}
//		return algs[0];
//	}

	public ExpDiscoveryAlgorithm[] getDiscoveryAlgs() {
		if (existingAlgs == null || existingAlgs.size() == 0) {
			return new ExpDiscoveryAlgorithm[] { new ExpDiscoveryInductiveMiner() };
		}
		return existingAlgs.toArray(new ExpDiscoveryAlgorithm[existingAlgs.size()]);
	}

	public EGMatchConfiguration getTraceMatchingConfiguration() {
//		EGMatchConfiguration config = EGMatchConfigurationFactory
//				.createDataBasedDefaultConfig();

//		EGMatchConfiguration config = EGMatchConfigurationFactory
//				.createSeqBasedDefaultConfig();
//		
		EGMatchConfiguration config = EGMatchConfigurationFactory
				.createSeqBasedGreedyMatchAsMuchAsPossibleConfig();
		
//		
//		EGMatchConfiguration config = EGMatchConfigurationFactory
//				.createTimeBasedPOGreedyConfig();

		return config;
	}
	
	public List<ExpLocRefiningDiscoveryAlgorithm> getHyberRelabelAlgs() {
		return hyberRelabelAlgs;
	}

	public void setHyberRelabelAlgs(List<ExpLocRefiningDiscoveryAlgorithm> hyberRelabelAlgs) {
		this.hyberRelabelAlgs = hyberRelabelAlgs;
	}

	public PrintStream getOutWriter() {
		if (writer == null) {
			String fileDirectory = "./dupResult/resultDupTask" + runday ;
			new File(fileDirectory).mkdirs();
			String fileName =  fileDirectory + "/res.csv";
//			writer = null;
			try {
				writer = new PrintStream(new File(fileName));
			} catch (IOException e) {
				e.printStackTrace();
			}
//			CsvWriter csvFusionWriter = new CsvWriter(writer, ';');
//			this.writers.put(file, csvFusionWriter);
		}
		return writer;
	}

	/* ----------------------------------------------------------------
	 *
	 *   Duplicated Pairs
	 *
	 * ---------------------------------------------------------------- */

	public ExpImprLabelSetI<String> getDupPair() {
		return dupPairs.get(0);
	}

	public void setDupTaskPair(ExpImprLabelPair<String> pair) {
		this.dupPairs.add(pair);
	}
	
	public List<ExpImprLabelPair<String>> getDupPairs() {
		return this.dupPairs;
	}

	public Set<ExpImprLabelPair<String>> getAllDuplicationMap(XEventClasses classes) {
		Set<ExpImprLabelPair<String>> mapsSourceToTarget = new THashSet<ExpImprLabelPair<String>>();
		mapsSourceToTarget.add(getDuplicationMap(classes));

//		int i = 0;
//		for (int i = 0; i < classes.size(); i++) {
//		for (int j = i + 1; j < classes.size(); j++) {
//
//			Map<XEventClass, XEventClass> mapSourceToTarget = new THashMap<XEventClass, XEventClass>();
//
//			XEventClass targetClass = classes.getByIndex(i);
//			XEventClass sourceClass = classes.getByIndex(j);
//			mapSourceToTarget.put(sourceClass, targetClass);
//			mapsSourceToTarget.add(mapSourceToTarget);
//		}
//		}
		return mapsSourceToTarget;
	}

	public ExpImprLabelPair<String> getDuplicationMap(XEventClasses classes) {
//		Map<XEventClass, XEventClass> mapSourceToTarget = new THashMap<XEventClass, XEventClass>();

//		// Fix one mapping for dup.
		XEventClass targetClass = classes.getByIdentity("C");
		XEventClass sourceClass = classes.getByIdentity("H");
//		mapSourceToTarget.put(sourceClass, targetClass);

		// Get one random mapping
//		int source = (int) (Math.random() * classes.size());
//		int target = (int) (Math.random() * classes.size());
//		while (target == source) {
//			target = (int) (Math.random() * classes.size());
//		}
//		XEventClass targetClass = classes.getByIndex(source);
//		XEventClass sourceClass = classes.getByIndex(target);
////		mapSourceToTarget.put(sourceClass, targetClass);

		ExpImprLabelPair<String> pair = new ExpImprLabelPair<String>(sourceClass.toString(),
				targetClass.toString());
		return pair;
	}

	public void setDupTaskPairs(Collection<ExpImprLabelPair<String>> pairs) {
		this.dupPairs.addAll(pairs);
		
	}

	private String modelName;
	public String getModelName() {
		return modelName;
	}

	public void setModelName(String modelName) {
		this.modelName = modelName;
	}

	public String getLogNum() {
		return logNum;
	}

	public void setLogNum(String logNum) {
		this.logNum = logNum;
	}

	private String logNum;

	public void setWriter(PrintStream writer) {
		this.writer = writer;
		
	}

	private ModelProperties modelProperties;
	private LogProperties logProperties;
	private String fileDir;
	public String getFileDir() {
		return fileDir;
	}



	public LogProperties getLogProperties() {
		if(logProperties == null){
			logProperties = new LogProperties();
		}
		return logProperties;
	}

	public void setModelProperties(ModelProperties modelProperties) {
		this.modelProperties = modelProperties;
	}
	public ModelProperties getModelProperties(){
		if(modelProperties == null){
			modelProperties = new ModelProperties();
		}
		return modelProperties;
	}

	public void setLogProperties(LogProperties logProperties) {
		this.logProperties = logProperties;
		
	}

	public void setDirectory(String fileDirectory) {
		this.fileDir = fileDirectory;
		
	}
	
	private boolean isDetected = false;
	private ModelQuality systemModelQuality;
	private boolean isUsingOracle;
	private boolean isAllowInLoop;
	private int numOfImprLeaves = 4;

	public void setNumOfImprLeaves(int numOfImprLeaves) {
		this.numOfImprLeaves = numOfImprLeaves;
	}

	public Boolean isDetected() {
		return isDetected;
	}

	public void setDetected(boolean b) {
		isDetected = b;
	}

	public PrintStream getRPWriter() {
			if (prwriter == null) {
				String fileDirectory = "./dupResult/resultDupTask" + runday ;
				new File(fileDirectory).mkdirs();
				String fileName =  fileDirectory + "/res.csv";
//				writer = null;
				try {
					prwriter = new PrintStream(new File(fileName));
				} catch (IOException e) {
					e.printStackTrace();
				}
//				CsvWriter csvFusionWriter = new CsvWriter(writer, ';');
//				this.writers.put(file, csvFusionWriter);
			}
			return prwriter;
		
	}

	public void setSystemBaseSet(ModelQuality mapTrcEnable) {
		this.systemModelQuality = mapTrcEnable;
	}

	public ModelQuality getSystemModelQuality() {
		return systemModelQuality;
	}

	public boolean isUsingOracle() {
		return this.isUsingOracle;
	}

	public void setUsingOracle(boolean isUsingOracle) {
		this.isUsingOracle = isUsingOracle;
	}

	public void setImprAllowInLoop(boolean b) {
		this.isAllowInLoop = b;
		
	}

	public boolean isAllowInLoop() {
		return isAllowInLoop;
	}

	public int getNumOfImprLeaves() {
		return this.numOfImprLeaves;
	}




}
