package org.processmining.xlu.relabel.experiment;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.in.XesXmlGZIPParser;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.out.XesXmlGZIPSerializer;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot2Image;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;
import org.processmining.processtree.ptml.exporting.PtmlExportTree;
import org.processmining.processtree.ptml.importing.PtmlImportTree;
import org.processmining.xlu.ageneral.UtilWriter;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryETM;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.experiment.generator.GenerateLog;
import org.processmining.xlu.experiment.generator.GenerateLogParameters;
import org.processmining.xlu.experiment.generator.GenerateTree;
import org.processmining.xlu.experiment.generator.GenerateTreeParameters;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.zexperiment.EGLogDiscoveryTestPluginImpl;
import org.processmining.xlu.matching.zexperiment.ModelQuality;

import gnu.trove.map.hash.THashMap;

@Plugin(name = "Log Relabel Experiment", returnLabels = { "Log Relabeled" }, returnTypes = { XLog.class }, parameterLabels = {
		"Event Log", "Parameters", "Gen Tree" }, categories = { PluginCategory.Discovery,
		PluginCategory.Analytics }, keywords = { "Relabeling" }, help = "TODO", userAccessible = true, level = PluginLevel.Local)
public class ExpRelabelingPlugin {

	Random r = new Random();

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Log Relabeling Inf. Exp", requiredParameterLabels = { 0 })
//	public XLog relabellingLogUsingSL(PluginContext context,
//			EGraphsWithMappings graphsWithMappings) {
//		XLog log = RelabelAlgUsingMatching.relabellingLogUsingSLeemansImpl(graphsWithMappings);
//		EGLogDiscoveryTestPluginImpl.runDefaultTest(context, log);
//
//		return log;
//	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Exp. Log Relabeling Dup", requiredParameterLabels = { 0 })
	public XLog relabellingLogUsingSLDup(PluginContext context, XLog origLog)
			throws IOException {

		ExpRelabelingParam param = new ExpRelabelingParamSBP();
		ExpRelabelingAlg.runExperimentDuplicatedLabelRelabeling(context, origLog, param);

		return null;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Exp. Log Relabeling RealLife Dup", requiredParameterLabels = { 0 })
	public XLog relabellingRealLifeLog(PluginContext context, XLog origLog) {

		ExpRelabelingParam param = new ExpRelabelingParam();
		ExpRelabelingAlg.runRealLifeLog(context, origLog, param);

		return null;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Exp. Log Relabeling RealLife Dup using REG", requiredParameterLabels = { 0 })
	public XLog relabellingRealLifeLog(PluginContext context, RepEGraphs origLog) {

		ExpRelabelingParam param = new ExpRelabelingParam();
		XLog orgLog = origLog.getLog();

		XFactory factory = XFactoryRegistry.instance().currentDefault();

		String runday = UtilWriter.getRunDate();
		PtmlExportTree exporter = new PtmlExportTree();
		XesXmlGZIPSerializer serializer = new XesXmlGZIPSerializer();

		String fileName = "./dupResult/resultDupTask" + runday + ".csv";
		PrintStream writer = null;

//		writer = null;
		try {
			writer = new PrintStream(new File(fileName));
		} catch (IOException e) {
			e.printStackTrace();
		}
		ExpWriterRelabelingUtil.printColumn(writer);

		param.setWriter(writer);

		int cluster = 0;
		for (RepEGraph repgraph : origLog.getGraphs()) {
			XLog newLog = factory.createLog(orgLog.getAttributes());

			for (int traceIndex : repgraph.getRepresentedGraphIds()) {
				XTrace orgTrace = orgLog.get(traceIndex);

				newLog.add((XTrace) orgTrace.clone());

			}
			param.setLogNum(String.valueOf(cluster));
//			param.setDupTaskPairs(pairs);
			param.setRunday(runday);
			param.setModelName("Real");
//			param.setModelProperties(modelProperties);
//			param.setLogProperties(logProperties);
//			param.setLogNum(String.valueOf(runNum));
//			param.setWriter(writer);

			ExpRelabelingAlg.runRealLifeLog(context, newLog, param);
		}
		return null;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Exp. Log Relabeling Dup Random", requiredParameterLabels = {})
	public XLog relabellingLogUsingSLGenerator(PluginContext context) throws Exception {
		Random random = new Random();
		String runday = UtilWriter.getRunDate();
		PtmlExportTree exporter = new PtmlExportTree();
		XesXmlGZIPSerializer serializer = new XesXmlGZIPSerializer();

		String fileDirectory = "./dupResult/resultDupTask" + runday;
		new File(fileDirectory).mkdirs();
		fileDirectory += "/";
		new File(fileDirectory + "images").mkdirs();
		new File(fileDirectory + "logs").mkdirs();
		new File(fileDirectory + "precsv").mkdirs();
		new File(fileDirectory + "mappingDump").mkdirs();
		
		new File(fileDirectory).mkdirs();
		String fileName = fileDirectory + "res.csv";
		String settingFileName = fileDirectory + "setting.txt";

//		String fileName = "./dupResult/resultDupTask" + runday + ".csv";
		PrintStream writer = null;
		PrintStream settingWriter = null;
//		writer = null;
		try {
			writer = new PrintStream(new File(fileName));
			settingWriter = new PrintStream(new File(settingFileName));
		} catch (IOException e) {
			e.printStackTrace();
		}

		ExpWriterRelabelingUtil.printColumn(writer);
//		CsvWriter csvFusionWriter = new CsvWriter(writer, ';');
//		this.writers.put(file, csvFusionWriter);

		boolean isGenerate = true;
		boolean isGenerateLabels = true;
		String readrunday = "mrt04-1713";
		String readfileDirectory = "./dupResult/resultDupTask" + readrunday + "/";
		Map<String, List<ExpImprLabelPair<String>>> mapModel2PairDuplicates = isGenerate ? null
				: readFromFile(readfileDirectory + "setting.txt");

//		int modelNum = 1;
		for (int modelNum = 0; modelNum < 1; modelNum++) {
//		int modelNum = 0;
			String modelName = UtilWriter.int2Letter(modelNum);

			GenerateTree treeGen = new GenerateTree();
			GenerateTreeParameters treeGenParam = new GenerateTreeParameters();
//		int numAct = random.nextInt(20) + 10;
			int numAct = 20;
			treeGenParam.setNumberOfActivities(numAct);
			ProcessTree tree = null;

			if (isGenerate) {
				String modelfilename = fileDirectory + modelName + "_ModelGen" + runday
						+ ".ptml";
				tree = treeGen.generateTree(treeGenParam);
				exporter.exportDefault(context, tree, new File(modelfilename));
			} else {
				String modelfilename = readfileDirectory + modelName + "_ModelGen"
						+ readrunday + ".ptml";
//						+ "mrt04-1713.ptml";
				tree = (ProcessTree) (new PtmlImportTree()).importFile(context,
						modelfilename);
			}

			addObjectToContext(context, tree, "Generated Tree", ProcessTree.class);
			ModelProperties modelProperties = new ModelProperties(tree, numAct);

			/* ---------------------------------
			 * Get duplicated pairs.
			 * --------------------------------- */
			Type[] types = new Type[] { Type.SEQ };
			Collection<ExpImprLabelPair<String>> pairs = new ArrayList<ExpImprLabelPair<String>>();
			
			ExpRelabelingParam param = new ExpRelabelingParam();
		
			param.setRunday(runday);
			param.setModelName(modelName);
			param.setModelProperties(modelProperties);
			
			
			param.setWriter(writer);
			param.setDirectory(fileDirectory);

			/*
			 * 
			 */
			param.setUsingOracle(true);
			param.setImprAllowInLoop(true);
			param.setNumOfImprLeaves(1);

			if (isGenerate || isGenerateLabels) {
				for (Type t : types) {
//					ExpImprLabelPair<String> pair = ExpDuplicatedTaskUtil
//							.getRandomDupTaskPair(tree, t);
					int numLeaves = param.getNumOfImprLeaves();
					ExpImprLabelPair<String> pair = ExpDuplicatedTaskUtil
							.getRandomDupTaskSet(tree, numLeaves, param.isAllowInLoop() ?  1 : 0);
					
					if (pair != null) {
						pair.setMinimalSharedParent(t.toString());
						pairs.add(pair);
						settingWriter.println(modelName + ";" + t.toString() + ";"
								+ StringUtils.join(pair.getNodeAs(), ",") + ";"
								+ pair.getNodeB());
					}
				}
			} else {
				// TODO 
				List<ExpImprLabelPair<String>> pair = mapModel2PairDuplicates.get(modelName);
				if(pair != null){ // cause by models have only loops and imprecise label not allowed in loop. 
					pairs.addAll(pair);
				}
			}
			
			
			if(pairs.isEmpty() ){
				continue;
			} 
			ExpImprLabelPair<String> pair =	 pairs.iterator().next();
			if( pair.getNodeB() == null || pair.getNodeAs().size() < param.getNumOfImprLeaves() - 1 ){
				continue;
			}
			param.setDupTaskPairs(pairs);
			
			
			int runMax = 1;
			ExpScriptGenerator.writeHTML(fileDirectory, runday, modelName, runMax, pairs);

//			for (int runNum = 1; runNum < runMax + 1; runNum++) {
			int runNum = 1;
			XLog log = null;

			String logfilename = fileDirectory + "logs/" + modelName + "_" + runNum
					+ "_Log.xes.gz";
			
			param.setLogNum(String.valueOf(runNum));
		
			if (isGenerateLabels) {

				GenerateLog logGen = new GenerateLog();
				GenerateLogParameters logGenParam = new GenerateLogParameters();
				logGenParam.setNoiseEventsPerTrace(0);
				logGenParam.setNumberOfTraces(1000);
				logGenParam.setMaxTraceLength(80);

				log = logGen.generateLog(tree, logGenParam);
			} else {
				String readlogfilename = readfileDirectory + "logs/" + modelName + "_"
						+ runNum + "_Log.xes.gz";

				XesXmlGZIPParser parser = new XesXmlGZIPParser();
				log = parser.parse(new File(readlogfilename)).get(0);
			}

			LogProperties logProperties = new LogProperties(log);

			addObjectToContext(context, log, "Generated Log", XLog.class);
//			context.getProvidedObjectManager().createProvidedObject("Generated Log", log,
//					XLog.class, context);

			OutputStream output = new FileOutputStream(logfilename);
			serializer.serialize(log, output);
			
			param.setLogProperties(logProperties);
			
			

			MarkedPetrinet net = ExpDiscoveryETM.convertProcessTreeToPetrinet(context,
					tree);
			export2Image(net, new File(fileDirectory + "images/" + modelName
					+ "_" + runNum + "_M0.png"));
			ModelQuality qualityL1M1 = EGLogDiscoveryTestPluginImpl
					.runReplayToGetFitnessPrecisionAndGeneralization(context, net, log,
							new XEventNameClassifier());
			qualityL1M1.precWrtMGen = 1.0;
			qualityL1M1.recallWrtMGen = 1.0;
			ExpWriterRelabelingUtil.printQuality(null, "M0_gen;L2_impr", qualityL1M1,
					null, writer, param, null);

			param.setSystemBaseSet(qualityL1M1);

//				ExpWriterRelabelingUtil.printQuality(null, "M0", qualityL1M1, alg, writer, param);
			ExpRelabelingAlg.runExperimentDuplicatedLabelRelabeling(context, log, param);

//			}
		}
		writer.close();
		settingWriter.close();
		return null;
	}

	private Map<String, List<ExpImprLabelPair<String>>> readFromFile(String fileName) {
		BufferedReader br = null;
		Map<String, List<ExpImprLabelPair<String>>> res = new THashMap<>();
		try {
			br = new BufferedReader(new FileReader(fileName));
			StringBuilder sb = new StringBuilder();
			String line = br.readLine();

			while (line != null) {
				String[] value = line.split(";");
//		    	if(value[0].equals(modelName)){
				List<ExpImprLabelPair<String>> set = res.get(value[0]);

				if (set == null) {
					set = new ArrayList<ExpImprLabelPair<String>>();
				}

				String[] nodeAs = value[2].split(",");
				
				ExpImprLabelPair<String> pair = new ExpImprLabelPair<String>(nodeAs,
						value[3]);
				
				pair.setMinimalSharedParent(value[1]);
				set.add(pair);
				res.put(value[0], set);

				line = br.readLine();
			}
//		    String everything = sb.toString();

			br.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return res;
	}

	@SuppressWarnings("unchecked")
	private void addObjectToContext(PluginContext context, Object obj, String name,
			Class type) {
//		context.getProvidedObjectManager().createProvidedObject(name, obj,
//				type, context);
	}

	public static void export2Image(MarkedPetrinet net, File file) {
//		InductiveVisualMinerState state = new InductiveVisualMinerState(xLog, null);
//		state.setPaths(1.0);
//		InductiveVisualMinerPanel panel = new InductiveVisualMinerPanel(context, state,
//				Classifiers.getClassifiers(xLog), VisualMinerWrapperPluginFinder.find(context, state.getMiner()), true);
//		InductiveVisualMinerController controller = new InductiveVisualMinerController(context, panel, state);
////		panel.paintComponents( Graphics);
//		
//		ExporterPNG exporter=	new ExporterPNG();
//		try {
//			exporter.export(panel.getGraph(), file);
//		} catch (Exception e) {
//			e.printStackTrace();
//		}	
//		GraphVisualizerAlgorithm alg = new GraphVisualizerAlgorithm();
		Dot dot = org.processmining.xlu.relabel.visualization.GraphvizPetriNet.convert(
				net.getPetrinet(), net.getInitialMarking(), net.getFinalMarking());
//		NavigableSVGPanel panel = new AnimatableSVGPanel(DotPanel.dot2svg(dot));
		DotPanel panel = new DotPanel(dot);
		dot = panel.getDot();

		Dot2Image.dot2image(dot, file,
				org.processmining.plugins.graphviz.dot.Dot2Image.Type.png);
	}

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "X. Lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Exp. Log Relabeling Dev or Inf", requiredParameterLabels = { 0 })
//	public XLog relabellingLogContainInfreqOrDev(PluginContext context, XLog origLog) {
//
//		ExpRelabelingParam param = new ExpRelabelingParam();
//		ExpRelabelingAlg.runExperimentInfreqBehaviorRelabeling(context, origLog, param);
//
//		return null;
//	}

}
