package org.processmining.xlu.relabel.experiment;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Collection;

import org.apache.commons.lang3.StringUtils;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;

/**
 * Generate the HTML that shows the overview of models.
 * 
 * @author xlu
 * 
 */
public class ExpScriptGenerator {

	public static void writeHTML(String fileDirectory, String date, String modelName,
			int runMax, Collection<ExpImprLabelPair<String>> pairs) {

		// input 
//		int modelNum = 2;
//		String date = "okt19-2141";		
//		String modelName = UtilWriter.int2Letter(modelNum);
//		String modelName = "J";

		File file = new File(fileDirectory + "Result_model" + modelName + ".html");
		PrintWriter printWriter = null;
		try {
			printWriter = new PrintWriter(file);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		ExpRelabelingParam param = new ExpRelabelingParam();
//		Type[] types = new Type[] { Type.SEQ, Type.AND, Type.XOR, Type.LOOPXOR };

		printWriter.println("<!DOCTYPE HTML>");
		printWriter.println("<html>");
		printWriter.println("<head>");
		printWriter
				.println("<link rel=\"stylesheet\" href=\"css/style.css\" type=\"text/css\"/>");
		printWriter.println("</head>");
		printWriter.println("<body>");
		printWriter.println("<table border=\"1\" style=\"width:100%\">");
		printWriter.println("");
		printWriter.println("");
		// for each run of a model 
		for (int runNum = 1; runNum < runMax + 1; runNum++) {

			// Print row
			printWriter.println("<tr>");
			// print M0 in cell 1
			String name = modelName + "_" + runNum + "_M0";
			printWriter.println("<th><figure><img src=\"images/" + name + ".png\""
					+ " alt=\"" + name
					+ "\" style=\"width:700px;height:228px;\"><figcaption>" + name
					+ "</figcaption></figure></th>");

			printWriter.println("</tr>");

//			for (Type joinType : types) {
			for (ExpImprLabelSetI<String> pair : pairs) {
				String joinType = pair.getMinimalSharedParent();

				for (ExpDiscoveryAlgorithm alg : param.getDiscoveryAlgs()) {
					// Print row
//					printWriter.println("<th>" + alg.toString() +"</th>");
					printWriter.println("<tr>");
					for (int modelType = 1; modelType < 4; modelType++) {
						name = modelName + "_" + runNum + "_M" + modelType + "_"
								+ alg.toString() + "_" + joinType;

						String description = modelType == 1 ? name : 
							name + "(" + StringUtils.join(pair.getNodeAs(), ",") + ";" + pair.getNodeB() + ")";
						

						printWriter.println("<th><figure><img src=\"images/" + name + "_"
								+ date + ".png\"" + " alt=\"" + name
								+ "\" style=\"width:700px;height:228px;\"><figcaption>"
								+ description + "</figcaption></figure></th>");

					}
					printWriter.println("</tr>");
				}

				for (ExpDiscoveryAlgorithm alg : param.getHyberRelabelAlgs()) {
					printWriter.println("<tr>");
					printWriter.println("<th>");
					printWriter.println("</th>");
					printWriter.println("<th>");
					printWriter.println("</th>");
					int modelType = 4;
//					for (int modelType = 1; modelType < 4; modelType++) {
					name = modelName + "_" + runNum + "_M" + modelType + "_"
							+ alg.toString() + "_" + joinType;
					
					String description = 
						name + "(" + StringUtils.join(pair.getNodeAs(), ",") + ";" + pair.getNodeB() + ")";
					
					printWriter.println("<th><figure><img src=\"images/" + name + "_"
							+ date + ".png\"" + " alt=\"" + name
							+ "\" style=\"width:700px;height:228px;\"><figcaption>"
							+ description + "</figcaption></figure></th>");

//					}
					printWriter.println("</tr>");
				}

			}

		}
		printWriter.println("");
		printWriter.println("</table>");
		printWriter.println("</body>");
		printWriter.println("</html>");
		printWriter.close();

	}

}
