package org.processmining.xlu.relabel.experiment;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.processmining.xlu.experiment.discovery.ExpModelSimplicityUtil;
import org.processmining.xlu.matching.zexperiment.ModelQuality;
import org.processmining.xlu.relabel.experiment.LogProperties.LogProperty;
import org.processmining.xlu.relabel.experiment.ModelProperties.ModelProperty;
import org.processmining.xlu.relabel.param.RelabelParam;

public class ExpWriterRelabelingUtil {

	
	public static void printColumn(PrintStream writer) {
		String s = "Mnum;Lnum;Source;Target;Shared;Path2A;Path2B;Path2Root;detected;UseDOracle;ImprInLoop;"
				+ "Algorithm;Model;Log;isImproved;Fitness;Precision;PrecisionDL;PrecisionDLT;" 
				+ ModelQuality.LOG_PREC_LB + ";" + ModelQuality.LOG_PREC + ";" + ModelQuality.LOG_PREC_TO_UB + ";" 
				+ "PrecWrtMGen;" 
				+ ModelQuality.SYS_PREC_LB + ";" + ModelQuality.SYS_PREC_TO_LB + ";" + ModelQuality.SYS_PREC_TO_UB + ";" 
				+ "RecallWrtMGen;F1WrtMGen;";
//		for(String key : properties){
//			s += key + ";";
//		}
		s += StringUtils.join(ExpModelSimplicityUtil.propertykeys, ";");

		s += ";numSM;numLM;numMM;numMMinv;Config";
		
		for(ModelProperty t : ModelProperty.values()){
			s +=";" + t.toString() ;
		}
		for(LogProperty t : LogProperty.values()){
			s += ";" + t.toString();
		}
		
		s += ";edgeCost;unfThres;runday"; 
		writer.println(s);
		writer.flush();
	}

	public static void printQuality(ExpImprLabelSetI<String> pair, String l,
			ModelQuality quality, String alg, PrintStream writer, ExpRelabelingParam param, RelabelParam<?,?> relabelParam) {
		

		List<String> res = new ArrayList<String>();

		res.add(param.getModelName());
		res.add(param.getLogNum());
		String[] ss = ExpDuplicatedTaskUtil.getStrings(pair);
		for (String s : ss) {
			res.add(s);
		}

		res.add(param.isDetected().toString());
		res.add(String.valueOf(param.isUsingOracle()));
		res.add(String.valueOf(param.isAllowInLoop()));
		res.add(alg != null? alg.toString(): "Sys");
		res.add(l);
		res.add(String.valueOf(quality.isImproved()));
		//NormPr(ABS)	NormToImpr	PrecWrtMGen	SysPr(ABS)	SysPrecToImpr

		res.add(String.valueOf(quality.fitness));
		res.add(String.valueOf(quality.precision));
		res.add(String.valueOf(quality.precisionDistinctlabels));
		res.add(String.valueOf(quality.precisionDistinctlabelsTau));
		
		res.add(quality.properties.get(ModelQuality.LOG_PREC_LB) != null ? quality.properties.get(ModelQuality.LOG_PREC_LB) : "");
		res.add(quality.properties.get(ModelQuality.LOG_PREC) != null ? quality.properties.get(ModelQuality.LOG_PREC) : "");
		res.add(quality.properties.get(ModelQuality.LOG_PREC_TO_UB) != null ? quality.properties.get(ModelQuality.LOG_PREC_TO_UB) : "");
		
		res.add(String.valueOf(quality.precWrtMGen));
		
		res.add(quality.properties.get(ModelQuality.SYS_PREC_LB) != null ? quality.properties.get(ModelQuality.SYS_PREC_LB) : "");
		res.add(quality.properties.get(ModelQuality.SYS_PREC_TO_LB) != null ? quality.properties.get(ModelQuality.SYS_PREC_TO_LB) : "");
		res.add(quality.properties.get(ModelQuality.SYS_PREC_TO_UB) != null ? quality.properties.get(ModelQuality.SYS_PREC_TO_UB) : "");
		
		res.add(String.valueOf(quality.recallWrtMGen));
		res.add(String.valueOf((quality.precWrtMGen + quality.recallWrtMGen)/2.0));
		
//		for (Entry<String, String> entry : propertykeys) {
//			res.add(entry.getValue());
//		}
		for(int i = 0; i < ExpModelSimplicityUtil.propertykeys.length; i++){
			res.add(quality.getProperties().get(ExpModelSimplicityUtil.propertykeys[i]));
		}
		
		res.add(String.valueOf(quality.getNumSM()));
		res.add(String.valueOf(quality.getNumLM()));
		res.add(String.valueOf(quality.getNumMM()));
		res.add(String.valueOf(quality.getNumMMinv()));

		res.add(param.getTraceMatchingConfiguration().toString());
		
		
		for (ModelProperty entry : ModelProperty.values()) {
			Object object = param.getModelProperties().getProperties().get(entry);
			res.add(object != null? object.toString() : "");
		}
		for (LogProperty entry : LogProperty.values()) {
			Object object = param.getLogProperties().getProperties().get(entry);
			res.add(object != null? object.toString() : "");
		}
		
		res.add(String.valueOf(relabelParam != null? relabelParam.getRelabelParam().getMaxCostOfEdge() : ""));
		res.add(String.valueOf(relabelParam != null? relabelParam.getRelabelParam().getUnfoldingThreshold(): ""));
		res.add(param.runday);
				
				
		String result = StringUtils.join(res, ";");

		writer.println(result);
		writer.flush();
	}

}
