package org.processmining.xlu.relabel.experiment;

import gnu.trove.map.hash.THashMap;

import java.util.Map;

import org.processmining.processtree.Block.And;
import org.processmining.processtree.Block.Seq;
import org.processmining.processtree.Block.Xor;
import org.processmining.processtree.Block.XorLoop;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;

public class ModelProperties {

	public enum ModelProperty {
		N_SEQ("nSeq"), N_AND("nAnd"), N_XOR("nXor"), N_LOOP("nLoop"), N_ACTIVITY("nAct");

		private final String text;

		private ModelProperty(final String text) {
			this.text = text;
		}

		public String toString() {
			return text;
		}

	}

	private Map<Object, Object> properties;

	public Map<Object, Object> getProperties() {
		return properties;
	}

	public void setProperties(Map<Object, Object> properties) {
		this.properties = properties;
	}

	public ModelProperties(ProcessTree tree, int numAct) {
		properties = new THashMap<Object, Object>();
		for (ModelProperty p : ModelProperty.values()) {
			properties.put(p, 0);
		}

		properties.put(ModelProperty.N_ACTIVITY, numAct);

		for (Node n : tree.getNodes()) {
			ModelProperty type = getType(n);
			if (type != null) {
				properties.put(type, (Integer) properties.get(type) + 1);
			}
		}
	}

	public ModelProperties() {
		// TODO Auto-generated constructor stub
		properties = new THashMap<Object, Object>();
	}

	private ModelProperty getType(Node n) {
		ModelProperty type = null;
		if (n instanceof Seq) {
			type = ModelProperty.N_SEQ;
		} else if (n instanceof And) {
			type = ModelProperty.N_AND;
		} else if (n instanceof Xor) {
			type = ModelProperty.N_XOR;
		} else if (n instanceof XorLoop) {
			type = ModelProperty.N_LOOP;
		}
		return type;
	}
}
