package org.processmining.xlu.relabel.plugin;

import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryInductiveMiner;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.matching.zexperiment.ModelQuality;
import org.processmining.xlu.relabel.experiment.ExpRelabelingAlg;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelParam;
import org.processmining.xlu.relabel.param.RelabelParamPT;
import org.processmining.xlu.relabel.plugin.detection.ILDetector;
import org.processmining.xlu.relabel.plugin.detection.RelabelScope;
import org.processmining.xlu.relabel.plugin.matching.ModelCreator;

/**
 * 
 * Created July 6th, 2015.
 * 
 * @author x.lu
 * 
 */

@Plugin(name = "Log Relabeling", returnLabels = { "Log Relabeled" }, returnTypes = {
		XLog.class }, parameterLabels = { "Event Log", "Parameters" }, categories = {
				PluginCategory.Discovery, PluginCategory.Analytics }, keywords = {
						"Relabeling" }, help = "TODO", userAccessible = true)
public class RelabelGeneralPlugin {

	PluginContext context = null;

	//	public static XLog relabelLogLocalDefault(XLog log) {
	//		return relabelLogGeneral(log, RelabelParamPT.createLocalDefaultParam());
	//	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Refine Labels Globally, Default", requiredParameterLabels = {
			0 })
	public XLog relabelLogGlobalDefault(PluginContext context, XLog log) {
		this.context = context;
		return relabelLogGlobalDefault(log);
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Refine Labels Globally, Adaptive", requiredParameterLabels = {
			0 })
	public XLog relabelLogGlobalAdaptive(PluginContext context, XLog log) {
		this.context = context;
		return relabelLogGlobalAdaptive(log);
	}

	public XLog relabelLogGlobalDefault(XLog log) {
		return relabelLogGeneral(log, RelabelParamPT.createGlobalDefaultParam());
	}

	public XLog relabelLogGlobalAdaptive(XLog log) {
		RelabelParamPT relabelParam = RelabelParamPT.createGlobalDefaultParam();

		// Discover model for current log as base-line (fall-through) 
		ExpDiscoveryInductiveMiner discoveryMiner = new ExpDiscoveryInductiveMiner();
		XEventClassifier classifier = new XEventNameClassifier();
		int numEvtClassesLabelLog = XEventClasses.deriveEventClasses(classifier, log)
				.size();
		ModelQuality qualityL2M2 = ExpRelabelingAlg.computeQuality(context, classifier,
				log, discoveryMiner.minePetrinet(context, log));

		// Init relabel parameters. 
		ILDetector<?, ?> detector = relabelParam.getDetector();
		RelabelScope<?, ?> scope = detector.detectImpreciseLabels(log,
				relabelParam.getDetectorParam());

		ModelCreator algCreator = relabelParam.getRelabeler();
		RelabelModel model = algCreator.createRelabelModel(log, scope.getLabelScope(),
				relabelParam);

		// Run the adaptive version.
		XLog log3R = (XLog) log.clone();
		boolean found = false;
		loop: for (int edgeCostThreshold = 5; edgeCostThreshold >= 0; edgeCostThreshold -= 1) {
			for (int unfthreshold = 0; unfthreshold < 60; unfthreshold += 10) {

				relabelParam.getRelabelParam().setMaxCostOfEdge(edgeCostThreshold);
				relabelParam.getRelabelParam().setUnfoldingThreshold(unfthreshold);

				log3R = algCreator.applyRelabeling(model, relabelParam);
				ModelQuality qualityL2M3 = computeQuality(relabelParam, log3R, scope, log,
						discoveryMiner, numEvtClassesLabelLog);

				if (qualityL2M3 != null
						&& qualityL2M3.precision > qualityL2M2.precision) {

					found = true;
					break loop;
				}
			}
		}

		if (!found) {
			return log;
		}

		return log3R;
	}

	private ModelQuality computeQuality(RelabelParam<?, ?> relabelParam, XLog log3R,
			RelabelScope<?, ?> scope, XLog log, ExpDiscoveryInductiveMiner discoveryMiner,
			int numEvtClassesLabelLog) {
		int numEvtClasses = XEventClasses
				.deriveEventClasses(relabelParam.getClassifier(), log3R).size();
		if (numEvtClasses < numEvtClassesLabelLog + scope.getLabelScope().size() * 10) {
			MarkedPetrinet net3 = discoveryMiner.minePetrinet(context, log3R);
			ExpRelabelingAlg.modifiedModel3ToL2(net3);

			if (net3.getPetrinet().getTransitions().size() < 40) {
				return ExpRelabelingAlg.computeQuality(context,
						relabelParam.getClassifier(), log, net3);

			}
		}
		return null;
	}

	public XLog relabelLogGeneral(XLog log, RelabelParam<?, ?> relabelParam) {
		int numEvtClassesLabelLog = XEventClasses
				.deriveEventClasses(relabelParam.getClassifier(), log).size();

		ExpDiscoveryInductiveMiner discoveryMiner = new ExpDiscoveryInductiveMiner();
		ModelQuality qualityL2M2 = ExpRelabelingAlg.computeQuality(context,
				relabelParam.getClassifier(), log,
				discoveryMiner.minePetrinet(context, log));

		ILDetector<?, ?> detector = relabelParam.getDetector();
		RelabelScope<?, ?> scope = detector.detectImpreciseLabels(log,
				relabelParam.getDetectorParam());

		if (context != null) {
			context.log("Detected imprecise label candidates: "
					+ StringUtils.join(scope.getLabelScope(), ","));
		}

		ModelCreator alg = relabelParam.getRelabeler();
		RelabelModel model = alg.createRelabelModel(log, scope.getLabelScope(),
				relabelParam);

		relabelParam.getRelabelParam().setMaxCostOfEdge(5);
		relabelParam.getRelabelParam().setUnfoldingThreshold(0);

		XLog relabeledLog = alg.applyRelabeling(model, relabelParam);

		ModelQuality qualityL2M3 = computeQuality(relabelParam, relabeledLog, scope, log,
				discoveryMiner, numEvtClassesLabelLog);

		boolean found = false;
		if (qualityL2M3 != null && qualityL2M3.precision > qualityL2M2.precision) {
			found = true;
		}
		if (!found) {
			return log;
		}

		return relabeledLog;
	}

}
