package org.processmining.xlu.relabel.plugin;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.model.RelabelVisModel;
import org.processmining.xlu.relabel.param.RelabelParamPT;
import org.processmining.xlu.relabel.plugin.detection.ILDetector;
import org.processmining.xlu.relabel.plugin.detection.ILDetectorIMGlobal;
import org.processmining.xlu.relabel.plugin.detection.RelabelScope;
import org.processmining.xlu.relabel.plugin.matching.ModelCreator;
import org.processmining.xlu.relabel.visualization.RelabelControllerModel;

@Plugin(name = "Interactive Relabeling", returnLabels = { "Log Relabeled" }, returnTypes = { RelabelVisModel.class }, parameterLabels = {
		"Event Log", "Parameters" }, categories = { PluginCategory.Discovery,
		PluginCategory.Analytics }, keywords = { "Relabeling" }, help = "TODO", userAccessible = true)
public class RelabelInteractivePlugin {

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Interactive Relabel Model", requiredParameterLabels = { 0 })
//	public RelabelModel relabellingLogGivenMapping(UIPluginContext context, EGraphsWithMappings graphsWithMappings) {
//		return RelabelAlgUsingMatching.createRelabelModel(graphsWithMappings);
//	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Interactive Relabel Global", requiredParameterLabels = { 0 })
	public RelabelVisModel relabelLogGlobal(PluginContext context, XLog log) {
//		MUDATAPreprocess.updateMUDataLog1econsult(log);
		
		RelabelParamPT param = RelabelParamPT.createGlobalDefaultParam();
//		param.setDetector(new BPIC2015_LOG_10());
//		param.setDetector(new ILDetectorMUMCDataset());
//		param.setDetector(new ILDetectorIMGlobalNaive());
		param.setDetector(new ILDetectorIMGlobal());
//		param.setDetector(new ILDetectorExpert("C"));
		// Detect imprecise labels
		ILDetector<ProcessTree, Node> detectionAlg = param.getDetector();
		RelabelScope<ProcessTree, Node> scope = detectionAlg.detectImpreciseLabels(log,
				param.getDetectorParam());

		// Create relabel model 
		ModelCreator alg = param.getRelabeler();
		
		// Does not work, everytime new config is returned. 
//		param.getRelabelerParam().getMatchingConfig().setPTraceParameter(new PTraceParameter(PTraceType.Non_Same_Day_Dependency));		
//		param.getRelabelerParam().getMatchingConfig().getCostStrategy().getCostConfiguration().setWeightMatched(1);
		RelabelModel model = alg.createRelabelModel(log, scope.getLabelScope(), param);

//		RelabelVisModel vismodel = createVisModel(log, param);
		return new RelabelControllerModel(model, scope, param);
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "Interactive Relabel Local", requiredParameterLabels = { 0 })
	public RelabelVisModel relabelLogLocal(PluginContext context, XLog log) {
		RelabelParamPT param = RelabelParamPT.createLocalDefaultParam();
//		param.setDetector(new BPIC2015_LOG_10_loc());
		// Detect imprecise labels
		ILDetector<ProcessTree, Node> detectionAlg = param.getDetector();
		RelabelScope<ProcessTree, Node> scope = detectionAlg.detectImpreciseLabels(log,
				param.getDetectorParam());

		// Create relabel model 
		ModelCreator alg = param.getRelabeler();
		RelabelModel model = alg.createRelabelModel(scope.getLogForRelabel(),
				scope.getLabelScope(), param);

//		RelabelVisModel vismodel = createVisModel(log, param);
		return new RelabelControllerModel(model, scope, param);
	}

//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "x.lu", email = "x.lu@tue.nl", uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "Interactive Relabel Model", requiredParameterLabels = { 0 })
//	public RelabelModel relabellingLogHYB(PluginContext context, XLog log) {
//		
//		EGMatchConfiguration conf = EGMatchConfigurationFactory.createSeqBasedGreedyMatchAsMuchAsPossibleConfig();
////		Set<String> labels = RelabelPlugin.detectCoarseLabel(context, conf.getClassifier(), log);
////		conf.setFocusOnMatchingLabels(labels);
//		
//		EGraphsWithMappings graphsWithMappings = (new EGMatchPlugin()).replayLog(context,
//				(XLog) log.clone(), conf);
//		return RelabelAlgUsingMatching.createRelabelModel(graphsWithMappings);
//	}
}
