package org.processmining.xlu.relabel.plugin;

import org.deckfour.xes.model.XLog;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelParam;
import org.processmining.xlu.relabel.plugin.detection.ILDetector;
import org.processmining.xlu.relabel.plugin.detection.RelabelScope;
import org.processmining.xlu.relabel.plugin.matching.ModelCreator;

public abstract class RelabelLocalAlg<M, S> {

	XLog refinedLog;
	private RelabelModel model;

	public RelabelModel getModel() {
		return model;
	}

	public XLog getRefinedLog() {
		return refinedLog;
	}

	public M relabelLog(XLog log, RelabelParam<M, S> relabelParam) {

		ILDetector<M, S> detector = relabelParam.getDetector();
		RelabelScope<M, S> scope = detector.detectImpreciseLabels(log, relabelParam.getDetectorParam());

		XLog projectedLog = scope.getLogForRelabel();
		if (scope.getLabelScope().isEmpty() && projectedLog == null) {
			// No imprecise labels detected
			return scope.getModel();
		}

		ModelCreator alg = relabelParam.getRelabeler();
		model = alg.createRelabelModel(projectedLog, scope.getLabelScope(), relabelParam);

		refinedLog = alg.applyRelabeling(model, relabelParam);

		return relabelParam.getReplacer().rediscover(scope, refinedLog);
	}
}
