package org.processmining.xlu.relabel.plugin.detection;

import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.mining.logs.IMLog;
import org.processmining.plugins.InductiveMiner.mining.postprocessor.PostProcessor;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task.Manual;
import org.processmining.xlu.relabel.plugin.hyber.IMSubLogPostProcessor;

public class ExpILDetectorOracleLocal implements ILDetector<ProcessTree, Node> {

	private Set<String> impreciseLabels = new THashSet<String>();

	public Set<String> getImpreciseLabels() {
		return impreciseLabels;
	}

	public void setImpreciseLabels(Set<String> impreciseLabels) {
		this.impreciseLabels = impreciseLabels;
	}

	public RelabelScope<ProcessTree, Node> detectImpreciseLabels(XLog log,
			ILDetectorParam param) {

//			List<Type> operators = param.getSelectedOperators();
		XEventClassifier classifier = param.getClassifier();
		MiningParametersIM imParam = new MiningParametersIM();
		imParam.setClassifier(classifier);

		IMSubLogPostProcessor postProcessor = new IMSubLogPostProcessor();
		imParam.setPostProcessors(new ArrayList<PostProcessor>(Arrays
				.asList(postProcessor)));

		//Get a model. 
		ProcessTree tree = IMProcessTree.mineProcessTree(log, imParam);

		Map<Node, IMLog> mapNode2Log = postProcessor.getMapNodeToLog();

		List<Node> selectedNodes = new ArrayList<Node>();
		for (Node n : tree.getNodes()) {
			if (impreciseLabels.contains(n.getName())) {
				selectedNodes.add(n);
			}
		}
		Node selectedNode = getSharedParent(selectedNodes, tree);
		IMLog selectedLog = mapNode2Log.get(selectedNode);
		
		while (selectedLog == null || selectedLog.size() > log.size() * 3){
			Block parent = selectedNode.getParents().iterator().next();
			selectedLog = mapNode2Log.get(parent);
			selectedNode = parent;
		}
		
		XLog projectedLog = selectedNode != null ? selectedLog.toXLog()
				: null;

		RelabelScopePT scope = RelabelScopePT.createScope(tree, selectedNode,
				impreciseLabels, log, projectedLog);

		return scope;
	}

	private Node getSharedParent(List<Node> selectedNodes, ProcessTree tree) {
		Node root = tree.getRoot();

		//TODO 
		if (selectedNodes.size() == 0 || selectedNodes.size() > 1) {
			System.err.println("No Imprecise Leaves");

			return root;
		}
		Node n = selectedNodes.get(0);
		Block parent = n.getParents().iterator().next();
		
		if(parent == root){
			return root;
		}
		
		int num = getNumberManualLeaves(parent);
		if(num > 1 && num < 5){
			Node newParent = parent.getParents().iterator().next();
			if(newParent != null && newParent != root ){
				return newParent;
			} else {
				return parent;
			}
		}
		
		
		

		
//		if (selectedNodes.size() == 1) {
//			return selectedNodes.get(0).getParents().iterator().next();
//		}
//		Map<Node, Collection<Block>> mapNtoAncestors = new THashMap<Node, Collection<Block>>();
//		
//		for(Node n : selectedNodes){
//			Collection<Block> parents = new THashSet<Block>(n.getParents());
//			mapNtoAncestors.put(n, parents);
//		}
//		
////		Node shared = null;
////		while(shared == null){
//			Iterator<Collection<Block>> i = mapNtoAncestors.values().iterator();
//			Set<Block> sharedSet = new THashSet<Block>(i.next());
//			
//			while (i.hasNext()){
//				sharedSet.retainAll(i.next());
//			}
//			
//			if(sharedSet.size() > 0){
//				return sharedSet.iterator().next();
//			} else {
//				for( Entry<Node, Collection<Block>> entry : mapNtoAncestors.entrySet()){
//					
//				}
//			}
		return parent;
	}

//	private boolean isChildOf(Node child, Node parent) {
//		
//		return false;
//	}

private int getNumberManualLeaves(Block parent) {
		Set<Node> children = new THashSet<Node>();
		getChildren(parent, children);
		
		return children.size();
	}

	private void getChildren(Block parent, Set<Node> children) {
		for(Node n : parent.getChildren()){
			if(n instanceof Block){
				getChildren((Block)n, children);
			} else if(n instanceof Manual){
				children.add(n);
			}
		}
	
	}

	public void addImpreciseLabel(String nodeB) {
		impreciseLabels.add(nodeB);
	}
}
