package org.processmining.xlu.relabel.plugin.detection;

import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.mining.logs.IMLog;
import org.processmining.plugins.InductiveMiner.mining.postprocessor.PostProcessor;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;
import org.processmining.processtree.Task.Automatic;
import org.processmining.processtree.Task.Manual;
import org.processmining.xlu.relabel.plugin.hyber.IMSubLogPostProcessor;

public class ILDetectorIMGlobalNaive implements ILDetector<ProcessTree, Node> {

	public RelabelScope<ProcessTree, Node> detectImpreciseLabels(XLog log,
			ILDetectorParam param) {
		if (param instanceof ILDetectorIMParam) {
			return detectImpreciseLabelsIM(log, (ILDetectorIMParam) param);
		} //
		return new RelabelScopePT();
	}

	/**
	 * 
	 * @param log
	 * @param classifier
	 * @param param
	 *            : specifies the operators and the maximal number of imprecise
	 *            labels
	 * @return
	 */
	public RelabelScope<ProcessTree, Node> detectImpreciseLabelsIM(XLog log,
			ILDetectorIMParam param) {

		List<Type> operators = param.getSelectedOperators();
		XEventClassifier classifier = param.getClassifier();
		MiningParametersIM imParam = new MiningParametersIM();
		imParam.setClassifier(classifier);
		IMSubLogPostProcessor postProcessor = new IMSubLogPostProcessor();
		imParam.setPostProcessors(new ArrayList<PostProcessor>(Arrays
				.asList(postProcessor)));

		//Get a model. 
		ProcessTree tree = IMProcessTree.mineProcessTree(log, imParam);

		List<Node> selectedNodes = new ArrayList<Node>();
		for (Node n : tree.getNodes()) {
			if (operators.contains(tree.getType(n))) {
				selectedNodes.add(n);
//				break;
			}
		}
//		Map<Node, Set<String>> labelSets = getLeafLabelSets(tree, selectedNodes);
//		Map<Node, IMLog> mapNode2Log = postProcessor.getMapNodeToLog();

		Map<Node, IMLog> mapNode2Log = postProcessor.getMapNodeToLog();

		Node selectedNode = null;
		Set<String> selectedlabels = new THashSet<String>();
		for (Node n : selectedNodes) {
			Set<String> labels = new THashSet<String>();
			ILDetectorIMGlobalNaive.recursivelyGetLeaves(n, labels);

			if (ILDetectorIMLocal.isEqualSets(n, selectedNode)
					&& labels.size() < selectedlabels.size()
					&& labels.size() <= param.getMaxNumLabels()) {
				selectedNode = n;
				selectedlabels = labels;
			} else if (ILDetectorIMLocal.isTauLoop(n)
					&& !ILDetectorIMLocal.isTauLoop(selectedNode)) {
				selectedNode = n;
				selectedlabels = labels;
			}
		}

//		XLog projectedLog = selectedNode != null ? mapNode2Log.get(selectedNode).toXLog()
//				: null;
//		Set<String> dupImprecises = new THashSet<String>();
//		if (projectedLog != null) {
//			for (XTrace t : projectedLog) {
//				MultiSet<String> dupImprecise = new HashMultiSet<String>();
//				for (XEvent e : t) {
//					dupImprecise.add(classifier.getClassIdentity(e));
//				}
//				for (String l : dupImprecise.baseSet()) {
//					if (dupImprecise.occurrences(l) > 1) {
//						dupImprecises.add(l);
//					}
//				}
//			}
//		}
//		if (selectedNode != null && selectedNode instanceof Block) {
//			ILDetectorIMLocal.removePreciseLoopLabels(selectedNode, dupImprecises);
//
//		}

		RelabelScopePT scope = RelabelScopePT.createScope(tree, tree.getRoot(),
				selectedlabels, log, log);

		return scope;
	}



	public static Map<Node, Set<String>> getLeafLabelSets(ProcessTree tree,
			List<Node> selectedNodes) {
		Map<Node, Set<String>> mapNode2LabelSets = new THashMap<Node, Set<String>>();
		// for each selected node which is a loop, obtain all its leaves. 
		nextNode: for (Node n : selectedNodes) {

			Set<String> newlabels = new THashSet<String>();
			recursivelyGetLeaves(n, newlabels);

			// handle loops contained in another loop
			boolean newlabelsContained = false;
			for (Iterator<Entry<Node, Set<String>>> iterator = mapNode2LabelSets
					.entrySet().iterator(); iterator.hasNext();) {
				Entry<Node, Set<String>> entry = iterator.next();
				Set<String> obtainedSet = entry.getValue();
				// if a small loop is contained in a larger loop, remove the smaller loop
				if (obtainedSet.containsAll(newlabels)) {
					newlabelsContained = true;
					break nextNode;
//					iterator.remove();
				} else if (newlabels.containsAll(obtainedSet)) {
					// if a larger loop is found, remove the smaller loops in the map.
					iterator.remove();
//					continue nextNode;
				}
			}
			if (!newlabelsContained) {
				mapNode2LabelSets.put(n, newlabels);
			}
		}

		return mapNode2LabelSets;
	}

	public static void recursivelyGetLeaves(Node n, Set<String> newlabels) {
		if (n instanceof Manual) {
			newlabels.add(((Manual) n).getName());
			return;
		} else if (n instanceof Automatic) {
			newlabels.add("XXTAUXX");
			return;
		} else if (n instanceof Block) {
			Block b = (Block) n;
			for (Node child : b.getChildren()) {
				recursivelyGetLeaves(child, newlabels);
			}
		}

	}

}
