package org.processmining.xlu.relabel.plugin.detection;

import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.util.collection.HashMultiSet;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.mining.logs.IMLog;
import org.processmining.plugins.InductiveMiner.mining.postprocessor.PostProcessor;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.XorLoop;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;
import org.processmining.processtree.Task.Automatic;
import org.processmining.xlu.relabel.plugin.hyber.IMSubLogPostProcessor;

public class ILDetectorIMLocal implements ILDetector<ProcessTree, Node> {

	public RelabelScope<ProcessTree, Node> detectImpreciseLabels(XLog log,
			ILDetectorParam param) {
		if (param instanceof ILDetectorIMParam) {
			return detectImpreciseLabelsIM(log, (ILDetectorIMParam) param);
		} //
		return new RelabelScopePT();
	}

	public RelabelScope<ProcessTree, Node> detectImpreciseLabelsIM(XLog log,
			ILDetectorIMParam param) {

		List<Type> operators = param.getSelectedOperators();

		MiningParametersIM imParam = new MiningParametersIM();
		XEventClassifier classifier = param.getClassifier();
		
		imParam.setClassifier(classifier);
		IMSubLogPostProcessor postProcessor = new IMSubLogPostProcessor();
		imParam.setPostProcessors(new ArrayList<PostProcessor>(Arrays
				.asList(postProcessor)));

		//Get a model. 
		ProcessTree tree = IMProcessTree.mineProcessTree(log, imParam);

		//Get a set of sets of transitions to relabel... 
		List<Node> selectedNodes = new ArrayList<Node>();
		for (Node n : tree.getNodes()) {
			if (operators.contains(tree.getType(n))) {
				selectedNodes.add(n);
//				break;
			}
		}
//		Map<Node, Set<String>> labelSets = getLeafLabelSets(tree, selectedNodes);
		Map<Node, IMLog> mapNode2Log = postProcessor.getMapNodeToLog();

		Node selectedNode = null;
		Set<String> selectedlabels = new THashSet<String>();
		for (Node n : selectedNodes) {
			Set<String> labels = new THashSet<String>();
			ILDetectorIMGlobal.recursivelyGetLeaves(n, labels);

			if (isEqualSets(n, selectedNode)
					&& labels.size() < selectedlabels.size()
					&& labels.size() <= param.getMaxNumLabels()) {
				selectedNode = n;
				selectedlabels = labels;
			} else if (isTauLoop(n) && !isTauLoop(selectedNode)) {
				selectedNode = n;
				selectedlabels = labels;
			}
		}

		
		XLog projectedLog = selectedNode != null ? mapNode2Log.get(selectedNode).toXLog()
				: null;
		Set<String> dupImprecises = new THashSet<String>();
		if (projectedLog != null) {
			for (XTrace t : projectedLog) {
				MultiSet<String> dupImprecise = new HashMultiSet<String>();
				for (XEvent e : t) {
					dupImprecise.add(classifier.getClassIdentity(e));
				}
				for (String l : dupImprecise.baseSet()) {
					if (dupImprecise.occurrences(l) > 1) {
						dupImprecises.add(l);
					}
				}
			}
		}
		if (selectedNode != null && selectedNode instanceof Block) {
			removePreciseLoopLabels(selectedNode, dupImprecises);

		}

		RelabelScopePT scope = RelabelScopePT.createScope(tree, selectedNode,
				dupImprecises, log, projectedLog);

		return scope;

	}

	public static void removePreciseLoopLabels(Node selectedNode, Set<String> dupImprecises) {
		if (selectedNode instanceof Block) {
			for (Node child : ((Block) selectedNode).getChildren()) {
				if (child instanceof XorLoop && !ILDetectorIMLocal.isTauLoop(child)) {
					Set<String> preciseLoopLabels = new THashSet<>();
					ILDetectorIMGlobal.recursivelyGetLeaves(child, preciseLoopLabels);
					dupImprecises.removeAll(preciseLoopLabels);
				} else {
					removePreciseLoopLabels(child, dupImprecises);
				}
			}
		}
	}
	
	public static boolean isEqualSets(Node n, Node selectedNode) {		
		return isTauLoop(n) == isTauLoop(selectedNode);
	}

	public static boolean isTauLoop(Node n) {
		if(n instanceof XorLoop){
			XorLoop loop = (XorLoop) n;
			if(loop.getChildren().get(1) instanceof Automatic){
				return true;
			}
		}
		return false;
	}

}
