package org.processmining.xlu.relabel.plugin.hyber;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.conversion.ProcessTree2Petrinet;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.InvalidProcessTreeException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.NotYetImplementedException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.PetrinetWithMarkings;
import org.processmining.xlu.experiment.discovery.ExpLocRefiningDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelParamPT;
import org.processmining.xlu.relabel.plugin.RelabelLocalAlg;
import org.processmining.xlu.relabel.plugin.RelabelLocalAlgPT;
import org.processmining.xlu.relabel.plugin.detection.ILDetectorParam;

public class RelabelHyberApproachIM implements ExpLocRefiningDiscoveryAlgorithm {

	private RelabelParamPT relabelParam = RelabelParamPT.createLocalDefaultParam();
	private XLog refinedSublog;
	private RelabelModel rmodel;

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log) {
		return minePetrinet(context, log,
				relabelParam != null ? relabelParam.getClassifier()
						: new XEventNameClassifier());
	}

	public MarkedPetrinet minePetrinet(PluginContext context, XLog log,
			XEventClassifier classifier) {

//		RelabelParamPT relabelParam = 
		ILDetectorParam param = relabelParam.getDetectorParam();
		param.setClassifier(classifier);

		RelabelLocalAlg<ProcessTree, Node> alg = new RelabelLocalAlgPT();
		ProcessTree tree = alg.relabelLog(log, relabelParam);
		refinedSublog = alg.getRefinedLog();
		rmodel = alg.getModel();

		//create Petri net connections
		MarkedPetrinet net = convertTreeInContext(context, tree);

		return net;
	}

	public static MarkedPetrinet convertTreeInContext(PluginContext context,
			ProcessTree tree) {
		MarkedPetrinet net = convertTree(tree);
		context.addConnection(new InitialMarkingConnection(net.getPetrinet(), net
				.getInitialMarking()));
		context.addConnection(new FinalMarkingConnection(net.getPetrinet(), net
				.getFinalMarking()));
		return net;
	}

	public static MarkedPetrinet convertTree(ProcessTree tree) {
		PetrinetWithMarkings pn = null;
		try {
			pn = ProcessTree2Petrinet.convert(tree);
		} catch (NotYetImplementedException e) {
			e.printStackTrace();
		} catch (InvalidProcessTreeException e) {
			e.printStackTrace();
		}
		return new MarkedPetrinet(pn.petrinet, pn.initialMarking, pn.finalMarking);
	}

	public String toString() {
		return "ReLocIM";
	}

	public RelabelParamPT getRelabelParam() {
		return relabelParam;
	}

	public void setRelabelParam(RelabelParamPT relabelParam) {
		this.relabelParam = relabelParam;
	}

	public XLog getLocalRefinedLog() {
		return refinedSublog;
	}

	public RelabelModel getRelabelModel() {
		return this.rmodel;
	}

}
