package org.processmining.xlu.relabel.plugin.matching;

import gnu.trove.map.hash.THashMap;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.model.XLog;
import org.processmining.xlu.matching.models.RepEGraph;
import org.processmining.xlu.matching.models.Tree;
import org.processmining.xlu.matching.models.collection.RepEGraphs;
import org.processmining.xlu.matching.plugins.EGFusionPlugin;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelCoreParam;

public class ClusterRefinerUsingFusing implements ClusterRefiner {

	protected RelabelModel model;

	public ClusterRefinerUsingFusing() {

	}

	@SuppressWarnings({ "cast", "unchecked" })
	public Map<Integer, Integer> cluster(XLog log, RelabelModel model,
			RelabelCoreParam param) {
		this.model = model;
		ClusterRefinerParam clusterParam = param.getClusterParam();

		Map<Integer, Integer> mapTrcId2ClsId = new THashMap<Integer, Integer>();

		if (clusterParam.getNumOfClusters() > 1) {

			if (model.getGraphs() == null) {
				EGFusionPlugin plugin = new EGFusionPlugin();
//				RepEGraphs repgraphs = plugin.fuseIncGraphs(null, model.getMappings());
				RepEGraphs repgraphs = plugin.fuseIncGraphsIncFilter(null, model.getMappings());
				model.setGraphs(repgraphs);
			}

			Tree<RepEGraph> tree = model.getGraphs().getHierarchy();
			List<Tree<RepEGraph>> parents = new ArrayList<Tree<RepEGraph>>();

			parents.add(tree);
			for (int i = 0; i < clusterParam.getNumOfClusters(); i++) {
				List<Tree<RepEGraph>> children = new ArrayList<Tree<RepEGraph>>();
				for (Tree<RepEGraph> parent : parents) {
					if (parent.getChildCount() == 0
							|| (parent.getObject() != null && parent.getObject()
									.getCost() == 0)) {
						children.add(parent);
						continue;
					}

					for (int j = 0; j < parent.getChildCount(); j++) {
						//				RepEGraph g = ((Tree<RepEGraph>) parent.getChildAt(j)).getObject();
						children.add(((Tree<RepEGraph>) parent.getChildAt(j)));
					}

				}
				parents = children;
				if (parents.size() >= clusterParam.getNumOfClusters()) {
					break;
				}
			}

			for (int i = 0; i < parents.size(); i++) {
				RepEGraph g = ((Tree<RepEGraph>) parents.get(i)).getObject();
				for (int id : g.getRepresentedGraphIds()) {
					mapTrcId2ClsId.put(id, i);
				}
			}
			model.setMapTrcToCls(mapTrcId2ClsId);
		}

		return mapTrcId2ClsId;
	}

}
