package org.processmining.xlu.relabel.plugin.matching;

import gnu.trove.map.hash.THashMap;

import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.processmining.plugins.InductiveMiner.graphs.ConnectedComponents;
import org.processmining.plugins.InductiveMiner.graphs.Graph;
import org.processmining.plugins.InductiveMiner.graphs.GraphImplQuadratic;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.param.RelabelCoreParam;

public class ClusterRefinerUsingMatching implements ClusterRefiner {

	protected RelabelModel model;

	public ClusterRefinerUsingMatching() {

	}

	@SuppressWarnings({ "cast", "unchecked" })
	public Map<Integer, Integer> cluster(XLog log, RelabelModel model,
			RelabelCoreParam param) {
		this.model = model;
		ClusterRefinerParam clusterParam = param.getClusterParam();
		int maxCost = clusterParam.getMaxCostOfEdge();

		Map<Integer, Integer> mapTrcId2ClsId = new THashMap<Integer, Integer>();

		Graph<Integer> orgGraph = model.getTraceGraph();

		/* ---------------------------------
		 * update graph with edges within the max cost
		 * --------------------------------- */
		Graph<Integer> graph = new GraphImplQuadratic<Integer>(Integer.class);
		for (Integer n : orgGraph.getVertices()) {
			graph.addVertex(n);
		}
		for (Long e : orgGraph.getEdges()) {

			int max = model.getMappingsInfo().getMax();
			if (max == 0 || (orgGraph.getEdgeWeight(e) - 1) * 100 / max <= maxCost) {
//			if (orgGraph.getEdgeWeight(e) - 1 <= maxCost) {
				graph.addEdge(orgGraph.getEdgeSource(e), orgGraph.getEdgeTarget(e),
						orgGraph.getEdgeWeight(e));
			}
		}

		Set<Set<Integer>> subgraphs = ConnectedComponents.compute(graph);
		int i = 0;
		for (Set<Integer> cls : subgraphs) {
			for (Integer id : cls) {
				mapTrcId2ClsId.put(id, i);
			}
			i++;
		}
		model.setMapTrcToCls(mapTrcId2ClsId);

		return mapTrcId2ClsId;
	}

}
