package org.processmining.xlu.relabel.plugin.matching;

import gnu.trove.map.hash.THashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XLog;
import org.processmining.plugins.InductiveMiner.graphs.ConnectedComponents;
import org.processmining.plugins.InductiveMiner.graphs.Graph;
import org.processmining.plugins.InductiveMiner.graphs.GraphImplQuadratic;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.model.TEvent;
import org.processmining.xlu.relabel.param.RelabelCoreParam;

public class LabelRefinerUsingMatching implements LabelRefiner {

	protected RelabelModel model;

	public Map<Integer, Map<Integer, Integer>> labelEvents(XLog log, RelabelModel model,
			RelabelCoreParam param) {
		this.model = model;

//		XLog log = (XLog) orglog.clone();
		Map<XEventClass, Graph<TEvent>> mapLabel2Graph = model.getMapLabel2Graph();

		Map<Integer, Map<Integer, Integer>> res = new THashMap<Integer, Map<Integer, Integer>>();

		LabelRefinerParam relabelerParam = param.getRelabelParam();
//		List<XEvent> tobeRemoved = new ArrayList<XEvent>();
		int threshold = relabelerParam.getUnfoldingThreshold();
		int maxCost = relabelerParam.getMaxCostOfEdge();

//		int progress = 0;

		for (Entry<XEventClass, Graph<TEvent>> entry : mapLabel2Graph.entrySet()) {
			if (!model.getClasses().contains(entry.getKey().toString())) {
				continue;
			}
////			if (progress % 50 == 0) {
//			System.out.println("Progressed graph " + progress);
////			}
//			progress++;

			XEventClass eclass = entry.getKey();
			Graph<TEvent> orgGraph = entry.getValue();

			/* ---------------------------------
			 * update graph with edges within the max cost
			 * --------------------------------- */
			Graph<TEvent> graph = new GraphImplQuadratic<TEvent>(TEvent.class);
			for (TEvent n : orgGraph.getVertices()) {
				graph.addVertex(n);
			}
			for (Long e : orgGraph.getEdges()) {
				
				int max = model.getMappingsInfo().getMax();
				if (max == 0 || (orgGraph.getEdgeWeight(e) - 1) * 100 / max  <= maxCost) {				
//				if (orgGraph.getEdgeWeight(e) - 1 <= maxCost) {
					graph.addEdge(orgGraph.getEdgeSource(e), orgGraph.getEdgeTarget(e),
							orgGraph.getEdgeWeight(e));
				}
			}

			Set<Set<TEvent>> subgraphs = null;

//			if (param.isClustering()) {

			/* --------------------------------------------------------
			 * Advanced way to group events : do cluster 
			 * ------------------------------------------------------- */
////			Cluster cluster = createClusters(graph);
//				Cluster cluster = mapLabel2Cluster.get(eclass);
//				int num = param.getEventClassNum(eclass);
//				subgraphs = getSubgraphs(cluster, num);
//			} else {
			/* --------------------------------------------------------
			 * Naive way to group events : take connected subgraphs 
			 * ------------------------------------------------------- */
			// !!BUG? Edge with cost 0 is not counted as an edge. 
			subgraphs = ConnectedComponents.compute(graph);
//			} 
			/*---------------------------------------------------------
			 *  Retain all
			 *---------------------------------------------------------*/

			int maxSizeSubgraph = Integer.MIN_VALUE;

			/* ----------------------------------
			 *  Compute all statistics
			 * ---------------------------------*/
			List<Collection<TEvent>> sortedSubgraphs = new ArrayList<>();
			for (Collection<TEvent> subgraph : subgraphs) {
				maxSizeSubgraph = subgraph.size() > maxSizeSubgraph ? subgraph.size()
						: maxSizeSubgraph;
				sortedSubgraphs.add(subgraph);
			}
			java.util.Collections.sort(sortedSubgraphs,
					new Comparator<Collection<TEvent>>() {
						public int compare(Collection<TEvent> o1, Collection<TEvent> o2) {
							int d = getAvgIndex(o1) - getAvgIndex(o2);
							return d;
						}

						private int getAvgIndex(Collection<TEvent> o1) {
							int avgIndex = 0;
							for (TEvent e : o1) {
								avgIndex += e.getEventIndex();
							}
							avgIndex = avgIndex / o1.size();
							return avgIndex;
						}
					});

			int i = 0;
			for (Collection<TEvent> subgraph : sortedSubgraphs) {
				if (subgraph.size() >= (maxSizeSubgraph * threshold / 100.0)) {
					i++;
				}
				for (TEvent eIndices : subgraph) {
					Map<Integer, Integer> mapEvt2Cls = res.get(eIndices.getTrIndex());

					if (mapEvt2Cls == null) {
						mapEvt2Cls = new THashMap<Integer, Integer>();
					}
					mapEvt2Cls.put(eIndices.getEventIndex(), i);
					res.put(eIndices.getTrIndex(), mapEvt2Cls);
				}
			}
		}
		return res;
	}

}
