package org.processmining.xlu.relabel.plugin.matching;

import edu.uci.ics.jung.graph.UndirectedSparseGraph;
import gnu.trove.map.hash.THashMap;
import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.swing.JFrame;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.InductiveMiner.graphs.Graph;
import org.processmining.plugins.InductiveMiner.graphs.GraphImplQuadratic;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.collection.EGraphsWithMappings;
import org.processmining.xlu.matching.models.collection.MatchMappings;
import org.processmining.xlu.matching.plugins.EGMatchPlugin;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.EGMatchConfiguration;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.model.TEvent;
import org.processmining.xlu.relabel.param.RelabelCoreParam;

import com.apporiented.algorithm.clustering.AverageLinkageStrategy;
import com.apporiented.algorithm.clustering.Cluster;
import com.apporiented.algorithm.clustering.ClusteringAlgorithm;
import com.apporiented.algorithm.clustering.DefaultClusteringAlgorithm;
import com.apporiented.algorithm.clustering.visualization.DendrogramPanel;

public class ModelCreatorUsingMatching implements ModelCreator {

	public static final String IMPRECISE_LABEL = "ImpreciseLabel";
	RelabelModel model;
	boolean isLazy = false;

//	public RelabelerUsingMatching(EGMatchConfiguration config){
//		this.conf = config;
//	}

	public RelabelModel createRelabelModel(XLog log, Set<String> labels,
			RelabelCoreParam param) {
		if (model != null && isLazy) {
			return model;
		}
//		if (labels == null || labels.isEmpty()) {
//			return null;
//		}
		RelabelModel model = new RelabelModel();

		EGMatchConfiguration conf = param.getRelabelerParam().getMatchingConfig();
		conf.setFocusOnMatchingLabels(labels);
		System.out.println("Matching using config : " + conf.toString());

		EGraphsWithMappings graphsWithMappings = (new EGMatchPlugin()).matchingTraces(
				(XLog) log.clone(), conf);

		EGraphMappingsInfo info = EGraphMappingsInfo
				.createMappingInfo(graphsWithMappings);
//		debugPrint(info);

		MatchMappings mapGraphsToMapping = graphsWithMappings.getMappings();
//		EGMatchConfiguration conf = graphsWithMappings.getMatchConfiguration();

		Map<MatchMapping, Integer> mapToNewCost = recomputeCostForMappings(graphsWithMappings);
		info.setMapToCost(mapToNewCost);
//		__testMappings(graphsWithMappings, mapToNewCost);

//		PLog plog = graphsWithMappings.getPlog();
//		XLog log = plog.getXLog();

		/*---------------------------------------------------------
		 *  Create a graph of events for each label (XEventClass)
		 *  that share the label across all traces
		 *---------------------------------------------------------*/
		Graph<Integer> traceGraph = computeTraceGraph(mapGraphsToMapping, log, mapToNewCost);

		
		/*---------------------------------------------------------
		 *  Create a graph of events for each label (XEventClass)
		 *  that share the label across all traces
		 *---------------------------------------------------------*/
		Map<XEventClass, Graph<TEvent>> mapLabel2Graph = computeRelatedGraphForEventClass(
				labels, conf.getClasses(), mapGraphsToMapping, log, mapToNewCost);

		model.setMappings(graphsWithMappings);
		model.setMappingsInfo(info);
//		if (param.isClustering()) {
//			// this currently cluster one event class, not subtraces..
////			Map<XEventClass, Cluster> mapLabel2Cluster = new THashMap<XEventClass, Cluster>();
////			for (Entry<XEventClass, Graph<TEvent>> entry : mapLabel2Graph.entrySet()) {
////				Cluster cluster = createClusters(entry.getValue());
////				mapLabel2Cluster.put(entry.getKey(), cluster);
////			}
////			model.setMapLabel2Cluster(mapLabel2Cluster);
//
//			EGFusionPlugin plugin = new EGFusionPlugin();
//			RepEGraphs repgraphs = plugin.fuseIncGraphs(null, graphsWithMappings);
//			model.setGraphs(repgraphs);
//
//		} // else 

		model.setMapLabel2Graph(mapLabel2Graph);

		model.setTraceGraph(traceGraph);
		
		model.setOrigLog(log);
		model.setClasses(labels);
		this.model = model;
		return model;
	}



	private Map<MatchMapping, Integer> recomputeCostForMappings(
			EGraphsWithMappings graphsWithMappings) {
		Map<MatchMapping, Integer> resMap = new THashMap<MatchMapping, Integer>();

		CostStrategy costfunction = graphsWithMappings.getMatchConfiguration()
				.getCostStrategy();

		MatchMappings mappings = graphsWithMappings.getMappings();
		int size = graphsWithMappings.getPlog().size();
		for (int ti = 0; ti < size; ti++) {
			for (int tj = ti + 1; tj < size; tj++) {
				MatchMapping mapping = mappings.getAnOptimalMapping(ti, tj);

				int totalCost = 0;
				for (Entry<Integer, Integer> match : mapping.getMatchs()) {
					String ilabel = mapping.getSourceGraph().getNodeById(match.getKey())
							.getLabel();
					if (graphsWithMappings.getMatchConfiguration().getLabels()
							.contains(ilabel)) {
						Integer costMatch = costfunction.getWeightedCostofMatchedPair(
								match.getKey(), match.getValue(), mapping);

						Integer costGlob = costfunction.getWeightedCostOfGlobalStructure(
								match.getKey(), match.getValue(), mapping);
						totalCost += costMatch + costGlob;
					}

				}

				for (Integer ei : mapping.getSourcesToNull()) {
					String ilabel = mapping.getSourceGraph().getNodeById(ei).getLabel();
					if (graphsWithMappings.getMatchConfiguration().getLabels()
							.contains(ilabel)) {
						Integer costNoMatch = costfunction.getWeightedCostNoMatch(ei,
								mapping.getSourceGraph(), mapping);

						totalCost += costNoMatch;
					}

				}

				for (Integer ej : mapping.getTargetToNull()) {
					String jlabel = mapping.getTargetGraph().getNodeById(ej).getLabel();
					if (graphsWithMappings.getMatchConfiguration().getLabels()
							.contains(jlabel)) {
						Integer costNoMatch = costfunction.getWeightedCostNoMatch(ej,
								mapping.getTargetGraph(), mapping);

						totalCost += costNoMatch;
					}

				}
				resMap.put(mapping, totalCost);
			}
		}
		return resMap;

	}


	public XLog applyRelabeling(RelabelModel model, RelabelCoreParam param) {
		ClusterRefiner cluster = param.getClusterMethod();
		LabelRefiner relabel = param.getRelabelMethod();

//		XLog origlog = (XLog) model.getOrigLog().clone();
//		XLog log = model.getOrigLog();
//		model.setOrigLog(origlog);
		XLog log = (XLog) model.getOrigLog().clone();

		Map<Integer, Integer> mapTrc2cls = cluster.cluster(log, model, param);
		Map<Integer, Map<Integer, Integer>> mapEvt2cls = relabel.labelEvents(log, model,
				param);

		Set<String> labels = model.getClasses();
		XEventClassifier classifier = param.getClassifier();

		for (int i = 0; i < log.size(); i++) {
			XTrace t = log.get(i);
			for (int j = 0; j < t.size(); j++) {
				XEvent e = t.get(j);
				String label = XConceptExtension.instance().extractName(e);
				XAttribute a = XFactoryRegistry.instance().currentDefault()
						.createAttributeLiteral(IMPRECISE_LABEL, label, null);
				e.getAttributes().put(IMPRECISE_LABEL, a);

				if (!labels.contains(classifier.getClassIdentity(e))) {
					continue;
				}

				String newLabel = label.toString();

				Map<Integer, Integer> map = mapEvt2cls.get(i);
				if (map != null && map.containsKey(j)) {
					newLabel = newLabel + "_X_" + map.get(j);
				}

				Integer traceCluster = mapTrc2cls.get(i);
				if (traceCluster != null) {
					newLabel = newLabel + "_C_" + traceCluster;
				}

				XConceptExtension.instance().assignName(e, newLabel);
			}
		}

		return log;
	}

//	public XLog applyRelabeling(RelabelModel model, RelabelParam param) {
//		XLog log = (XLog) model.getOrigLog().clone();
//		Map<XEventClass, Graph<TEvent>> mapLabel2Graph = model.getMapLabel2Graph();
//		
////		Map<XEventClass, Cluster> mapLabel2Cluster = model.getMapLabel2Cluster();
//
////		if (param.isClustering() && param.getNumOfClusters() > 1) {
////			updateClusters(model, param);
////		}
//
//		List<XEvent> tobeRemoved = new ArrayList<XEvent>();
//		int threshold = param.getUnfoldingThreshold();
//		int maxCost = param.getMaxCost();
//		int progress = 0;
//		for (Entry<XEventClass, Graph<TEvent>> entry : mapLabel2Graph.entrySet()) {
//			if (!model.getClasses().contains(entry.getKey().toString())) {
//				continue;
//			}
////			if (progress % 50 == 0) {
//			System.out.println("Progressed graph " + progress);
////			}
//			progress++;
//
//			XEventClass eclass = entry.getKey();
//			Graph<TEvent> orgGraph = entry.getValue();
//
//			Graph<TEvent> graph = new GraphImplQuadratic<TEvent>(TEvent.class);
//			for (TEvent n : orgGraph.getVertices()) {
//				graph.addVertex(n);
//			}
//			for (Long e : orgGraph.getEdges()) {
//				if (orgGraph.getEdgeWeight(e) - 1 <= maxCost) {
//					graph.addEdge(orgGraph.getEdgeSource(e), orgGraph.getEdgeTarget(e),
//							orgGraph.getEdgeWeight(e));
//				}
//			}
//
//			Set<Set<TEvent>> subgraphs = null;
//
////			if (param.isClustering()) {
//
//			/* --------------------------------------------------------
//			 * Advanced way to group events : do cluster 
//			 * ------------------------------------------------------- */
//////			Cluster cluster = createClusters(graph);
////				Cluster cluster = mapLabel2Cluster.get(eclass);
////				int num = param.getEventClassNum(eclass);
////				subgraphs = getSubgraphs(cluster, num);
////			} else {
//			/* --------------------------------------------------------
//			 * Naive way to group events : take connected subgraphs 
//			 * ------------------------------------------------------- */
//			// !!BUG? Edge with cost 0 is not counted as an edge. 
//			subgraphs = ConnectedComponents.compute(graph);
////			} 
//			/*---------------------------------------------------------
//			 *  Retain all
//			 *---------------------------------------------------------*/
//
//			int maxSizeSubgraph = Integer.MIN_VALUE;
//
//			/* ----------------------------------
//			 *  Compute all statistics
//			 * ---------------------------------*/
//			List<Collection<TEvent>> sortedSubgraphs = new ArrayList<>();
//			for (Collection<TEvent> subgraph : subgraphs) {
//				maxSizeSubgraph = subgraph.size() > maxSizeSubgraph ? subgraph.size()
//						: maxSizeSubgraph;
//			}
////			java.util.Collections.sort(list, c);
//
//			int i = 0;
//			for (Collection<TEvent> subgraph : subgraphs) {
//				if (subgraph.size() >= (maxSizeSubgraph * threshold / 100.0)
//						|| param.isTraceModelMode()) {
////				if (subgraph.size() == maxSizeSubgraph) {
//					i++;
//					String label = eclass.toString() + "_X_" + i;
//
//					for (TEvent eIndices : subgraph) {
//						String eventLabel = label;
//						if (param.isClustering() && param.getNumOfClusters() > 1) {
//							int clusterId = model.getMapTrcToCls().get(
//									eIndices.getTrIndex());
//							eventLabel += "_C_" + clusterId;
//						}
//
//						XEvent e = log.get(eIndices.getTrIndex()).get(
//								eIndices.getEventIndex());
//						XConceptExtension.instance().assignName(e, eventLabel);
//					}
//				} else {
//					/* ----------------------------------
//					 *  Remove all infrequent events
//					 * ---------------------------------*/
////					for (TEvent eIndices : subgraph) {
////						XEvent e = log.get(eIndices.getTrIndex()).get(
////								eIndices.getEventIndex());
////						tobeRemoved.add(e);
////					}
//				}
//			}
//
////			/*---------------------------------------------------------
////			 *  Retain only the largest connected subgraph
////			 *---------------------------------------------------------*/
////			int maxSize = Integer.MIN_VALUE;
////			Collection<XEvent> maxSubgraph = null;
////			for (Collection<XEvent> subgraph : subgraphs) {
////				if (subgraph.size() > maxSize) {
////					maxSize = subgraph.size();
////					maxSubgraph = subgraph;
////				}
////			}
////			int i = 1;
////			for (Collection<XEvent> subgraph : subgraphs) {
////				String label = eclass.toString() + "-" + i++;
////				if (subgraph.size() <= 1) {
////					tobeRemoved.add(subgraph.iterator().next());
////				} else if (subgraph.equals(maxSubgraph)) {
////					for (XEvent eIndices : subgraph) {
////						XConceptExtension.instance().assignName(eIndices, label);
////					}
////				} else {
////					for (XEvent e : subgraph) {
////						tobeRemoved.add(e);
////					}
////				}
////			}
//
////			/*---------------------------------------------------------
////			 *  Do Hierachical cluster subgraphs
////			 *---------------------------------------------------------*/
//
////			/*---------------------------------------------------------
////			 *  Visualize subgraphs
////			 *---------------------------------------------------------*/
////			edu.uci.ics.jung.graph.Graph<TEvent, Long> visGraph = transformGraph(graph);
////			VisualizationImageServer vs =
////				      new VisualizationImageServer(
////				        new KKLayout(visGraph), new Dimension(400, 400));
////				 
////				    JFrame frame = new JFrame();
////				    frame.getContentPane().add(vs);
////				    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
////				    frame.pack();
////				    frame.setVisible(true);	
//
//		}
//
//		for (XTrace t : log) {
//			t.removeAll(tobeRemoved);
//		}
//		log.getClassifiers().add(new XEventAttributeClassifier("Relabel", "Relabel"));
//		return log;
//	}

//	private static void updateClusters(RelabelModel model, RelabelParam param) {
//		//		List<RepEGraph> clusters = repgraphs.getGraphs();	
//		Tree<RepEGraph> tree = model.getGraphs().getHierarchy();
//		//		List<RepEGraph> clusters = new ArrayList<RepEGraph>();
//		//		if(tree.getChildCount() > param.getNumOfClusters() && param.getNumOfClusters() > 1){
//		List<Tree<RepEGraph>> parents = new ArrayList<Tree<RepEGraph>>();
//		parents.add(tree);
//		for (int i = 1; i < param.getNumOfClusters(); i++) {
//			List<Tree<RepEGraph>> children = new ArrayList<Tree<RepEGraph>>();
//			for (Tree<RepEGraph> parent : parents) {
//				if (parent.getChildCount() == 0
//						|| (parent.getObject() != null && parent.getObject().getCost() == 0)) {
//					children.add(parent);
//					continue;
//				}
//
//				for (int j = 0; j < parent.getChildCount(); j++) {
//					//				RepEGraph g = ((Tree<RepEGraph>) parent.getChildAt(j)).getObject();
//					children.add(((Tree<RepEGraph>) parent.getChildAt(j)));
//				}
//
//			}
//			parents = children;
//		}
//
//		Map<Integer, Integer> mapTrcId2ClsId = new THashMap<Integer, Integer>();
//		for (int i = 0; i < parents.size(); i++) {
//			RepEGraph g = ((Tree<RepEGraph>) parents.get(i)).getObject();
//			for (int id : g.getRepresentedGraphIds()) {
//				mapTrcId2ClsId.put(id, i);
//			}
//		}
//		model.setMapTrcToCls(mapTrcId2ClsId);
//	}

	/*
	 * TODO: this function belongs in the Cluster, make a cluster wrapper
	 */
	private static Set<Set<TEvent>> getSubgraphs(Cluster cluster, int num) {
//		System.out.println(cluster.getName());
		List<Cluster> clusters = new ArrayList<Cluster>();
		clusters.add(cluster);
		List<Cluster> childClusters = new ArrayList<Cluster>();
		while (num-- > 0) {

			for (Cluster c : clusters) {
				System.out.println(c.getDistance());
				if (c.getDistance() == null) {
					// Means c.size() = 1;
//					System.out.println("asdfae");
				}
				if (c.getChildren().size() > 0) {
					childClusters.addAll(c.getChildren());
				} else {
					childClusters.add(c);
				}
			}

			clusters.clear();
			clusters = childClusters;
			childClusters = new ArrayList<Cluster>();
		}

		Set<Set<TEvent>> subgraphs = new THashSet<>();
		for (Cluster c : clusters) {
			Set<TEvent> subgraph = new THashSet<>();

			String[] events = c.toString().substring(8).split("&");
			for (String eString : events) {
				int split = eString.indexOf(",");
				int traceIndex = Integer.parseInt(eString.substring(1, split));
				int eventIndex = Integer.parseInt(eString.substring(split + 1,
						eString.length() - 1));
				TEvent e = new TEvent(traceIndex, eventIndex);
				subgraph.add(e);
			}
			subgraphs.add(subgraph);
		}

		return subgraphs;
	}

	private static Cluster createClusters(Graph<TEvent> graph) {
		int size = graph.getVertices().length;
		String[] names = new String[size];
		double[][] distances = new double[size][size];
		for (int j = 0; j < graph.getVertices().length; j++) {
			for (int k = 0; k < graph.getVertices().length; k++) {
				distances[j][k] = 5;
			}
		}

		for (int j = 0; j < graph.getVertices().length; j++) {
			names[j] = graph.getVertexOfIndex(j).toString();
			distances[j][j] = 0;
		}

		for (Long l : graph.getEdges()) {
			distances[graph.getEdgeSourceIndex(l)][graph.getEdgeTargetIndex(l)] = 1;
			distances[graph.getEdgeTargetIndex(l)][graph.getEdgeSourceIndex(l)] = 1;
		}
		ClusteringAlgorithm alg = new DefaultClusteringAlgorithm();
		com.apporiented.algorithm.clustering.Cluster cluster = alg.performClustering(
				distances, names, new AverageLinkageStrategy());

		boolean isVisualizeDendrogram = false;
		if (isVisualizeDendrogram) {
			DendrogramPanel dp = new DendrogramPanel();
			dp.setModel(cluster);
			JFrame frame = new JFrame();
			frame.getContentPane().add(dp);
			frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
			frame.pack();
			frame.setVisible(true);
		}
		return cluster;
	}

	private static edu.uci.ics.jung.graph.Graph<TEvent, Long> transformGraph(
			Graph<TEvent> graph) {
		edu.uci.ics.jung.graph.Graph<TEvent, Long> visGraph = new UndirectedSparseGraph<TEvent, Long>();
		for (TEvent e : graph.getVertices()) {
			visGraph.addVertex(e);
		}

		for (Long l : graph.getEdges()) {
			visGraph.addEdge(l, graph.getEdgeSource(l), graph.getEdgeTarget(l));
		}
		return visGraph;
	}
	
	private Graph<Integer> computeTraceGraph(MatchMappings mapGraphsToMapping, XLog log,
			Map<MatchMapping, Integer> mapToNewCost) {
		Graph<Integer> graph = new GraphImplQuadratic<Integer>(Integer.class);
		for (int traceIndex = 0; traceIndex < log.size(); traceIndex++) {
			graph.addVertex(traceIndex);
			for (int oTraceIndex = 0; oTraceIndex < traceIndex; oTraceIndex++) {
				MatchMapping mapping = mapGraphsToMapping.getAnOptimalMapping(
						oTraceIndex, traceIndex);
				
				int cost = mapToNewCost.get(mapping);
				graph.addEdge(traceIndex, oTraceIndex, cost + 1); // '+1' for because there is a bug
			}
			
		}
		return graph;
	}

	private static Map<XEventClass, Graph<TEvent>> computeRelatedGraphForEventClass(
			Set<String> labels, XEventClasses classes, MatchMappings mapGraphsToMapping,
			XLog log, Map<MatchMapping, Integer> mapToNewCost) {

		Map<XEventClass, Graph<TEvent>> mapLabel2Graph = new THashMap<>();

		for (int traceIndex = 0; traceIndex < log.size(); traceIndex++) {
			if (traceIndex % 100 == 0) {
				System.out.println("Relabel started with trace " + traceIndex);
			}
			XTrace t = log.get(traceIndex);

			/*---------------------------------------------------------
			 *  Create nodes for each trace. 
			 *---------------------------------------------------------*/
			for (int eventIndex = 0; eventIndex < t.size(); eventIndex++) {
				XEvent e = t.get(eventIndex);
				TEvent indexedEv = new TEvent(traceIndex, eventIndex);
				XEventClass eclass = classes.getClassOf(e);
				if (!labels.contains(eclass.toString())) {
					continue;
				}

				Graph<TEvent> graph = mapLabel2Graph.get(eclass);
				if (graph == null) {
					graph = new GraphImplQuadratic<TEvent>(TEvent.class);
					mapLabel2Graph.put(eclass, graph);
				}
				graph.addVertex(indexedEv);
			}

			/*---------------------------------------------------------
			 *  Adding edges.  
			 *---------------------------------------------------------*/
			for (int oTraceIndex = 0; oTraceIndex < traceIndex; oTraceIndex++) {
				XTrace otherTrace = log.get(oTraceIndex);

				MatchMapping mapping = mapGraphsToMapping.getAnOptimalMapping(
						oTraceIndex, traceIndex);

				for (Entry<Integer, Integer> pair : mapping.getMatchs()) {
					if (pair.getValue() >= t.size()) {
						continue;
					}
					XEvent sourceEvt = otherTrace.get(pair.getKey());
//					XEvent targetEvt = t.get(pair.getValue());
					TEvent sourceEv = new TEvent(oTraceIndex, pair.getKey());
					TEvent targetEv = new TEvent(traceIndex, pair.getValue());
					XEventClass eclass = classes.getClassOf(sourceEvt);
					if (!labels.contains(eclass.toString())) {
						continue;
					}
					Graph<TEvent> graph = mapLabel2Graph.get(eclass);

//					if (!param.isTraceModelMode() || mapping.getCost() == 0) {
//					int cost = mapping.getCost();
					int cost = mapToNewCost.get(mapping);
					graph.addEdge(sourceEv, targetEv, cost + 1); // '+1' for because there is a bug
//					} //else {
//						System.out.println("Edge not added " + mapping.getCost());
					//}
				}
			}
		}

		return mapLabel2Graph;
	}

}
