package org.processmining.xlu.relabel.visualization;

import java.util.Collection;
import java.util.List;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.partialorder.plugins.vis.PartialVisualType;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IMProcessTree;
import org.processmining.plugins.inductiveVisualMiner.InductiveVisualMiner.InductiveVisualMinerLauncher;
import org.processmining.processtree.ProcessTree;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryAlgorithm;
import org.processmining.xlu.experiment.discovery.ExpDiscoveryInductiveMiner;
import org.processmining.xlu.experiment.discovery.MarkedPetrinet;
import org.processmining.xlu.matching.plugins.EGMatchFusionAndRelabelPlugin;
import org.processmining.xlu.relabel.model.RelabelModel;
import org.processmining.xlu.relabel.model.RelabelVisModel;
import org.processmining.xlu.relabel.param.RelabelParam;
import org.processmining.xlu.relabel.plugin.detection.RelabelScope;
import org.processmining.xlu.relabel.plugin.matching.ReplacerIMLocal;

import gnu.trove.set.hash.THashSet;

public class RelabelControllerModel implements RelabelVisModel {
	public RelabelModel model;
	public RelabelParam param;
	public RelabelScope scope;
	public MarkedPetrinet net;
	public XLog newlog;
	
	
	private Object visType = new ExpDiscoveryInductiveMiner();

	public RelabelControllerModel(RelabelModel model, RelabelScope scope, RelabelParam param) {
		this.model = model;
		this.scope = scope;
		this.param = param;
	}

	public Collection<String> getAllLabels() {

		XEventClasses classes = XEventClasses.deriveEventClasses(param.getClassifier(), model.getOrigLog());
		THashSet<String> values = new THashSet<>();
		for (XEventClass c : classes.getClasses()) {
			values.add(c.toString());
		}

		return values;
	}

	public Collection<String> getSelectedLabels() {
		return scope.getLabelScope();
	}

	public int getMaxCostOfEdge() {
		return param.getRelabelParam().getMaxCostOfEdge();
	}

	public int getUnfoldingThreshold() {
		return param.getRelabelParam().getUnfoldingThreshold();
	}

	public Petrinet getPetrinet() {
		return net.getPetrinet();
	}

	public XLog getNewlog() {
		if(newlog == null){
			return computeNewLog();
		}
		return newlog;
	}

	public void setMaxCostOfEdge(int threshold) {
		param.getRelabelParam().setMaxCostOfEdge(threshold);		
	}

	public void setNumOfClusters(int threshold) {
		param.getClusterParam().setNumOfClusters(threshold);		
	}

	public void setUnfoldingThreshold(int threshold) {
		param.getRelabelParam().setUnfoldingThreshold(threshold);
		
	}

	public void setEventClassNumb(XEventClass evc, int threshold) {
		param.getRelabelParam().setEventClassNumb(evc, threshold);
		
	}

	public XLog computeNewLog() {
		newlog = param.getRelabeler().applyRelabeling(model, param);
		return newlog;
	}

	public Object computeNewProcessModel(PluginContext context) {
		computeNewLog();
		Object net = null;
		if(getDiscoveryAlg() != null){
			this.net = getDiscoveryAlg().minePetrinet(context, newlog);
			net = this.net;
		} else if(isUsingIvm()){ // is using Inductive Visual Miner
			// Get a process tree
//			ProcessTree tree = null;
			MiningParametersIM param = new MiningParametersIM();
//			param.setClassifier(classifier);
			ProcessTree tree = IMProcessTree.mineProcessTree(newlog, param);
			ReplacerIMLocal.changeTransitionLabels(tree);
			ExpDiscoveryInductiveMiner.postProcessing(tree.getRoot());
			// 
			net = InductiveVisualMinerLauncher.launcherPro(this.scope.getOrigLog(), tree);
			
		}else {
			EGMatchFusionAndRelabelPlugin alg = new EGMatchFusionAndRelabelPlugin();
			net = alg.computeTraceMatchingDefault(context, newlog);
			
		}
		return net;
	}


	public RelabelScope getRelabelScope() {
		return this.scope;
	}

	public RelabelModel getRelabelModel() {
		return this.model;
	}

	public RelabelParam getRelabelParam() {
		return this.param;
	}

	public void setImpreciseLabelCandidate(List<String> newSelected) {
		this.scope.setLabelScope(newSelected);
		model = param.getRelabeler().createRelabelModel(model.getOrigLog(), this.scope.getLabelScope(), param);
		
	}

	public PartialVisualType getREGVisType() {
		if(isUsingREGs()){
			return (PartialVisualType)visType;
		}
		return null;
	}
	
	
	public Object getVisType() {
		return visType;
	}
	
	public ExpDiscoveryAlgorithm getDiscoveryAlg(){
		if(visType instanceof ExpDiscoveryAlgorithm){
			return (ExpDiscoveryAlgorithm)visType;
		}
		return null;
	}

	public void setVisType(Object visType) {
		this.visType = visType;
	}

	public boolean isUsingREGs() {
		return visType instanceof PartialVisualType;
	}
	
	private boolean isUsingIvm() {
		// TODO Auto-generated method stub
		return visType.equals("IvM (show deviations)");
	}

}