package org.processmining.xlu.relabel.visualization;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.fluxicon.slickerbox.ui.SlickerScrollBarUI;

public class RelabelImpreciseLabelCandidateView {

	String title;

	Map<String, Color> map = new HashMap<>();

	JPanel cpanel;

//	private PGraphController<?> controller;

	private Collection<String> values;
	private Collection<String> selectedValues;

	private RelabelMainPanel parent;

	private RelabelControllerModel controller;

	public RelabelImpreciseLabelCandidateView(RelabelMainPanel parent,
			RelabelControllerModel controller) {
		this.parent = parent;
		this.values = controller.getAllLabels();
		this.selectedValues = controller.getSelectedLabels();
		this.controller = controller;

	}

	public JPanel getPanel() {
		if (cpanel == null) {
			cpanel = new JPanel();
			cpanel.setLayout(new BorderLayout());
			cpanel.setForeground(Color.WHITE);
			cpanel.setBackground(Color.DARK_GRAY);
			cpanel.setAlignmentX(Component.LEFT_ALIGNMENT);
			
			JLabel title = RelabelMainPanel.createLabel(cpanel, "Imprecise Label Candidates :");

			cpanel.add(title, BorderLayout.NORTH);
			
			String[] array = values.toArray(new String[values.size()]);
			final JList<String> list = new JList<String>(array);
			JComponent comp = configureMultiSelectionList(list, "Select " + title,
					"Visualize ");

			list.getSelectionModel()
					.addListSelectionListener(new ListSelectionListener() {

						@SuppressWarnings("unchecked")
						public void valueChanged(ListSelectionEvent e) {
							if (e.getValueIsAdjusting() == false) {
								if (list.getSelectedIndex() == -1) {
									//No selection, disable fire button.
									//fireButton.setEnabled(false);
								} else {
//							if(Arrays.equals(selectedValues,list.getSelectedIndices())){
									List<String> newSelected = list
											.getSelectedValuesList();
									/* ---------------------------------
									 * Tell the controller model has changed
									 * --------------------------------- */
									controller.setImpreciseLabelCandidate(newSelected);

									/* ---------------------------------
									 * Tell the parent view they need to update
									 * --------------------------------- */
									parent.updateNewModel();
								}
							}
						}
					});

//			int index = list.
			int[] selected = new int[selectedValues.size()];
			int j = 0;
			for (int i = 0; i < array.length; i++) {
				if (selectedValues.contains(array[i])) {
					selected[j] = i;
					j++;
				}
			}
			list.setSelectedIndices(selected);

			list.setCellRenderer(new SelectedListCellRenderer());

			cpanel.add(comp, BorderLayout.CENTER);
		}

		return cpanel;
	}

	@SuppressWarnings("rawtypes")
	public static JComponent configureMultiSelectionList(JList list, String title,
			String description) {
		cofigureListBasic(list, title, description);
		list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		list.setSelectionInterval(0, list.getModel().getSize() - 1);
		return configureAnyScrollable(list, title, description);
	}

//	protected static Color colorListBgSelected = Color.LIGHT_GRAY;
	
	
	@SuppressWarnings("rawtypes")
	public static void cofigureListBasic(JList list, String title, String description) {
		list.setFont(list.getFont().deriveFont(13f));
		list.setBackground(Color.GRAY);
		list.setForeground(Color.WHITE);
		list.setFont(list.getFont().deriveFont(12f));
	}

	protected static JComponent configureAnyScrollable(JComponent scrollable,
			String title, String description) {
		JScrollPane listScrollPane = new JScrollPane(scrollable);
		listScrollPane.setVerticalScrollBarPolicy(
				ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
		listScrollPane.setHorizontalScrollBarPolicy(
				ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		listScrollPane
				.setViewportBorder(BorderFactory.createEmptyBorder());
		listScrollPane.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
		JScrollBar vBar = listScrollPane.getVerticalScrollBar();
		vBar.setUI(new SlickerScrollBarUI(vBar, Color.GRAY,
				new Color(30, 30, 30), new Color(80, 80, 80), 4, 12));
		listScrollPane.setBackground(Color.DARK_GRAY);

		return listScrollPane;
	}

	public class SelectedListCellRenderer extends DefaultListCellRenderer {
		private static final long serialVersionUID = -2089347684127611653L;

		@SuppressWarnings("rawtypes")
		@Override
		public Component getListCellRendererComponent(JList list, Object value, int index,
				boolean isSelected, boolean cellHasFocus) {
			Component c = super.getListCellRendererComponent(list, value, index,
					isSelected, cellHasFocus);

//			c.setForeground(Color.YELLOW);
			if(isSelected){
				c.setBackground(Color.DARK_GRAY.brighter());
				c.setForeground(Color.WHITE);
			}
			return c;
		}
	}
	
	protected static JComponent packLeftAligned(JComponent component) {
		JPanel packed = new JPanel();
		packed.setOpaque(false);
		packed.setBorder(BorderFactory.createEmptyBorder());
		packed.setLayout(new BoxLayout(packed, BoxLayout.X_AXIS));
		packed.add(component);
		packed.add(Box.createHorizontalGlue());
		return packed;
	}

}
